// -*- C++ -*-

// AudioIO.h
//
// Classes for sources and sinks of data
// that interface to real-time audio hardware
// for Audio Pattern Recognition Language (aprl)
// dpwe@media.mit.edu 1995feb02
//
/* $Header: /u/drspeech/src/aprl/RCS/AudioIO.h,v 1.3 1999/01/09 02:30:56 dpwe Exp $ @media.mit.edu Machine Listening Group
 */

#include "DataPipe.h"
extern "C" {
#include <dpwelib/audIO.h>
}

#define DFLT_SRATE 	(22050.0)
#define DFLT_BUFTIME 	(0.100)

class AudioSrc : public DataSource {
private:
    AUSTRUCT *au;
    int status;    // 0 = not started; 1 = running; -1 = paused
    double lastTime;
    double bufferTime;
    void init(void);
    double scale; 		// multiply the output by this.
    int chans;
public:
    AudioSrc(double srate = DFLT_SRATE, double buftime = DFLT_BUFTIME, int chs=1)
	: bufferTime(buftime), DataSource(srate, (int)(srate*chs*buftime)), chans(chs)
	{ au = 0; status = 0; scale = 1; /* fprintf(stderr, "ASr:0x%lx\n", this); */ };
    ~AudioSrc(void);

    // Provide fns to open and close audio on the fly
    void auOpen(void);
    void auClose(void);

    int Update(double time);
    int Bufconts(void)
        { int f = 0; if(au) f = AUBufConts(au); return f; }
    double getScale(void)	{ return scale; }
    void setScale(double sc)	{ scale = sc; }
};
	
class AudioSink : public DataSink {
private:
    AUSTRUCT *au;
    void init(void);
    double scale; 		// multiply the input by this.
    int chans;
    double sampleRate;
public:
    AudioSink(DataSource *in=0, int chs=1, double sr = 0) 
	: DataSink(in), chans(chs), sampleRate(sr)
        { au = 0; scale=1; /* fprintf(stderr,"ASk:0x%lx(0x%lx)\n",this,in); */}
    ~AudioSink(void);
    int Update(double time);
    int Bufconts(void)
        { int f = 0; if(au) f = AUBufConts(au); return f; }
    double getScale(void)	{ return scale; }
    void setScale(double sc)	{ scale = sc; }
};

	
