// -*- C++ -*-

// SoundfileIO.cc
//
// Classes for sources and sinks of data
// that interface to soundfile read/write
// for Audio Pattern Recognition Language (aprl)
// dpwe@media.mit.edu 1995feb02
//
#ifndef lint
static char rcsid[] = "$Header: /u/drspeech/src/tcltk/aprl/RCS/SoundfileIO.cc,v 1.2 1997/07/19 01:27:27 dpwe Exp $ @media.mit.edu Machine Listening Group";
#endif /* lint */

#include <stdio.h>
#include <assert.h>
#include "SoundfileIO.h"

int SoundfileSrc::newfile(char *filename)
{   /* A new filename has been specified.  Open the associated soundfile
       if it looks like it might exist.  Return 1 on success, else 0. */
    int rc = 0;
    if(filename != 0 && strlen(filename)>0) {
	if(isOpen()) {
	    close();
	}
	if(open(filename)>0) {
	    rc = 1;
	}
    }
    DataSource::SetSampleRate(Sound::getSampleRate());
    Sound::setFormat(DATAFORMAT); 
    Sound::setChannels(1);    
    return rc;
}

SoundfileSrc::SoundfileSrc(char *filename, double buftime /* = DFLT_BUFTIME */)
	: bufferTime(buftime), Sound("r")
{ 
    newfile(filename);
    Resize((int)(bufferTime * SampleRate())); 
    loop = 0; 
    scale = 1.0; 
}

int SoundfileSrc::Update(double time)
{ 
    int desSamps = min(Size(), (int)((time-lastTime)*SampleRate()));
    if(desSamps <= 0)	desSamps = Size(); // of the circ. buffer
    lastTime = time;
    int got = -1, remain = desSamps, totalGot = 0, thistime, n;
    DATA *d;
    PushHead();
    double sc = scale;
    int doscale = (sc!=1.0 && sc!=0.0);
    while(got && remain)  {
	assert(n = GetHeadPtr(&d));
	thistime = min(remain, n);
	if(Sound::isOpen()) {
	    got = Sound::readFrames(d, thistime);
	} else {
	    // Sound is not open - simulate read of nothing
	    got = 0;
	}
	if(doscale)   DATAscale(d, d, got, sc);
	if(loop) {
	    if(got == 0)  {
		Sound::frameSeek(0);
		got = -1;			// keep loopin'
	    }
	} else if(got < thistime) {
	    // not looping & didn't get everything;  
	    if(got==0 && Sound::isOpen()) {
		// got to EOF and not looping: close
		Sound::close();
	    }
	    // clear rest of buffer and pretend we filled it.
	    memset((void*)(d+got), 0, (thistime-got)*sizeof(DATA));
	    got = thistime;
	}
	UpdateHead(got);
	remain -= got;
	totalGot += got;
    }
    return totalGot; 
}

int SoundfileSink::Update(double /* time */)
{ 
    int totalDone = 0;
    if(ThereIsMore())  {
	if(!isOpen)
	    init();
	assert(Input()->PopTail()>=0);
	int todo  = Input()->Avai();
	DATA *d;
	int n;
	double sc = scale, isc = (1.0/scale);
	int doscale = (sc != 1.0 && sc != 0.0);
	while(todo)  {
	    n = min(todo, Input()->GetTailPtr(&d));
	    if(doscale)   DATAscale(d, d, n, sc);
	    writeFrames(d, n);
	    if(doscale)   DATAscale(d, d, n, isc);
	    /* have to restore input values as other objects 
	       may look at the data after us - we have no right 
	       to modify it */
	    Input()->UpdateTail(n);
	    todo -= n;
	    totalDone += n;
	}	
    }
    return totalDone; 
}

#ifdef MAIN //---------------------------- test code ----------------------

extern "C" {
    char *programName;
};

#include "AudioIO.h"
#include "FilterPipe.h"

void testSFIO(char *infile)
{
    SoundfileSrc  ss(infile);
    BandpassFilterPipe bp(1000.0, 8, BPF_2NDO, &ss);
    AudioSink     ak(&bp);
    SoundfileSink sk("tmp.aif", &bp);
    
    Scheduler s;
    
    s.Enschedule(&ss);
    s.Enschedule(&bp);
    s.Enschedule(&ak);
    s.Enschedule(&sk);
    
    while(ss.ThereIsMore())  {
	s.Update(0.0);
    }
}

void main(int argc, char **argv)
{
    programName = argv[0];
    if(argc < 2)  {
	fprintf(stderr, "Usage: %s infile\n", programName);
	exit(-1);
    }
    testSFIO(argv[1]);
}

#endif				/* MAIN */
