//
// tclScheduler.cc
//
// tcl interface to Scheduler object
//
// for aprl - audio pattern recognition language
// dpwe@media.mit.edu 1995feb08
//
static char rcsid[] = "$Header: /n/crab/da/dpwe/src/aprl/tclScheduler.cc,v 1.1 1996/11/18 18:42:29 dpwe Exp $ @media.mit.edu Machine Listening Group";

#include "tclAprlMisc.h"

#include "Scheduler.h"

// tcl object data structure

typedef struct {
    Tcl_Interp 	*interp;
    Tk_Window	tkwin;		// fake, needed for Tk_Config
    double	timestep;
    double	nowtime;
    Scheduler	*sched;		// the actual C++ object
    char	*command;	// Tcl expression evaluated each step
    int		doCmd;		// only eval command if this is nonzero
} tclScheduler;

// Config options

static Tk_ConfigSpec schedulerConfigSpecs[] = {
    {TK_CONFIG_DOUBLE, "-timestep", "timestep", "Timestep",
	"0.1", Tk_Offset(tclScheduler, timestep), 0},
    {TK_CONFIG_DOUBLE, "-nowtime", "nowtime", "Nowtime",
	"0.0", Tk_Offset(tclScheduler, nowtime), 0},
    {TK_CONFIG_STRING, "-command", "command", "Command", 
        "after 10 { %s step }", Tk_Offset(tclScheduler, command), 0},
    {TK_CONFIG_INT, "-docmd", "docmd", "DoCmd", 
        "0", Tk_Offset(tclScheduler, doCmd), 0}, 
    // have to default to no command for upwards compat - eccchh
    {TK_CONFIG_END, NULL, NULL, NULL, NULL, 0, 0}
};

extern "C" {

// tcl interface function definitions

int Scheduler_Init(Tcl_Interp *interp);
    /* add the tcl commands defined in this package to the interpreter */
int schedulerCommand(ClientData clientData, Tcl_Interp *interp, 
		     int argc, char **argv);
int schedulerProc(ClientData clientData, Tcl_Interp *interp, 
		  int argc, char **argv);
static int schedulerConfigure(Tcl_Interp *interp, tclScheduler *schedPtr, 
			      char *cmdName, int argc, char **argv, int flag);
void schedulerDelete(ClientData clientData);
}

// The single, global scheduler to which all procs are added
Scheduler *globalScheduler = 0;

int Scheduler_Init(Tcl_Interp *interp)
{   /* add the tcl commands defined in this package to the interpreter */
    Tcl_CreateCommand(interp, "scheduler", schedulerCommand, (ClientData) NULL,
		      (void (*)(ClientData)) NULL);
    // initialize the datasrc hash table
    Tcl_InitHashTable(&datasrcTable, TCL_STRING_KEYS);
    return TCL_OK;
}

int schedulerCommand(ClientData /* clientData */, Tcl_Interp *interp, 
		     int argc, char **argv)
{	/* tcl command to instance a new scheduler object */
    if(argc < 2) {
	Tcl_AppendResult(interp, "wrong # args; should be \"", 
			 argv[0], " objName ?options?\"", NULL);
	return TCL_ERROR;
    }
    // allocate a new object
    tclScheduler *schedulerPtr = new tclScheduler;
    schedulerPtr->sched = 0;	/* don't create object until configure */
    schedulerPtr->interp = interp;
    schedulerPtr->nowtime = 0.0;
    schedulerPtr->timestep= 0.0;	// had better get updated
    schedulerPtr->command = NULL;
    schedulerPtr->tkwin = Tk_MainWindow(interp);   // dummy for Tk_Config
    // put it into the tcl command namespace
    Tcl_CreateCommand(interp, argv[1], schedulerProc, 
		      (ClientData)schedulerPtr, schedulerDelete);
    // pass rest of command to configure
    if(schedulerConfigure(interp,schedulerPtr,argv[1],argc-2,argv+2,0)
       !=TCL_OK) {
	return TCL_ERROR;
    }
    interp->result = argv[1];
    return TCL_OK;
}

#define STRLEN 256

static int ExpandPercents(char *input, char *output, int outlen, 
			   int nsubs, char *subchrs, char **substrs)
{	/* Take a string containing %x type strings, and expand them.  
	   Recognize <nsubs> substitutions, defined by subchrs[i]; 
	   %<subchrs[i]> becomes <substrs[i]>. 
	   <output> is pre-allocated output string, <outlen> long; 
	   return 0 if that's not big enough. */
    int outrem = outlen;
    int ncpy;

#define ADDOUT(N,FROM)	if(N>=outrem)	return 0;	   \
                        else {  strncpy(output, FROM, N);  \
				output += N; *output = 0; outrem -= N; }

    *output = 0;
    while(1)  {
	char *pct = strchr(input, '%');
	if(pct==NULL)  {
	    ncpy = strlen(input);
	    ADDOUT(ncpy, input);
	    *output++ = 0;
	    return 1;
	} else {	// found a percent; copy up to it
	    ncpy = pct-input;
	    ADDOUT(ncpy, input);
	    input += ncpy; 
	    // find the substitution
	    char subc = input[1];
	    int i=0, didsub = 0;
	    while(i<nsubs && !didsub)  {
		if(subchrs[i] == subc)  {
		    ncpy = strlen(substrs[i]);
		    ADDOUT(ncpy,substrs[i]);
		    didsub = 1;
		}
		++i;
	    }
	    if(!didsub)  {
		ADDOUT(2, input);
	    }
	    // skip over the "%x" in input, however we dealt with it.
	    input += 2;
	}
    }
#undef ADDOUT
}

int schedulerProc(ClientData clientData, Tcl_Interp *interp, 
		  int argc, char **argv)
{	/* handle the commands supported by the schedulerObj */
    tclScheduler *schd = (tclScheduler *)clientData;
    int		result = TCL_OK;
    char	c;
    int		length;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " <command> ?args?\"", NULL);
	return TCL_ERROR;
    }
    c = argv[1][0];
    length = strlen(argv[1]);
    if( (c=='c') && (strncmp(argv[1], "configure", length)==0) )  {
	if(argc==2) {
	    result =Tk_ConfigureInfo(interp, schd->tkwin,schedulerConfigSpecs, 
				      (char *)schd, NULL, 0);
	} else if(argc==3) {
	    result =Tk_ConfigureInfo(interp, schd->tkwin,schedulerConfigSpecs, 
				      (char *)schd, argv[2], 0);
	} else {
	    result =schedulerConfigure(interp, schd, argv[0], argc-2, argv+2, 
				       TK_CONFIG_ARGV_ONLY);
	}
    } else if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)) {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " cget option\"", (char *)NULL);
	    return TCL_ERROR;
	}
	result = Tk_ConfigureValue(interp, schd->tkwin, schedulerConfigSpecs,
				   (char *) schd, argv[2], 0);
    } else if( (c=='s') && strncmp(argv[1], "step", length)==0)  {
	if(argc>2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " step\"", NULL);
	    return TCL_ERROR;
	}
	// Actually execute a step on the scheduler
	double newtime = schd->nowtime+schd->timestep;
	schd->sched->Update(newtime);
	schd->nowtime = newtime;
	// If there is a TCL command defined, execute that too
	int rslt = TCL_OK;
	if(schd->doCmd && schd->command && strlen(schd->command)>0)  {
	    // first, do % substitutions
	    char output[STRLEN];
	    char *pstrs[2];
	    char timestr[32];
	    sprintf(timestr, "%.3f", schd->nowtime);
	    pstrs[0] = timestr; // %t gets the current time
	    pstrs[1] = argv[0];	// %s gets the name of the command
	    if(ExpandPercents(schd->command, output, STRLEN, 2, "ts", pstrs)){
		rslt = Tcl_VarEval(interp, output, 0);
	    } else {
		Tcl_AppendResult(interp, "error in % substitutions for \"", 
				 schd->command, "\"", NULL);
		rslt = TCL_ERROR;
	    }
	}    
	return rslt;
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1], 
			 "\":  configure/cget/step",
			 (char *) NULL);
	result = TCL_ERROR;
    }
    return result;
}

static int schedulerConfigure(Tcl_Interp *interp, tclScheduler *schd, 
			      char * /* cmdName */, int argc, char **argv, 
			      int flags)
{	/* Handle arguments specifically to configure */
    int rslt = TCL_OK;

    if(Tk_ConfigureWidget(interp, schd->tkwin, schedulerConfigSpecs, 
			  argc, argv, (char *)schd, flags)!= TCL_OK)  {
	return TCL_ERROR;
    }
    if(schd->sched == 0)  {
	schd->sched = new Scheduler;
	if(globalScheduler != NULL)  {
	    fprintf(stderr, "schedulerConfigure: warning: globalScheduler changed\n");
	}
	globalScheduler = schd->sched;
    }
    return rslt;
}

void schedulerDelete(ClientData clientData)
{	/* a particular schedulerObj is being deleted */
    tclScheduler *schd = (tclScheduler *)clientData;
    if(schd->sched)
	delete schd->sched;
    delete schd;
}

