//
// tclUserCmd.cc
//
// provide a hook for easily adding a pattern-recognition function
// to APRL.
//
// eds@media.mit.edu 10 feb 1995

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "double.Vec.h"
#define _UNISTD		// inhibit tcl's unistd.h which conflicts with g++

extern "C" {

#include <tcl.h>
#include <tk.h>

}

#include "UserCmd.h"

Tk_OptionParseProc parse_MS;
Tk_OptionPrintProc print_MS;

static Tk_CustomOption MultiStringtype = // new config type
  { &parse_MS, &print_MS, (ClientData) NULL };

typedef struct {
  Tcl_Interp *interp;
  Tk_Window tkwin;
  char **input; // many inputs
  char *cname; // which usercmd to call
  UserCmd *uc;
  double srate;
  double buftime;
} tclUserCmd;

#define NONESTR "<none>"

static Tk_ConfigSpec usercmdConfigSpecs[] = {
  {TK_CONFIG_CUSTOM,"-inputs","inputs","Inputs",
   NONESTR, Tk_Offset(tclUserCmd,input),0,&MultiStringtype},
  {TK_CONFIG_STRING,"-cname","cname","Cname",
   "1",Tk_Offset(tclUserCmd,cname),0,NULL},
  {TK_CONFIG_DOUBLE,"-srate","srate","Srate",
   "0.0",Tk_Offset(tclUserCmd,srate),0,NULL},
  {TK_CONFIG_DOUBLE,"-buftime","buftime","Buftime",
   "0.0",Tk_Offset(tclUserCmd,srate),0,NULL},
  {TK_CONFIG_END, NULL, NULL, NULL, NULL, 0, 0}
};

extern "C" { int UserCmd_Init(Tcl_Interp *interp); }
static int usercmdCommand(ClientData clientData, Tcl_Interp *interp,
		     int argc, char **argv);
static int usercmdProc(ClientData clientData, Tcl_Interp *interp, 
		  int argc, char **argv);
static int usercmdConfigure(Tcl_Interp *interp, tclUserCmd *uc, 
			      char *cmdName, int argc, char **argv, int flag);
static void usercmdDelete(ClientData clientData);
	   

extern Scheduler *globalScheduler;
extern Tcl_HashTable datasrcTable;


int UserCmd_Init(Tcl_Interp *interp) {
  // register with the interpreter
//  union malloptarg v;  // malloc-debug lines
//  v.i =1;
  
//  mallopt(MALLOC_WARN,v);
//    mallopt(MALLOC_FATAL,v);
//    mallopt(MALLOC_CKCHAIN,v);
//    mallopt(MALLOC_CKDATA,v);
    
  Tcl_CreateCommand(interp,"usercmd",usercmdCommand,(ClientData)NULL,
		   (void (*)(ClientData)) NULL);
  return TCL_OK;
}

static int usercmdCommand(ClientData /* clientData */, Tcl_Interp *interp, 
		   int argc, char **argv) {
  // new usercmd object
  
  if (argc < 2) {
    Tcl_AppendResult(interp, "wrong # args; should be \"", 
			 argv[0], " objName ?options?\"", NULL);
	return TCL_ERROR;
    }
  // allocate a new object
  
  tclUserCmd *ptr = new tclUserCmd;
  ptr->interp = interp;
  ptr->tkwin = Tk_MainWindow(interp);
  ptr->input = NULL;
  ptr->cname = NULL;
  ptr->uc = NULL;

  Tcl_CreateCommand(interp,argv[1],usercmdProc,(ClientData)ptr,
		    usercmdDelete);

  if (usercmdConfigure(interp,ptr,argv[1],argc-2,argv+2,0)
      != TCL_OK) {
    return TCL_ERROR;
  }
  interp->result = argv[1];
  return TCL_OK;
}

static int usercmdProc(ClientData clientData, Tcl_Interp *interp, 
		int argc, char **argv) {
  // only configure
  tclUserCmd *ptr = (tclUserCmd *)clientData;
  int	result = TCL_OK;
  char	c;
  int	length;
    
  if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " <command> ?args?\"", NULL);
	return TCL_ERROR;
    }
    c = argv[1][0];
    length = strlen(argv[1]);
    if( (c=='c') && (strncmp(argv[1], "configure", length)==0) )  {
	if(argc==2) {
	    result =Tk_ConfigureInfo(interp, ptr->tkwin, usercmdConfigSpecs, 
				      (char *)ptr, NULL, 0);
	} else if(argc==3) {
	    result =Tk_ConfigureInfo(interp, ptr->tkwin, usercmdConfigSpecs, 
				      (char *)ptr, argv[2], 0);
	} else {
	    result =usercmdConfigure(interp, ptr, argv[0], argc-2, argv+2, 
				       TK_CONFIG_ARGV_ONLY);
	}
    } else if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)) {
	if(argc != 2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " cget option\"", (char *)NULL);
	    return TCL_ERROR;
	}
	result = Tk_ConfigureValue(interp, ptr->tkwin, usercmdConfigSpecs,
				   (char *) ptr, argv[2], 0);
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1], 
			 "\":  configure/cget",
			 (char *) NULL);
	result = TCL_ERROR;
    }
    return result;
}

static int usercmdConfigure(Tcl_Interp *interp, tclUserCmd *ptr,
			    char *cmdName, int argc, char **argv, int flags) {
  char **oldinputs = ptr->input;

  if (Tk_ConfigureWidget(interp,ptr->tkwin,usercmdConfigSpecs,
			 argc, argv, (char *)ptr,flags) != TCL_OK) {
    return TCL_ERROR;
  }
  if (ptr->input != oldinputs) { // been reconfigured
    int i=0;
    
    DataSource **in = 0;
    int ct = 0;
    for (i=0;ptr->input[i];i++,ct++);
    
    in = (DataSource **)calloc(ct+1,sizeof(DataSource *));
    
    for (i=0;ptr->input[i];i++)
      if (strcmp(ptr->input[i],NONESTR) != 0) {
	Tcl_HashEntry *entryPtr;
	entryPtr = Tcl_FindHashEntry(&datasrcTable,ptr->input[i]);
	if (entryPtr == NULL) {
	  Tcl_AppendResult(interp,"input \"", ptr->input[i],
			   "\" does not match any data sources.",NULL);
	  return TCL_ERROR;
	}
	in[i] = (DataSource *)Tcl_GetHashValue(entryPtr);
      }
    in[i] = NULL;
    if (ptr->uc == NULL) {
      ptr->uc = new UserCmd(ptr->input,in,ct,ptr->cname);
      if (globalScheduler==0) {
	fprintf(stderr,"usercmdConfigure:Error: no scheduler defined\n");
      } else {
	globalScheduler->Enschedule(ptr->uc);
      }
      // Register as a source
      int isnew;
      Tcl_HashEntry *entryPtr;
      entryPtr = Tcl_CreateHashEntry(&datasrcTable,cmdName,&isnew);
      Tcl_SetHashValue(entryPtr,(DataSource *)(ptr->uc));
    } else {
      ptr->uc->SetInputs(in,ct);
    }
  }
  return TCL_OK;
}

static void usercmdDelete(ClientData clientData)
{
  tclUserCmd *ptr = (tclUserCmd *)clientData;
  if (ptr->uc)
    delete ptr->uc;
  delete ptr;
}
  

int parse_MS(ClientData /* cd */, Tcl_Interp * /* interp */,
			    Tk_Window /* tkwin */, char *value, char *widgRec,
			    int offset) {
  //
  // value contains a string which is a comma-delimited list of
  // substrings.  split it and put it in **input, which starts at
  // widgRec + offset
  //
  
  fflush(stdout);

  char **pos;
  char *p = value,*oldp;
  char *s;
  int ct = 0,i;

  
  while (p && ++ct) { // count commas
    p = strchr(p,',');
    if (p) p++  ;
  }
  ct--;

  pos = (char **)calloc(ct+2,sizeof(char *));
  *(char ***)(widgRec + offset) = pos;
  oldp = p = value;
  for (i=0;i!=ct;i++) {
     p = strchr(p,',');
     if (!p)
       p = oldp+strlen(oldp);
     s = (char *)malloc(((int)(p-oldp)+1)*sizeof(char));
     strncpy(s,oldp,p-oldp);
     s[p-oldp] = 0;
     pos[i] = s;
     oldp = p;
     oldp++; // skip comma
    }
  // do the last one
  s = (char *)malloc((strlen(oldp)+1)*sizeof(char));
  strcpy(s,oldp);
  pos[i++] = s;
  pos[i] = NULL;
  
  return TCL_OK;
}

char *print_MS(ClientData cd, Tk_Window /* tkwin */,
			    char * /* widgRec */, int /* offset */,
			    Tcl_FreeProc **freeProcPtr) {
   static char output[500];
   char **p = (char **)cd;
//   char **pos = (char **)widgRec + offset;
   int i=0;

   // as per instructions, build up a string from value in
   // widgrec + offset, and return it.
   
   for (i=0;p[i];i++) {
     strcat(output,p[i]);
     if (p[i++])
       strcat(output,",");
   }
   *freeProcPtr = NULL; // since output is static, don't need to free it

   return (char *)output;
 }
