#!@TCLSH@
#!/usr/local/bin/tclsh
# -*- Tcl -*-
#
# send_afe_msg
#
# Send a message to an audio_frontend process via a socket
#
# 1996nov20 dpwe@icsi.berkeley.edu
# $Header: /u/drspeech/src/audio_frontend/RCS/send_afe_msg.tcl,v 1.5 1999/10/08 00:40:24 dpwe Exp $

# Which port?
set resultport 7777

# What message (empty default -> use stdin)
set message ""

# Echo to stdout also?
set echo 0

# Should we wait for startup?
set wait 0

# Load our common bits from this directory
source "[file dirname [info script]]/afe_utils.tcl"

# Parse command-line args=val options
GetICSIargs

proc SendMessage {message} {
    # Actually send a message
    global resultport wait
    set firsttime 1
    set sock ""
    while {$sock == "" && ($firsttime || $wait)} {
	if {[catch {set sock [socket localhost $resultport]} rslt]} {
	    if {![string match "*connection refused*" $rslt]} {
		puts stderr "Error: $rslt"
		exit -1
	    } else {
		# pause before trying again
		exec sleep 1
	    }
	}
	set firsttime 0
    }
    puts $sock $message
    close $sock
}

proc HandleStdin {} {
    # Handle input on stdin
    global inpipe echo
    # Disable handler for now
    fileevent $inpipe readable ""
    if {[gets $inpipe line] < 0 || [eof $inpipe]} {
	# Error/eof on input
	exit 0
    }
    global filternumbers
    if {$line != ""} {
	SendMessage $line
    }
    if {$echo} {
	puts stdout $line
    }
    fileevent $inpipe readable HandleStdin
}

if {$message == ""} {
    # No message, take stdin
    puts stderr "$argv0: listening to stdin"
    set inpipe stdin
    fileevent $inpipe readable HandleStdin
    vwait dummy
    exit
} else {
    # just send one message
    # .. but map unescaped "&"s in the message to spaces, since they're 
    # hard to do otherwise
    regsub -all {\\&} $message "\n" message
    regsub -all {&} $message { } message
    regsub -all "\n" $message {\&} message
    SendMessage $message
}

exit 0
