// $Header$
//
// $Log$
//
// Online normalization

#include <string.h>
#include <math.h>

#include "OnlfeatureInput.h"
#include "OnlfeatureOutput.h"
#include "Online_Norm.h"

#include "arguments.h"

// general args
// 
size_t ioblocking=8000;

// Online norm params
//
float meantau = 2000.0;
float vartau = 2000.0;
float vareps = 0.001;
char* inhist = NULL;
char* outhist = NULL;

float framerate=100;
int nchan=9;

ARGS ARGS::Args[] = {
  // general
  // 
  ARGS("ioblocking",ARGS::Opt,ioblocking,"The I/O blocking size (samples)[8000]"),
  ARGS("framerate",ARGS::Opt,framerate,"Frame Rate (Hz)[100]"),
  ARGS("nchan",ARGS::Opt,nchan,"Number of channels[9]"),

  // Online norm parameters
  //
  ARGS("meantau",ARGS::Opt,meantau,"mean estimation time constant, in ms.[2000]"),
  ARGS("vartau",ARGS::Opt,vartau,"variance estimation time constant, in ms.[2000]"),
  ARGS("vareps",ARGS::Opt,vareps,"epsilon for variance normalization[0.001]"),
  ARGS("inhist",ARGS::Opt,inhist,"input history file"),
  ARGS("outhist",ARGS::Opt,outhist,"output history file"),
  
  ARGS()
};

int
main(int argc, char** argv)
{
  ARGS::parse(argc,argv);

  OnlfeatureInput ofi( (size_t) ioblocking, nchan, "-");
  OnlfeatureOutput ofo(nchan, "-");

  Online_Norm norm(ioblocking,
                   nchan,
                   framerate,
                   meantau,
                   vartau,
                   vareps,
                   inhist,
                   outhist);

  bool EOFile = false;

  // Loop over all of the input utterances
  //

  while (!EOFile)
    {
      bool EOUtt = false;
      bool EOProc = false;
      while (!EOProc)
        {
          const CMatrix b1 = ofi.read(EOUtt, EOProc, EOFile);
          const CMatrix b2 = norm.processBlock(b1,EOFile);
          ofo.write(b2, EOProc);
        }
    }
}


