#include "dbquery.h"
#include <string.h>
#include "gui.h"
#include <ctype.h>

#include <stdlib.h>	/* for exit() */

inline char *
newstrdup(char *old) {
  char *ret;

  if (old) {
    ret = new char[strlen(old)+1];
    if (ret) strcpy(ret,old);
  } else {
    ret = NULL;
  }
  return ret;
}

// do the same thing as strtok, only allow 0-length strings to return;
char *
strparse(char *string,const char *seps) {
  static char *current;
  char *ret;

  if (string) current=string;
  ret=current;

  current=strpbrk(ret,seps);
  if (current) *(current++)='\0';
  
  return ret;
}

berp_database::berp_database() {
    isopen = 0;
}

void berp_database::open_db_server(void) {
    if (isopen) {
	return;
    }
    isopen = 1;

  int code;
  int child;
  char db_args[BUFSIZ];

  code=pipe(fd);
  if (code == -1) {
    perror("dbquery: pipe");
    exit(1);
  }

#ifdef DEBUG
  fprintf(stderr,"opened pipe 1\n");
#endif
  code=pipe(fd+2);
  if (code == -1) {
    perror("dbquery: pipe");
    exit(1);
  }
#ifdef DEBUG
  fprintf(stderr,"opened pipe 2\n");
#endif

  if ((child=fork()) > 0 ) {
#ifdef DEBUG
    fprintf(stderr,"inside parent\n");
#endif

    close(READ_COMMAND);    
    close(WRITE_RESULT);
#ifdef DEBUG
    fprintf(stderr,"closed descriptors\n");
#endif

    F_READ=fdopen(READ_RESULT,"r");
    F_WRITE=fdopen(WRITE_COMMAND,"w");
#ifdef DEBUG
    fprintf(stderr,"fdopened read/write\n");
#endif
  } else if (child==0) {
#ifdef DEBUG
    fprintf(stderr,"inside child\n");
#endif
    close(WRITE_COMMAND);    
    close(READ_RESULT);
#ifdef DEBUG
    fprintf(stderr,"closed descriptors\n");
#endif
  
    if (dup2(READ_COMMAND,STDIN_FILENO) == -1) {
      perror("dbquery: child: dup2");
    }
#ifdef DEBUG
    fprintf(stderr,"renamed stdin\n");
#endif
    if (dup2(WRITE_RESULT,STDOUT_FILENO) == -1) {
      perror("dbquery: child: dup2");
    }
#ifdef DEBUG
    fprintf(stderr,"renamed stdout\n");
#endif
#ifdef DEBUG
    fprintf(stderr,"execing process\n");
#endif

    // init_db_paths();
    sprintf(db_args,"database=%s",database);
#ifdef DEBUG
    fprintf(stderr, "execl(%s, %s)\n", db_prog, db_args);
#endif
    execl(db_prog,db_prog,db_args,NULL);
    perror("dbquery: execl");
    exit(1);
  } else {
    perror("dbquery: fork");
    exit(1);
  }
}

berp_database::~berp_database() {
  fclose(F_READ);
  fclose(F_WRITE);
}

berp_query_result *
berp_database::query(char *q){
  berp_query_result *result=new berp_query_result;
  int num_matches;
  int i;

  open_db_server();

  fputs(q,F_WRITE);
  fflush(F_WRITE);
  fgets(buf, 5000, F_READ);

  if (sscanf(buf,"Matches: %d",&num_matches) == 1 && num_matches>0) {
    result->num_matches=num_matches;
    result->restaurants=new berp_restaurant*[num_matches];
    for (i=0;i<num_matches;i++) {
      fgets(buf, 5000, F_READ);
      result->restaurants[i]=new berp_restaurant(buf);
    }
  } else {
    result->num_matches=0;
    result->restaurants=NULL;
  }
  return result;
}

berp_query_result::~berp_query_result() {
  if (restaurants) delete[] restaurants;
}

berp_restaurant::berp_restaurant(char *string) {
  char *p;

  p=strparse(string,"|\n");
  rest_type=newstrdup(p=strparse(NULL,"|\n"));
  cost=newstrdup(p=strparse(NULL,"|\n"));
  sunday=newstrdup(p=strparse(NULL,"|\n"));
  monday=newstrdup(p=strparse(NULL,"|\n"));
  tuesday=newstrdup(p=strparse(NULL,"|\n"));
  wednesday=newstrdup(p=strparse(NULL,"|\n"));
  thursday=newstrdup(p=strparse(NULL,"|\n"));
  friday=newstrdup(p=strparse(NULL,"|\n"));
  saturday=newstrdup(p=strparse(NULL,"|\n"));
  minutes=newstrdup(p=strparse(NULL,"|\n"));
  name=newstrdup(p=strparse(NULL,"|\n"));
  street_num=newstrdup(p=strparse(NULL,"|\n"));
  street_name=newstrdup(p=strparse(NULL,"|\n"));
  comments=newstrdup(p=strparse(NULL,"|\n"));
  other_types=newstrdup(p=strparse(NULL,"|\n"));
  phonenumber=newstrdup(p=strparse(NULL,"|\n"));
  cross_street=newstrdup(p=strparse(NULL,"|\n"));
}

static char OutputBuf[9000];
static berp_database *berpdb;

extern "C" void init_dbquery(void) {
    // backendInit can call this after setting up the environment
    // variables, and we will open the "static" berp database server.
    berpdb = new berp_database;
    berpdb->open_db_server();
#ifdef DEBUG
    fprintf(stderr, "init_dbquery complete\n");
#endif /* DEBUG */
}
		      
char *
query(char *QueryBuf) {
  int i;
  berp_query_result *res;
  char buf[BUFSIZ];
  char tempbuf[BUFSIZ];

  buf[0] = 0;

  res=berpdb->query(QueryBuf);
  guiproc("delrl");

  if (res->num_matches) {
    sprintf(OutputBuf,"\nHere are the restaurants matching your query:\n\n");
    strcat(OutputBuf,"@bold(Name)                      | @bold(Address)                |  @bold(Cost) | @bold(Min. from ICSI)\n");
    strcat(OutputBuf,"---------------------------------------------------------------------------\n");    
    for (i=0;i<res->num_matches;i++) {
      sprintf(tempbuf,"%s %s",
	      res->restaurants[i]->street_num,
	      res->restaurants[i]->street_name);
      sprintf(buf,"@bold(%25s) | %22s | %5s | %s\n",
	      res->restaurants[i]->name,
	      tempbuf,
	      res->restaurants[i]->cost,
	      res->restaurants[i]->minutes);
      strcat(OutputBuf, buf);
      guiproc_strs("addrl",
		   "\"", res->restaurants[i]->name, "\"",
		   "\"", tempbuf,"\"",
		   "\"", res->restaurants[i]->cost, "\"",
		   "\"", res->restaurants[i]->minutes, "\"",
		   0);
    }
      strcat(OutputBuf,"\n@italic(You may ask for additional information on any of the above restaurants).\n");
      guiproc_strs("app", "You may ask for additional information on any of the above restaurants.", 0);
  } else {
      sprintf(OutputBuf,"Sorry, no restaurants match your query.\n");
  }
  delete res;
  return OutputBuf;
}

extern "C" int myfree(char *);

char * 
RestInfo(char *RestName)
{
  int i;
  char buf[BUFSIZ];
  berp_query_result *res;
  char costbuf[BUFSIZ];
  char tmpbuf[BUFSIZ];
  char *tatebuf;

  for(i = 0; i < strlen(RestName); i++)  {
    RestName[i] = toupper(RestName[i]);
    if (RestName[i] == '-')
      RestName[i] = ' ';
  }

  sprintf(buf,"REST.name = \"%s\"\n",RestName);

  res=berpdb->query(buf);
 
  if (res->num_matches) {
    // this assumes that only one restaurant is returned
    sprintf(buf,"Name           : @bold(%s)\n",res->restaurants[0]->name);
    strcat(OutputBuf,buf);
    sprintf(buf,"Type of Food   : @bold(%s)\n",res->restaurants[0]->rest_type);
    strcat(OutputBuf,buf);
    sprintf(buf,"Other types    : @bold(%s)\n",
	    res->restaurants[0]->other_types);
    strcat(OutputBuf,buf);
    sprintf(buf,"Address        : @bold(%s %s)\n",
	    res->restaurants[0]->street_num,
	    res->restaurants[0]->street_name);
    strcat(OutputBuf,buf);
    sprintf(buf,"Cross Street   : @bold(%s)\n",
	    res->restaurants[0]->cross_street);
    strcat(OutputBuf,buf);
    sprintf(buf,"Phone number   : @bold(%s)\n",
	    res->restaurants[0]->phonenumber);
    strcat(OutputBuf,buf);

    sprintf(costbuf,"%s",res->restaurants[0]->cost);
    if(!strcmp(costbuf,"$")) 
      sprintf(buf,"Cost           : @bold(less than $6.00 per person)\n");
    else if(!strcmp(costbuf,"$$"))
      sprintf(buf,"Cost           : @bold($6.00-$12.00 per person)\n");
    else if(!strcmp(costbuf,"$$$"))
      sprintf(buf,"Cost           : @bold($12.00-$16.00 per person)\n");
    else 
      sprintf(buf,"Cost           : @bold(more than $16.00 per person)\n");

    strcat(OutputBuf,buf);
    tatebuf = strdup(buf);
    if(!strcmp(costbuf,"$")) 
      sprintf(tatebuf,"less than $6.00 per person");
    else if(!strcmp(costbuf,"$$"))
      sprintf(tatebuf,"$6.00-$12.00 per person");
    else if(!strcmp(costbuf,"$$$"))
      sprintf(tatebuf,"$12.00-$16.00 per person");
    else 
      sprintf(tatebuf,"more than $16.00 per person");

    sprintf(buf,"Min. walk from ICSI: @bold(%s)\n",res->restaurants[0]->minutes);
    strcat(OutputBuf,buf);
    sprintf(buf,"Hours:(B=breakfast,L=lunch,D=dinner,N=night)\n");
    strcat(OutputBuf,buf);
    sprintf(buf,"Mon     Tue     Wed     Thu     Fri     Sat     Sun\n");
    strcat(OutputBuf,buf);
    sprintf(buf,"---     ---     ---     ---     ---     ---     ---\n");
    strcat(OutputBuf,buf);
    sprintf(buf,"@bold(@bold(%3s))     @bold(%3s)     @bold(%3s)     @bold(%3s)     @bold(%3s)     @bold(%3s)     @bold(%3s)\n",
	    res->restaurants[0]->monday,
	    res->restaurants[0]->tuesday,
	    res->restaurants[0]->wednesday,
	    res->restaurants[0]->thursday,
	    res->restaurants[0]->friday,
	    res->restaurants[0]->saturday,
	    res->restaurants[0]->sunday);
    strcat(OutputBuf,buf);
    for (i=0;i<BUFSIZ;i++)
	buf[i]=0;
    strcpy(tmpbuf,res->restaurants[0]->comments);
    guiproc_strs("restinfo",
		 "\"", res->restaurants[0]->name, "\"",
		 "\"", res->restaurants[0]->rest_type, "\"",
		 "\"", res->restaurants[0]->other_types, "\"",
		 "\"", res->restaurants[0]->street_num,
		       res->restaurants[0]->street_name, "\"",
		 "\"", res->restaurants[0]->cross_street, "\"",
		 "\"", res->restaurants[0]->phonenumber, "\"",
		 "\"", tatebuf, "\"",
		 "\"", res->restaurants[0]->minutes, "\"",
		 "\"", tmpbuf, "\"",
		 "\"", res->restaurants[0]->monday, "\"",
		 "\"", res->restaurants[0]->tuesday, "\"",
		 "\"", res->restaurants[0]->wednesday, "\"",
		 "\"", res->restaurants[0]->thursday, "\"",
		 "\"", res->restaurants[0]->friday, "\"",
		 "\"", res->restaurants[0]->saturday, "\"",
		 "\"", res->restaurants[0]->sunday, "\"",
		 0);
    myfree(tatebuf);

    if (strlen(tmpbuf) < 62)
	    sprintf(buf,"\n\nComments       : @bold(%s)\n",tmpbuf);
    else {
      sprintf(buf,"\n\nComments       : ");
      strcat(buf,"@bold(");
      strncat(buf,tmpbuf,62);
      strcat(buf,")");
      strcat(buf,"\n");
      for (i=62;i<strlen(tmpbuf);i+=79) {
	strcat(buf,"@bold(");
	strncat(buf,tmpbuf+i,79);
	strcat(buf,")");
	strcat(buf,"\n");
      }
      //strcat(buf,"\n");
    }
    strcat(OutputBuf,buf);
  } else { // num_matches == 0
    sprintf(OutputBuf,"Sorry, no restaurants match your query.\n");
  }

  delete res;
    
  return(OutputBuf);
}

