extern "C" {
#include<stdio.h>
#include<errno.h>
#include<unistd.h>
}

/* File descriptors:
   fd[0]: CHILD database program READS commands
   fd[1]: PARENT program WRITES commands
   fd[2]: PARENT program READS results
   fd[3]: CHILD database program WRITES results
   */

#define READ_COMMAND fd[0]
#define WRITE_COMMAND fd[1]
#define READ_RESULT fd[2]
#define WRITE_RESULT fd[3]

/* The environment variable name for $SPEECH_DIR */
#define SPEECHDIR_ENV "SPEECH_DIR"
/* What to use for $SPEECH_DIR if it isnt defined in the environment */
#define DEF_SPEECHDIR "/u/drspeech"
/* prototype for a function that will return a pointer to a string 
   containing the accepted value for $SPEECH_DIR */
extern const char *speech_dir(void);
/* function to copy *in to *out, prepending $SPEECH_DIR if *in 
   doesn't start with "/" */
extern void maybe_prepend_speechdir(char *to, const char *from);
/* function to read environment variable or substitute default */
extern "C" void getenv_default(const char *envname, const char *deflt, char *out);

/* Where to find the database prog and the database itself.
   If they don't start with /, $SPEECH_DIR is prepended. */
#define DB_PROG  "share/bin/simpdb"
#define DATABASE "demo/berp/data/restaurants"
/* Actual current values for these parameters */
extern char *db_prog;
extern char *database;
/* static initializer for these paths */
extern "C" void init_db_paths(void);

/* intialize the berp database (&server) */
extern "C" void init_dbquery(void);

class berp_restaurant {
public:
  char *rest_type;
  char *cost;
  char *sunday;
  char *monday;
  char *tuesday;
  char *wednesday;
  char *thursday;
  char *friday;
  char *saturday;
  char *minutes;
  char *name;
  char *street_num;
  char *street_name;
  char *comments;
  char *other_types;
  char *phonenumber;
  char *cross_street;
  berp_restaurant(char *);
  ~berp_restaurant();
};

class berp_query_result {
public:
  int num_matches;
  berp_restaurant **restaurants;
  ~berp_query_result();
};

class berp_database {
  FILE *F_READ;
  FILE *F_WRITE;
  int fd[4];
  char buf[5000];
  int isopen;

public:
  void open_db_server(void);  
  berp_query_result *query(char *);
  berp_database();
  ~berp_database();
};

