/*******************************************************************************

	gramcheck.C:

		The main routine for the gramcheck tool.

		Author:		Dan Jurafsky
		Date:		Jan 1994
		History:	DSJ	6-5-93	Added header


******************************************************************************/

#include "grammar.h"
#include "parser.h"

/********************************* GLOBAL VARIABLES ************************/

static Grammar gr("ss",0);
static ChartParser p(&gr);
static FILE *fp;

/*  These are for finding memory leaks */
int News = 0;
int Deletes = 0;
int NodeDeletes = 0;
int NodeCreates = 0;
int ArcNews = 0;
int ArcDeletes = 0;
int RuleNews = 0;
int RuleDeletes = 0;
int SymbolNews = 0;
int SymbolDeletes = 0;

/********************************* Main ************************/

main(int argc, char *argv[])
{
	double pr;
	char buf[200];
	char *s;
	int flags = 0;

	float pfollowsize,unpfollowsize;
	int i = 0;
	int len = 0;
	double entropy = 0;
	FILE * errorfp = fopen("parse.err","w+");

	// Read in the grammar

	gr.ReadFile("berp.grammar");
	p.ComputeLCArray();

	while (fgets(buf,BUFSIZ,stdin) != NULL) {
		s = p.TopDownParser(buf,&pr,0,0,0,&pfollowsize,&unpfollowsize);
		//s = p.TopDownParser(buf,&pr,0,PRINTCHART,0,&pfollowsize,&unpfollowsize);
		//cout << "pr of sentence " << buf << " is " << pr << "\n" << flush;
		if (pr != 0.0) 
			printf("%s\n",buf);
		else
			fprintf(errorfp,"%s\n",buf);
		p.Reset();
	}

	gr.Reset();

	exit(0);
}
	


void *operator new(size_t sz)
{
	++News;
	return malloc(sz);
}

void operator delete(void *p)
{
	++Deletes;
	free((char *)p);
}
