#ifdef __GNUC__
#include <pfstream.h>
#else /* !__GNUC__*/
#include <iostream.h>
#endif /* __GNUC__ */

#define BOTTOMUP 1
#define TOPDOWN 2

const int MAXRHSRULES = 20;
const int FIRSTRHS = 1;
const int MAXARCS = 4000;
const int MAXAGENDA = 4000;
const int PRINTCHART = 2;
const int PRINTPROB = 1;
const int TOPFOLLOW = 1;
const int BOTTOMFOLLOW = 2;
const int MAXGUESS = 70;

const int MAX_SENTENCE_LEN = 200;

const int STATE_NONE = 0;
const int STATE_SCAN = 1;
const int STATE_COMPLETION = 2;
const int STATE_PREDICTION = 3;

class Chart;
class ArcIterator;
class ParseTreeIterator;

class Arc : public Rule {

        int	From;
	int	To;
	int	IsComplete;
	Symbol	*NextSym;
	int	dot;
	Tree *	ArcTree;
	double	prefixprob;
	double	insideprob;
	int		state;
	

public:
	        Arc();
	        Arc(Rule*);
	        ~Arc();
	        Arc(Symbol*,int,int);
	Arc &	operator=(Arc&);
	void	AdvanceDot();
	void	print(ostream&os=cout);
	void    print(FILE *);
	void	Reset();
	void	BuildSem(int, char*);
	int	Disjoint(int *);

	// Added by Dan Gildea
	char *  GetSemFromDag();

	friend	ArcIterator;
	friend	ParseTreeIterator;
	friend	Chart;
	friend	ChartParser;
};

class KeyStack {
	Arc*	TheStack[200];
	int		index;
public:
			KeyStack();
	void		Push(Arc*);
	Arc*	Pop();
	Arc*	Peek();
};

class AgendaList {
	Arc*	q[MAXAGENDA+1];
	int	head;
	int	tail;
public:
		AgendaList();
	void	Add(Arc*);
	Arc*	Get();
};


class ArcIterator;
class ArcCollection {
	int	iArcs;
	Arc	*Arcs[MAXARCS];
public:
		ArcCollection();
	void	AddArc(Arc*);
	void	Reset();
	void 	print();
	friend	ArcIterator;
	friend	Chart;
};

class ArcIterator {
	int	current;
	ArcCollection *c;
public:
                 ArcIterator();
                 ArcIterator(ArcCollection *);
	Arc *	 operator()(Symbol*,int);
	Arc *	 operator()(Symbol*);
	Arc *	 operator()();
	Arc *	 operator()(Symbol*, int, int);
	Arc * 	NextStart(Symbol* Cat, int Start);
	void	 Reset();
};
#ifdef NOTDEF
class ParseTree : public Tree {
	int	junk;
	ParseTree(ParseNode n);
	friend	Chart;
};


class ParseNode : public Node {
	void 	ExpandNode(Arc *);
	friend	Chart;
	ParseNode(Symbol*,float);
};

#endif

class ChartParser;

class Chart {
        ArcCollection		Arcs;
	ArcCollection		Constits;
	KeyStack		Keys;
	AgendaList		Agenda;
	Symbol			*Start;
	int			Len;
	Tcl_HashTable		HashTable;
	char 			ArcIndex[MAXARCS][MAXSYMLEN+4];
	int			arccount;
	int                     Parsetype;
public:	
	        Chart();
	void	Reset();
	void	ExtendArc(Arc*,int,Tree*,char*);
	void	AddConstit(Symbol*,int,int);
	void	AddConstit(Arc *);
	void	NewArc(Rule*,int,int,Tree*,char*);
	Arc *	PopKey();
	Tree	*GetParseTree(Symbol*,int);
	char 	*GetSem(Symbol*, int, double *);
	Arc 	*GetBestSem(int*);
	char    *GetCoverSem(int *);
//	Tree	*MakeParseTree(Symbol*);
//	int		RecurseMakeParseTree(Node*,int);
	int	GetLen() {return Len;}
	void	print();
	friend	ChartParser;
	friend	ParseTreeIterator;
	int 	GetNumberOfParses(Symbol*, int);
	double 	GetSentenceProb();
	double  GetBottomUpSentenceProb();
	Arc * 	GetArc(int,int,int);
	void	TopDownExtendArcDown(Arc* Top, int To, Tree * t, char *s );
	void	TopDownExtendArcUp(Arc* a, int To, Tree * t, char *s,double,int);
	void	TopDownNewArc(Rule *r, int From, int To, Tree * t, char *s,int iscomplete,double,int);
	void	Register(Arc *ap);
	void	UpdateArcProb(Rule *r, int From, int To, double prob);
#ifdef TATEYAMA
	void sendAllParses2ptw(char *);
#endif
};


class ChartParser {
	Chart	c;
	Grammar	*grammar;
	double	*RuleCountArray;
	int	*RuleCountWork;
	Symbol  *GetWord(char*&);
	float 	termarray[MAXSYMBOLS];
public:
	Symbol	*tokens[70];
	Symbol	*last_token;
	int	ntokens;
	Symbol	*UnkWord;
	Symbol	*EndWord;
	Symbol	*Pause;
	Bigram	*bg;
		ChartParser(Grammar *);
	int	GetLen() {return c.GetLen();}
	void	Reset();
	void	CountRulesInit();
	void 	CountRules();
	char 	*Parser(char *words, double*prob, int countflag, int printlevel, int *guessflag,int *guesslist);
	void 	PrintRuleCount();
	void 	NormalizeRules();
	int ComputeFollowSet(int);
	char *TopDownParser(char *words,double*pr, int countflag, int printlevel);
	void TopDownInit();
	Symbol * Tokenizer(char*& );
	void Propose(Symbol*,int,double );
	void 	CopyOutTermArray(float * out);
	void 	ComputeLCArray();
};


class ParseTreeIterator {
	ArcIterator next;
	Symbol	*start;
	int	len;
public:
		ParseTreeIterator(Chart *c);
	Tree * Next();
	void  Reset();
};

