#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#ifdef OLD_STUFF
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#ifdef __SVR4
#include <sun/audioio.h>
#else
#include <multimedia/libaudio.h>
#include <multimedia/audio_device.h>
#include <sun/audioio.h>
#endif

/* Local variables */
char buf[BUFSIZ];

/*
 * This defines the tolerable sample rate error as a ratio between the
 * sample rates of the audio data and the audio device.
 */
#define	SAMPLE_RATE_THRESHOLD	(.01)


char		*Audio_dev = "/dev/audio";

int		Audio_fd = -1;		/* file descriptor for audio device */
Audio_hdr	Dev_hdr;		/* audio header for device */
Audio_hdr	File_hdr;		/* audio header for file */


/*
 * Play a file
 */
int
play_sparc(char* fname)
{
  int		cnt;
  int		err;
  int		ifd;
  struct stat	st;

  /* Validate and open the audio device */
  err = stat(Audio_dev, &st);
  if (err < 0) {
    fprintf(stderr, "cannot stat ");
    perror(Audio_dev);
    return(FALSE);
  }

  /* Try to open it quickly */
  Audio_fd = open(Audio_dev, O_WRONLY | O_NDELAY);
  if ((Audio_fd < 0) && (errno == EBUSY)) {
    fprintf(stderr, "%s is busy\n", Audio_dev);
    return(FALSE);
  }
  if (Audio_fd < 0) {
    fprintf(stderr, "error opening ");
    perror(Audio_dev);
    return(FALSE);
  }

  /* Get the device output encoding configuration */
  if (audio_get_play_config(Audio_fd, &Dev_hdr) != AUDIO_SUCCESS) {
    fprintf(stderr, "%s is not an audio device\n",  Audio_dev);
    return(FALSE);
  }

  if ((ifd = open(fname, O_RDONLY, 0)) < 0) {
    fprintf(stderr, "cannot open ");
    perror(fname);
    return(FALSE);
  }

  /* Check to make sure this is an audio file */
  err = audio_read_filehdr(ifd, &File_hdr, (char *)NULL, 0);
  if (err != AUDIO_SUCCESS) {
    fprintf(stderr, "%s is not a valid audio file\n", fname);
    close(ifd);
    return(FALSE);
  }

  /* Check the device configuration */
  if (audio_cmp_hdr(&Dev_hdr, &File_hdr) != 0) {
    /*
     * The device does not match the input file.
     * Wait for any old output to drain, then attempt
     * to reconfigure the audio device to match the
     * input data.
     */
    if (audio_drain(Audio_fd, FALSE) != AUDIO_SUCCESS) {
      perror("AUDIO_DRAIN error");
      return(FALSE);
    }
    if (!reconfig(fname)){
      close(ifd);
      return(FALSE);
    }
  }

  /*
   * At this point, we're all ready to copy the data.
   */
  while ((cnt = read(ifd,buf, sizeof (buf))) >= 0) {
    /* If input EOF, write an eof marker */
    err = write(Audio_fd,buf, cnt);

    if (err != cnt) {
      fprintf(stderr, "output error: ");
      perror("");
      break;
    }
    if (cnt == 0)
      break;
  }
  if (cnt < 0) {
    fprintf(stderr, "error reading ");
    perror(fname);
  }

  close(ifd);		/* close input file */

  /*
   * Though drain is implicit on close(), it's performed here
   * for the sake of completeness
   */
  audio_drain(Audio_fd, FALSE);
	
  close(Audio_fd);	/* close output */
  return(TRUE);
}


/*
 * Try to reconfigure the audio device to match the file encoding.
 * If this fails, we should attempt to make the input data match the
 * device encoding.  For now, we give up on this file.
 *
 * Returns TRUE if successful.  Returns FALSE if not.
 */
int
reconfig(char* file)
{
  int	err;
  char	msg[AUDIO_MAX_ENCODE_INFO];

  Dev_hdr = File_hdr;
  err = audio_set_play_config(Audio_fd, &Dev_hdr);

  switch (err) {
  case AUDIO_SUCCESS:
    return (TRUE);

  case AUDIO_ERR_NOEFFECT:
    /*
     * Couldn't change the device.
     * Check to see if we're nearly compatible.
     * audio_cmp_hdr() returns >0 if only sample rate difference.
     */
    if (audio_cmp_hdr(&Dev_hdr, &File_hdr) > 0) {
      double	ratio;

      ratio = (double) abs((int)
			   (Dev_hdr.sample_rate - File_hdr.sample_rate)) /
			     (double) File_hdr.sample_rate;
      if (ratio <= SAMPLE_RATE_THRESHOLD) {
	return (TRUE);
      }
      fprintf(stderr, "%s sample rate %d not available\n",file, File_hdr.sample_rate);
      return (FALSE);
    }
    (void) audio_enc_to_str(&File_hdr, msg);
    fprintf(stderr, "%s encoding not available: %s\n",file, msg);
    return (FALSE);

  default:
    fprintf(stderr, "i/o error (set config)\n");
    return(FALSE);
  }
} 

#else /* OLD_STUFF */

/* #ifdef __SVR4
   #define PLAY_CMD "/bin/audioplay"
   #else
   #define PLAY_CMD "/bin/play"
   #endif
*/
/* #define PLAY_CMD "/u/dpwe/bin/sun4/play.aif" */
#define PLAY_CMD_DFLT "sndplay"
/* which env var to query for the play prog binary */
#define PLAY_CMD_ENV "PLAYPROG"

/*
 * Play a file
 */
#define MAX_CMD_LEN 1024

int play_sparc(char *path, char *fname) {
  char buf[MAX_CMD_LEN];
static char playCmd[MAX_CMD_LEN];
static int havePlayCmd = 0;

  if(!havePlayCmd) {
      char *s;
      getenv_default(PLAY_CMD_ENV, PLAY_CMD_DFLT, playCmd);
      havePlayCmd = 1;
  }

  sprintf(buf,"%s %s/%s", playCmd, path, fname);
  system(buf);

/* free(buf); */
}

#endif
