#include "unify_hash.h"

int HashTable::hash(char *s) {
	const char *p;	// assumes 8-bit char
	unsigned int h;	// assumes 32-bit int
	h = 0;
	for (p = s; *p != '\0'; p++) {
		h = (h << 4) + *p;
		h = (h ^ ((h & 0xf0000000) >> 24)) & 0x0fffffff;
	}
	return h % sizeOfTable;
}

int prime(int n) {
        int i;
	for (i = 2; i * i < n; i++)
		if (n % i == 0)
			return 0;
	return 1;
}

int nextPrime(int n) {
        int i;
	for (i = n + 1; !prime(i); i++)
		;
	return i;
}

void HashTable::rehash() {
	int oldSize = sizeOfTable;
	Link **oldBuckets = buckets;
	sizeOfTable = nextPrime(sizeOfTable * 2);
	buckets = new Link *[sizeOfTable];
        int i;
	for (i= 0; i < sizeOfTable; i++)
		buckets[i] = NULL;
	for (i = 0; i < oldSize; i++) {
		Link *cur, *next;
		for (cur = oldBuckets[i]; cur != NULL; cur = next) {
			int newIndex = hash(cur->s);
			next = cur->next;
			cur->next = buckets[newIndex];
			buckets[newIndex] = cur;
		}
	}
	delete [] oldBuckets;
}


char *HashTable::entry(char *str) {
	if (count >= sizeOfTable)
		rehash();
	int i = hash(str);
	Link **l = buckets + i;
	while (*l)
		if (!strcmp(str, (*l)->s))
			return (*l)->s;
		else
			l = &((*l)->next);
	*l = new Link(str, *l);
	count++;	/* one more string in table */
	return (*l)->s;
}

HashTable *stringTable;
char *StringTableEntry(char *str) {
  if (!stringTable) stringTable=new HashTable; /* default size 223 buckets */
  return stringTable->entry(str);
}
