#!/usr/local/bin/wish
#
# berpdemo98.tcl
#
# New version of the BeRP demo has audio_frontend, probs & labels display,
# and backend all (or almost all) embedded within a single application.
# (Originally called integ_frontend.tcl)
#
# 1997jun19 dpwe@icsi.berkeley.edu
# $Header: /u/drspeech/src/berpdemo/share/RCS/berpdemo98.tcl,v 1.11 1999/10/08 01:39:32 dpwe Exp $

# Variables defined in configuration, passed to us through the environment

# Debug error message
set _debug 0
proc Debug {msg} {
    global _debug
    if {$_debug} {
	puts stderr "Debug: $msg"
    }
}

if {[array names env DPWE_DEBUG] != ""} {
    set _debug 1
    rename load load_orig
    proc load {args} {
	puts "** load $args"
	uplevel 1 "load_orig $args"
    }
    rename source source_orig
    proc source {args} {
	puts "** source $args"
	uplevel 1 "source_orig $args"
    }
}

# Where the berp data (restaurant database etc) is
set BERP_DATA_DIR     $env(BERP_DATA_DIR)
# Where the simpleui data (net, decoder grammar etc) is
set SIMPLEUI_DATA_DIR $env(SIMPLEUI_DATA_DIR)

# The CTM files are written in this directory; we want them 
# visible across the network (when running decode on a remote machine), 
# so this is now part of the install tree!  "make install" creates this 
# directory.
set TMPDIR "$BERP_DATA_DIR/tmp"

set FEACALC "rasta -k -S \$samplerate -m 12 -n 13 -L -F -q 2"
#set FEACALC "feacalc -rasta log -plp 12 -hpf -samplerate $samplerate -domain cep -clsave no -output - -opformat online -"
set QNSFWD  "qnsfwd"
set NOWAY   "noway"
# .. but watch out for when I may prepend them with "rsh $machine..." below

# Basic utility package (including getenv & SpeechArch)
package require Dpwe_Utilfns
# Load this now so we can grab the timeline widget below
package require APRL
# Stuff for spectrogram
package require StftSound
# .. and the stuff for displaying them
package require Dpweutils_tk
# Canvas display and label file access
package require Dpwe_Labelfilefns
# load definitions of breadf and breadi
package require Dpweutils_tcl
# Setup the loading of audio_frontend.tcl on audio_frontend_init
package require AudioFrontend

# Pick up any command line params passed in
set decodemachine ""
set fwpmachine ""

GetICSIargs "decodemachine fwpmachine"

if {$fwpmachine != ""} {
    set QNSFWD  "rsh $fwpmachine $QNSFWD"
}
if {$decodemachine != ""} {
    set NOWAY   "rsh $decodemachine $NOWAY"
}

# Assume I can find my scripts alongside myself
set DEMOSHAREDIR [file dirname [info script]]

# Pick up the readproc management routines (alongside probs.tcl)
#source "$DEMOSHAREDIR/rdprocmgr.tcl"
package require Dpwe_Rdprocmgr

# Global reference for secs per feature frame (dep. on front end)
set stepTime 0.010	;# 10 ms window step
# Skew between beginning of soundfile and beginning of features 
# (to align display)
set padtime 0.0

# Overall widget root (define ife_root before sourcing to embed)
if {[info var ife_root] == ""} {
    set  ife_root ""
}

# Base frames
set  fe_frame [frame $ife_root.fe]
set  afe_cont [frame $fe_frame.cont]
set  afe_root [frame $afe_cont.afe -relief groove -borderwidth 2]
set  pr_frame [frame $fe_frame.pr]

# sgram stuff
set  sg_root  [frame $pr_frame.sgm -relief groove -borderwidth 2]
set  sg_cnv   [canvas $sg_root.cnv -height 75]
# (add a timescale - APRL provided widget)
set  sg_tlb   [timeline $sg_root.tlb -orient horiz -minVal 0 -tixTopLeft 1 \
	                -maxVal 2.5 -relief flat -bg \#c8c8c8 -height 20 \
			-padX 0 -padY 0]
pack $sg_tlb  -side bottom -fill x -expand 0
pack $sg_cnv  -side top    -fill both -expand 1
pack $sg_root -side top -expand 0 -fill x -padx 3 -pady 3

# Features display canvas
set  fea_root  [frame $pr_frame.fea -relief groove -borderwidth 2]
# Feadisp height is 1x39 + a bit
set nfeatures  39	;# i.e. -n 13 x (-q 2 +1)
set fea_htfact 1
set  fea_cnv   [canvas $fea_root.cnv -height [expr 11+$fea_htfact*$nfeatures]]
pack $fea_cnv  -side top    -fill both -expand 1
pack $fea_root -side top -expand 0 -fill x -padx 3 -pady 3

# probs stuff
set  prb_root [frame $pr_frame.prb -relief groove -borderwidth 2]
pack $prb_root -side top -expand 0 -fill x -padx 3 -pady 3

# phonetic label stuff
set  lab_root [frame $pr_frame.lab -relief groove -borderwidth 2]
set  lab_cnv [canvas $lab_root.cnv -height 75]
pack $lab_cnv -fill both -expand 1
pack $lab_root -side top -expand 0 -fill x -padx 3 -pady 3

# Pack it all in
pack $afe_root -side top -padx 5 -pady 5
pack $afe_cont -side left -fill y -padx 3 -pady 3
pack $pr_frame -side left -expand 1 -fill both -padx 5 -pady 5
set  be_root [frame $ife_root.be -relief groove -borderwidth 2]
pack $fe_frame $be_root -side top -expand 1 -fill both

# Stuff for spectrogram
# (make sure packages are loaded, so that the Sound command exists, and FAImgs)
eval $auto_index(stftsound)
eval $auto_index(traceplot)

# Routine to update occaisionally
proc myUpdate {} {
    # Call updates (for the screen only)
    rdprocDisable
    Debug "**abt to update"
    update
    rdprocEnable
}

# Select the cute colormap
floatArrayCmap "esps"
# Set up the feadisp structures
set feaFa [floatArray \#auto]
set feaIm [image create photo feaImage -format FloatArray]
# -palette 256
set feaImLarger [image create photo feaImageLarger]
#  -palette 256
set feaTag "feaim"

proc feadispReset {soundDuration} {
    # Prepare to receive a new set of feature vectors
    global feaCol feaFa nfeatures stepTime feaIm feaImLarger fea_cnv feaTag
    global prb_timebasescale fea_htfact
    set feaCol 0
    $feaFa Resize $nfeatures [expr int($soundDuration/$stepTime)]
    $feaFa Fill 0.5
    # Reset the images too
    $feaIm config -width 0
    $feaImLarger config -width 0
puts stderr "larger im width: [$feaImLarger cget -width] fa size=[$feaFa Size]"
    $fea_cnv delete $feaTag
    # Put the image up straight away
    set rowfact $fea_htfact
    set colfact $prb_timebasescale
    lassign [$feaFa Size] imrows imcols
    # Display the array
    $feaIm config  -data $feaFa
    $feaIm config -width $imcols
    # Zoom between images
    $feaImLarger copy $feaIm -zoom $colfact $rowfact
    $feaImLarger config -width [expr $colfact*$imcols]
    # Place on the canvas
    $fea_cnv create image 0 5 -anchor nw -image $feaImLarger -tags $feaTag
}    

proc feadispDraw {} {
    # Draw up some of the features
    global fea_cnv feaIm feaImLarger feaFa prb_timebasescale feaTag fea_htfact
    # magnification of each sample pt
    set rowfact $fea_htfact
    set colfact $prb_timebasescale
    # Rescale the array
#    lassign [$feaFa Size] imrows imcols
#    $feaFa Resize [expr $imrows*$rowfact] [expr $imcols*$colfact]
    # Display the array
    $feaIm config  -data $feaFa
#    $feaIm config -width $imcols
    # Zoom between images
    $feaImLarger copy $feaIm -zoom $colfact $rowfact
#    $feaImLarger config -width [expr $colfact*$imcols]
#    $fea_cnv delete $feaTag
#    $fea_cnv create image 0 5 -anchor nw -image $feaImLarger -tags $feaTag
    myUpdate
}

# How often to update the image
set feaCol 0
set feaUpdate 5

proc feadispAddCol {feavec} {
    # Add the next column to features from a floatVec
    global feaFa feaCol feaUpdate
    $feaFa SetCol $feaCol $feavec
    incr feaCol
#    puts stderr "feadispAddCol: col $feaCol of [$feaFa Cols] stored"
    # Update the display every so often
    if {($feaCol % $feaUpdate) == 0} {
	feadispDraw
    }
}

# Set up the sgram structures
floatArrayCmap "esps"
set sgSo [Sound \#auto]
set sgFa [floatArray \#auto]
set sgIm [image create photo sgramImage -format FloatArray]
set sgIm2 [image create photo sgramImage2]

proc sgramInit {soundfile} {
    # Initialize drawing the spectrogram
    global afe_soundfile sg_cnv sg_tlb sgSo sgFa sgIm sgIm2 
    global soundDuration stepTime sg_vfact
    # from probs.tcl - pixels per 10ms frame
    global prb_timebasescale padtime
    # Set up elsewhere
    global samplerate

    # Even though this is just the beginning, we can, in fact, 
    # load the sound now.  In fact, we may as well compute the sgram too.
    $sgSo open $soundfile
    #set samplerate [expr 0.0+[$sgSo getSampleRate]]
    # DON'T use (guessed) srate from abbot audio stream - use global!
    # (but DO make sure it's cast to a float - else truncated labels!)
    set soundDuration [expr [$sgSo getFrames]/($samplerate+0.0)]
    # hence, step size = samples per pixel
    set sampsperpix [expr round($stepTime*$samplerate/$prb_timebasescale)]
    # Setup the time axis labels as well
    set tlb_max [expr $stepTime \
	                 *([winfo width $sg_tlb]/(0.0+$prb_timebasescale))]
    $sg_tlb config -maxVal $tlb_max
    # Choose FFT order and vertical scaling factor to track sampleRate
    set order [expr int(64*$samplerate/8000)]
    set sg_vfact [expr 128/$order]
    stftsound $sgSo $sgFa 0 0 $order $sampsperpix $order
    lassign [$sgFa Size] farows facols
    # Scale the data to be the top 30dB, scaled 0..1
    # .. but push 10dB at the very top into clipped white
    set amax [$sgFa Max]
    set dbrange 30.0
    set topclipdb 5.0
    $sgFa Offset [expr ($dbrange + $topclipdb) - $amax]
    $sgFa Scale  [expr 1.0/$dbrange]
    # Do the clip for the values above 1.0
    $sgFa Ceil 1.0
    # We have a 2-stage imaging scheme.  Make the first image, 
    # then copy it a bit at a time to the 2nd displayed image
    $sgIm config  -height $farows -data $sgFa
    # Create the displayed image which we will update
    $sgIm2 config -width $facols -height [expr $sg_vfact*$farows]
    $sgIm2 blank
    set sgTag "sgim"
    $sg_cnv delete $sgTag
    set xbase [expr int($padtime/2*$samplerate/$sampsperpix)]
    $sg_cnv create image $xbase 5 -anchor nw -image $sgIm2 -tags $sgTag
}

proc sgramUpdate {base nframes} {
    # Update the sgram image by copying some of it from sgIm to sgIm2
    global sgIm sgIm2 sg_vfact
    global samplerate prb_timebasescale stepTime
    # Convert the frames to pixels
    set sampsperpix [expr round($stepTime*$samplerate/$prb_timebasescale)]
    set basepix [expr int($base/$sampsperpix)]
    if {$nframes < 0} {
	set lastpix [$sgIm2 cget -width]
    } else {
	set lastpix [expr int(($base+$nframes)/$sampsperpix)]
    }
    set imht [$sgIm cget -height]
    # Do the copy
    $sgIm2 copy $sgIm -from $basepix 0 $lastpix $imht \
	    -to $basepix 0 -zoom 1 $sg_vfact
    # Update the display
    myUpdate
}

# Stuff for output labels
# Dpwe_Labelfilefns relies on having a font global defined:
set canvasFont 6x13

proc labelsReset {} {
    # Clear the labels panel prior to display
    global lab_cnv
    $lab_cnv delete "labels"
}

proc labelsDraw {labfile {ctmfile ""}} {
    # Draw up the output labels
    global lab_cnv samplerate soundDuration stepTime
    # How wide to make it? from probs.tcl
    global prb_timebasescale prb_xbase

    # Ever so slight horiz fix
    set xbase [expr $prb_xbase - 2]
    set ybase 5
    set tmin 0
    set tmax [expr $soundDuration - 2*$xbase*$stepTime/$prb_timebasescale]

#    set labs [ReadLabelFileNW $labfile $stepTime]
# noway-2.4 uses the CTM's step time to write actual times to the output; 
# also, writes durs not end times - use a new function
    set labs [ReadLabelFileNW2 $labfile]
    $lab_cnv delete "labels"

    # Maybe expand limits to encompass last label?
    lassign [lindex $labs e] lstart ldur
    set endtime [expr $lstart+$ldur+$stepTime/$prb_timebasescale]
    if {$endtime > $tmax} {
	puts "labelsDraw: duration extended from $tmax to $endtime"
	set tmax $endtime
    }

    set height 40
    set width [expr (($tmax-$tmin)/$stepTime)*$prb_timebasescale]

    CanvAddLabels $lab_cnv $labs $xbase $ybase $width $height $tmin $tmax 1 "labels"
    Debug "phnlabs: $labs"

    # Maybe draw up the words from the CTM file too?
    if {$ctmfile != ""} {
	set wlabs [ReadLabelFileCTM $ctmfile]
	Debug "wrdlabs (from $ctmfile): $wlabs"
	CanvAddLabels $lab_cnv $wlabs $xbase [expr $ybase+$height] $width $height $tmin $tmax 1 "labels"
    }

    feadispDraw
}

# Flag to detect new startups
set frnt_active 0
# Count of total frames
set frnt_frames 0

proc ReadFEPipe {pipe} {
    # Get samples from audio_frontend and write them to the 
    # forward-pass pipe
    global feapipe
    global frnt_active frnt_frames
    global afe_recordTime afe_soundfile
    Debug "ReadFEPipe: frames=$frnt_frames"

    if {!$frnt_active} {
	puts stderr "ReadFEP: new input, afe_recTime=$afe_recordTime"
	sgramInit $afe_soundfile
#	sgramUpdate 0 -1
	feadispReset $afe_recordTime
	# probsReset ??
	labelsReset

	set frnt_active 1
	set frnt_frames 0
    }
    set n [bfilecopy $pipe $feapipe 2048]
#    puts stderr "ReadFEPipe: out $n"
    set nframes [expr $n/2]	;# 2 bytes per sample
    sgramUpdate $frnt_frames $nframes
    incr frnt_frames $nframes

    if {[eof $pipe]} {
	rdprocClosePipe $pipe
	puts stderr "ReadFEPipe: closed ($n)"
#	DrawSgram
    flush $feapipe
	set frnt_active 0
    }
}

set decode_models 	"$SIMPLEUI_DATA_DIR/simple.noway.phmodels"
set decode_priors	"$SIMPLEUI_DATA_DIR/simple.noway.priors"
set decode_dict 	"$SIMPLEUI_DATA_DIR/simple.noway.dict"
set decode_trigram	"$SIMPLEUI_DATA_DIR/simple.noway.trigram"

set fwp_norm		"$SIMPLEUI_DATA_DIR/simple.norms"
set fwp_weights		"$SIMPLEUI_DATA_DIR/simple.wts"

set tmpphnfile "$TMPDIR/nowaytmpphn[pid].labels"
set ctmbase    "$TMPDIR/nowaytmpwrd[pid].labels"

set samplerate 16000

# swbforward doesn't work here !?! - writes a null output to probs??
set FWPASS "$QNSFWD \
    ftr1_file=- \
    ftr1_format=onlftr \
    ftr1_width=$nfeatures \
    ftr1_ftr_start=1 \
    ftr1_ftr_count=[expr $nfeatures - 1] \
    ftr1_norm_file=${fwp_norm} \
    window_extent=9 \
    ftr1_window_len=9 \
    init_weight_file=${fwp_weights} \
    mlp3_input_size=342 \
    mlp3_hidden_size=1000 \
    mlp3_output_size=56 \
    mlp3_output_type=softmax \
    realtime=true \
    activation_file=- \
    activation_format=rapbin \
    log_file=$TMPDIR/berplog[pid] \
    debug=1 \
    verbose=true"

set DECODE "$NOWAY \
    -phone_decode $tmpphnfile \
    -phone_models ${decode_models} \
    -priors       ${decode_priors} \
    -dictionary   ${decode_dict} \
    -trigram      ${decode_trigram} \
    -bin \
    -n_hyps 3 \
    -beam 3 \
    -state_beam 1 \
    -ctm $ctmbase \
    -frame_shift [expr int(1000*$stepTime)] \
    -new_lub \
    -acoustic_scale 0.125 \
    -prob_min 0.0000751 "
# for incremental output, add:
#     -inc_output 20 

# Remove any existing CTM (word alignment output, from -ctm) files
if {[set oldctms [glob -nocomplain "$ctmbase*"]] != ""} {
    foreach f $oldctms {
	puts stderr "removing old ctm '$f'..."
	file delete $f
    }
}

# Have to eval to get the $samplerate in $FEACALC to be expanded
set feapipe [eval "open \"| $FEACALC\" w+"]
fconfigure $feapipe -translation binary -buffering none -blocking 0
#full -buffersize 512
rdprocAdd $feapipe "ReadFeacalcPipe $feapipe"

set fwdpipe [open "| $FWPASS" "w+"]
fconfigure $fwdpipe -translation binary -buffering none -blocking 0
#full  -buffersize 512
rdprocAdd $fwdpipe "ReadBerpPipe $fwdpipe"

set decodepipe [open "| $DECODE" "w+"]
fconfigure $decodepipe -translation binary -buffering none -blocking 0
rdprocAdd $decodepipe "ReadDecodePipe $decodepipe"

set backendpipe [open "| cat" "w+"]

proc ReadNorms {normfile} {
    # Read a normfile, return two floatVecs, which are the offsets 
    # and the norm factors for to be applied to the feature vectors.
    # Norm files are "vec <size>\n<vals...>"
    if {[set f [Open $normfile "r"]] == ""} {
	Warn "ReadNorms: couldn't read '$normfile'"
	return ""
    }
    set vecs ""
    while {![eof $f]} {
	set line [gets $f]
	if {![eof $f]} {
	    # Read the "vec nn" line
	    if {![regexp {vec ([0-9]+)} $line all size]} {
		Warn "ReadNorms: header '$line' in '$normfile' not \"vec nn\""
		close $f
		return ""
	    }
	    # Read the vec vals
	    lappend vecs [set vec [floatVec \#auto $size]]
	    for {set i 0} {$i < $size} {incr i} {
		$vec Set $i [gets $f]
	    }
	}
    }
    close $f
    puts stderr "Read [llength $vecs] vectors from '$normfile' (len=[[lindex $vecs 0] Size])"
    return $vecs
}

# Set up the vectors for feature reading
set feavec [floatVec \#auto $nfeatures]
# Read in the norm vectors
lassign [ReadNorms $fwp_norm] fea_offs fea_scale

set fea_lastflag ""
set fea_col 0

proc ReadFeacalcPipe {pipe} {
    # Grab online features & copy to output
    global fwdpipe nfeatures
    global feavec fea_offs fea_scale fea_col
    global fea_lastflag
    # Read the flag
    set flag [breadi $pipe 1]
    if {$flag != $fea_lastflag} {
#   	   puts stderr "ReadFCP: lastflag=$fea_lastflag flag=$flag"
	if {$flag == 0} {
	    set fea_col 0
	    set fea_lastflag $flag
	}
    }
    Debug "ReadFeaCalcPipe: col=$fea_col flag=$flag"
    incr fea_col

#  puts stderr "ReadFCP:flag=$flag"
    if {[eof $pipe]} {
	rdprocClosePipe $pipe
	puts "Feacalc closed"
	# Propagate on to fwdpipe?
    } else {
	set feas {}
	for {set i 0} {$i < $nfeatures} {incr i} {
	    set f [breadf $pipe]
	    lappend feas $f
	    $feavec Set $i $f
	}
#  puts stderr "ReadFCP: read $i feas (like $f)"
	# Write to any further pipe
	if {$fwdpipe != ""} {
	    bwritei $fwdpipe $flag 1
	    foreach fea $feas {
		bwritef $fwdpipe $fea
	    }
	    flush $fwdpipe
	}
	# Condition the feature vector & send to display
	$feavec Sub $fea_offs
	$feavec Mul $fea_scale
	# map -2..2 to 0..1
	$feavec Offset 2.0
	$feavec Scale 0.25
	# Save
#  puts stderr "ReadFCP: adding [$feavec Size] pts to fea store"
	feadispAddCol $feavec
    }
#    puts stderr "ReadFCP: done ($flag)"
}

proc ReadDecodePipe {pipe} {
    # Output from the decoder
    global backendpipe tmpphnfile ctmbase

    set line [gets $pipe]
    Debug "ReadDecodePipe: $line"
    if {[eof $pipe]} {
	rdprocClosePipe $pipe
	puts "Decode closed"
    }
    # Strip the leading "count" returned by noway
    #regsub {^[0-9]* *} $line "" line
    regexp {([0-9]*) *(.*)} $line all uttnum line
    # Pass it to the backend (but not the first line)
    if {![string match "noway*" $line]} {
	puts $backendpipe $line
	# This is also a good time to display the labels
	set ctmfile $ctmbase.$uttnum
	while {![file exists $ctmfile] || [file size $ctmfile] == 0} {
	    puts stderr "Waiting for ctmfile '$ctmfile'..."
	    exec sleep 1
	}
	labelsDraw $tmpphnfile $ctmfile
	# Delete the CTM file and truncate the phnfile
	set dodelete 1
	if {$dodelete} {
	    file delete $ctmfile
	    set f [open $tmpphnfile "w"]
	    close $f
	}
    }
}

# Delete tmp files on exit
proc _cleanup {} {
    # Remove any temporary files just before quitting
    global tmpphnfile ctmfile
    catch {file delete $tmpphnfile}
    catch {file delete $ctmfile}
}

# Trap "exit" proc (if we haven't trapped it already)
if {[info command "orig_exit"] == ""} {

    rename exit orig_exit

    proc exit {{code 0}} {
	# Do the cleanup on exit
	_cleanup
	orig_exit $code
    }
}


# Making prb_inpipe "" means we have to set up our own fileevent
set prb_inpipe ""
# specifying prb_outpipe will get all input copied here
set prb_outpipe $decodepipe
# Height of the probs display
set prb_height 60
# Offset at left, to align with spectrogram
set prb_xbase  20
# Don't force updates? (only makes sync worse)
#set prb_updates ""
source "$DEMOSHAREDIR/probs.tcl"

set afe_outputHandler ReadFEPipe
set argv {useloadsave=1 debug=0}

#source "$AFE_TCL_DIR/audio_frontend.tcl"
audio_frontend_init

# Turn off the recording during the initial prompt
$afe_audiosrc close

# Load the berp backend
set grammar   "$BERP_DATA_DIR/berp.grammar"
set database  "$BERP_DATA_DIR/restaurants"
set audiodir  "$BERP_DATA_DIR/sounds"
set unify     "false"
set db_prog   "$DEMOSHAREDIR/simpdb"
set parsetree "true"
set ptprog    "$DEMOSHAREDIR/ptbrowser"
set be_inpipe $backendpipe

source        "$DEMOSHAREDIR/berpbackend.tcl"

# Turn on the fileevents
rdprocEnable
$afe_audiosrc open
