#!/usr/local/bin/wish
#
# probs.tcl
#
# Berpdemo component to display the phone-probabilities as they 
# come out of the NN classifier.
#
# A replacement for fosler's version in perl
#
# dpwe@icsi.berkeley.edu 1996aug19
# $Header: /u/drspeech/src/berpdemo/share/RCS/probs.tcl,v 1.15 1999/06/09 22:36:14 dpwe Exp $

# Wait for keypress at end of stream?
if {[file tail $argv0] == "probs.tcl"} {
    set prb_eou_wait 1
} else {
    set prb_eou_wait 0
}

# Capture the RCS version for display
set prb_revision {$Revision: 1.15 $}

# Defaults for input format - VERY confusing names, but 'standard'
#  - see InitBerpPipe/ReadBerpPipe for interpretation
set prb_ascii  0
set prb_binary 1

# Get the fns to read icsi-style args
package require Dpwe_Utilfns
# Get the fns to manage enabling/disabling "file readable" procs
#source "[file dirname [info script]]/rdprocmgr.tcl"
package require Dpwe_Rdprocmgr

# load definitions of breadf and breadi in shared library (Tcl 7.5 up)
package require Dpweutils_tcl

# Global currframe is set to specify the x-axis
set prb_currframe -

set prb_width 700
# The vertical excursion of the graphs - excludes labelline and decoration
if {[info vars "prb_height"] == ""} {
    set prb_height 100
}
if {[info vars "prb_canvasFont"] == ""} {
    set prb_canvasFont "fixed"
# The space left for the line of text labels at the top
    set prb_lht 16
}
# timebase scaling
set prb_timebasescale 4
# Vertical scaling
set prb_heightscale [expr $prb_height/100.0]
if {[info vars "prb_xbase"] == ""} {
    # Start x at 1.5 frames in, to align for processing swallowage
    set prb_xbase [expr 2+int(1.5*$prb_timebasescale)]
}
set prb_ybase 2

# If we're passed in $prb_root, use it as the basis for our widgets
# else make it nothing
if {[info vars "prb_root"] == ""} {
    set prb_root ""
}
# File ID of input stream
if {[info vars "prb_inpipe"] == ""} {
    set prb_inpipe "stdin"
}
# Stream to copy input to (defaults to none)
if {[info vars "prb_outpipe"] == ""} {
    set prb_outpipe ""
}
# Interval for doing updateds; 0 = none; default = 5 (input frames)
if {[info vars "prb_updates"] == ""} {
    set prb_updates "5"
}

if {![winfo exist $prb_root.top]} {
    # Set up basic widgets, including display of $prb_currframe
    frame $prb_root.top

    # Note the revision small on the screen
    regexp {:(.*)\$} $prb_revision all revision
    label $prb_root.top.version -text "[file tail [info script]] $revision" -font 5x8
    pack  $prb_root.top.version -side right -fill y -expand 0

    label $prb_root.top.label -relief flat -text "Frame:"
    label $prb_root.top.frame -relief sunken -textvariable prb_currframe
    pack $prb_root.top.label $prb_root.top.frame -side left -fill y -expand 0
    pack $prb_root.top -side top -fill x -expand 0

    set prb_cnv [canvas $prb_root.cnv -width [expr $prb_width+2*$prb_xbase] \
			 -height [expr $prb_height+2*$prb_ybase+$prb_lht] \
			 -borderwidth 2 -relief sunken]
    pack $prb_root.cnv -side top -fill both -expand 1
}

# If embedder hasn't defined an occaisional update routine, do it here
if {[info proc myUpdate] == ""} {
    proc myUpdate {} {
	# Call updates (for the screen only)
	rdprocDisable
#puts stderr "probs: abt to update"
        update
        rdprocEnable
    }
}

if {[info proc Debug] == ""} {
    set _debug 0
    proc Debug {msg} {
	global _debug
	if {$_debug} {
	    puts stderr "Debug: $msg"
	}
    }
}

# prb_valArray() holds the most recent value for each bin
set prb_valArray(0) 0
set prb_labeledArray(0) 0
set prb_nameArray(0) ""
set prb_colorArray(0) black
set prb_colors {black blue red limegreen turquoise4 magenta yellow orange slategray4}
# initialize the colorArray
set prb_maxColors 60

proc init_pset {{psetfile "%d/icsi56.phset"}} {
    # Read in the phoneset file & setup the arrays
    global prb_colorArray prb_nameArray prb_labeledArray prb_valArray 
    global prb_colors
    global BERP_DATA_DIR
    # "%d" in the file name becomes $BERP_DATA_DIR
    regsub -all {%d} $psetfile $BERP_DATA_DIR psetfile

    if {[set psf [open $psetfile "r"]] == ""} {
	puts stderr "probs: couldn't read phoneset file $psetfile"
	exit -1
    }
    # First line is number of phonemes
    set nn [gets $psf]
    # Remaining lines are phoneme definitions
    for {set i 0} {$i < $nn} {incr i} {
	set x [expr $i % [llength $prb_colors]]
	set prb_colorArray($i) [lindex $prb_colors $x]
	set prb_nameArray($i) [lindex [gets $psf] 0]
	# labeledArray keeps track of whether this one has been labeled lately
	set prb_labeledArray($i) 0
	set prb_valArray($i) 0
    }
    close $psf
}

if {[info vars BERP_DATA_DIR] == ""} {
    set BERP_DATA_DIR "/u/drspeech/demos/berp"
}

init_pset

# plot is called by the perl script
set prb_labthrsh 50
set prb_lowthrsh 10
set prb_showthrsh 5
proc plot {num val} {
    # Update trace $num to value $val
    global prb_cnv prb_currframe 
    global prb_valArray prb_colorArray prb_labeledArray prb_nameArray
    global prb_xbase prb_ybase prb_height prb_heightscale prb_timebasescale 
    global prb_lht prb_canvasFont
    global prb_labthrsh prb_lowthrsh prb_showthrsh
#    puts stderr "plot $num = $val @ $prb_currframe"
#    if {$prb_currframe > 0} {
	set lastval $prb_valArray($num)
	if {$lastval > $prb_showthrsh || $val > $prb_showthrsh} {
	    set lasty [expr $prb_ybase + $prb_height + $prb_lht - $prb_heightscale*$lastval]
	    set lastx [expr $prb_xbase + $prb_timebasescale*($prb_currframe - 1)]
	    set x [expr $prb_xbase + $prb_timebasescale*$prb_currframe]
	    set y [expr $prb_ybase + $prb_height + $prb_lht - $prb_heightscale*$val]
	    $prb_cnv create line $lastx $lasty $x $y -tag t$num \
		    -fill $prb_colorArray($num)
	    if {($val > $prb_labthrsh || $lastval > $prb_labthrsh) \
		    && $val < $lastval \
		    && $prb_labeledArray($num) == 0 } {
		$prb_cnv create text $x $prb_ybase -anchor n -tag t$num \
			-font $prb_canvasFont \
			-fill $prb_colorArray($num) -text $prb_nameArray($num)
		set prb_labeledArray($num) 1
	    }
	    set prb_valArray($num) $val
	    if {$val < $prb_lowthrsh} {
		set prb_labeledArray($num) 0
	    }
	}
#    }
#    update
    # update was needed in the driven-from-perl version (maybe), but now
    # have internal version where plot is driven from a fileevent
}

proc clearcanvas {} {
    # Clear the canvas prior to a new stream
    global prb_cnv
    $prb_cnv delete all
}

# Number of tokens in each time frame (row) - read from stream
set prb_rowlen 0

set prb_pipeInited 0

proc InitBerpPipe {pipe} {
    # Start off reading the berp pipe
    global prb_rowlen prb_ascii prb_binary prb_pipeInited prb_outpipe
    # Clear EOL translation on input pipe if not ascii (Tcl7.5-ism)
    if {!$prb_ascii && [info commands fconfigure] != ""} {
	fconfigure $pipe -translation binary -buffering none -blocking 0
	if {$prb_outpipe != ""} {
	    fconfigure $prb_outpipe -translation binary \
		    -buffering none -blocking 0
	}
    }
    if {$prb_binary} {
	# binary number starts stream?
	set prb_rowlen [breadi $pipe]
	if {$prb_rowlen == ""} {
	    Debug "ReadBerpInit: nothing"
	    exit
	}
	if {$prb_outpipe != ""} {bwritei $prb_outpipe $prb_rowlen}
    } else {
	# skip input line
	set dummy [gets $pipe]
	if {$prb_outpipe != ""} {puts $prb_outpipe $dummy}
	set prb_rowlen  [gets $pipe]
	if {$prb_outpipe != ""} {puts $prb_outpipe $prb_rowlen}
    }
    Debug "InitBerpPipe: rowlen=$prb_rowlen"
    set prb_pipeInited 1
}    

proc ReadBerpPipe {pipe} {
    # Read a single row of data from the input pipe
    global prb_done prb_outpipe prb_updates
    global prb_currframe prb_ascii prb_binary
    global prb_rowlen prb_pipeInited
    global prb_showthrsh prb_valArray
    global prb_eou_wait

#    Debug "ReadBerpPipe: currframe=$prb_currframe"
    # Disable the fileevent for now
    #fileevent $pipe readable {}
    if {$prb_pipeInited == 0} {
	InitBerpPipe $pipe
    } else {
	set prb_done 0
	if {$prb_ascii} {
            if {[gets $pipe line] < 0} {
		puts stderr "ReadBerpPipe: stdin error"; rdprocClosePipe $pipe; set prb_done 1
	    }
	    if {$prb_outpipe != ""} {puts $prb_outpipe $line}
            if {[regexp {^-?[0-9]+$} $line]} {
		# Line is just a number - it's the frame
		set prb_currframe $line
		# reset the phone index
		set counter 0
	    } elseif {[regexp {^ *$}]} {
		# just whitespace
	    } else {
		# it's a row of parameter values
		foreach parm $line {
		    plot $counter [expr int($parm*100.0)]
		    incr counter
		}
	    }
	} else {
	    # Not ascii input - pure binary
	    set prb_currframe [breadi $pipe]
	    Debug "ReadBerpPipe: currframe=$prb_currframe"
	    if {$prb_currframe == ""} {
		Debug "ReadBerp: nothing"
		# Will take it as EOF
	    }
	    if {$prb_outpipe != "" && $prb_currframe != {}} {
		bwritei $prb_outpipe $prb_currframe
	    }
	    # (will return {} on eof)
	    if {$prb_currframe != -1 && $prb_currframe != {}} {
	        for {set i 0} {$i < $prb_rowlen} {incr i} {
		    #set v [expr int([breadf $pipe]*100.0)]
		    set f [breadf $pipe]
		    if {$prb_outpipe != ""} {bwritef $prb_outpipe $f}
		    set v [expr int($f*100.0)]
		    if {$prb_valArray($i) > $prb_showthrsh \
			    || $v > $prb_showthrsh} {
			plot $i $v
		    }
                }
	    }
	    if {$prb_currframe == -1 && $prb_eou_wait} {
		# Optional pause at end of utterance
		rdprocDisable
		if {[tk_dialog .askyn "Continue?" "EOU reached - go on?"  "" 0 "OK" "Quit"] == 1} {
		    exit
		}
		rdprocEnable
	    }
	}
	# Try to speed it up by only re-rendering the display every 5 ticks
	if {$prb_updates != "" && $prb_updates > 0 \
		&& $prb_currframe != {} \
		&& ($prb_currframe % $prb_updates) == 0 } { 
	    # Have to disable the filehandler while updating, else 
	    # just finish the file before doing the screen
	    myUpdate
	}
    }
    # Clear canvas if currframe == 0 (start of new utterence)
    if {$prb_currframe == 0} {
	clearcanvas
    }

    # Kill the application if the input goes EOF
    if {[eof $pipe]} {puts stderr "ReadBerpPipe: EOF"; rdprocClosePipe $pipe; if {$prb_outpipe != ""} {rdprocClosePipe $prb_outpipe}; set prb_done 1;}
}

if {$prb_inpipe != ""} {
    # Either no pipe was specified (& we're using stdin)
    # or we were given an actual pipe
    #fileevent $prb_inpipe readable "ReadBerpPipe $prb_inpipe"
    rdprocAdd $prb_inpipe "ReadBerpPipe $prb_inpipe"
    set prb_done 0
    rdprocEnable
    vwait prb_done
#    exit
}

# If this is embedded & sourced with prb_inpipe set to "", 
# the pipe doesn't actually get setup - 
# it's up to the caller to setup their own fileevent
