. $Header: /u/drspeech/src/dr_scripts/RCS/dr_embed.man,v 1.8 1997/12/15 19:25:47 fosler Exp fosler $
.\" A few notes:
.\" i) Don't add extra blank lines - these appear in the output and
.\"    leaves too much space when display on the screen
.\"
.\" This title line needs to be changed to contain your program name,
.\" but the date will be changed automatically by RCS.
.TH dr_embed 1 "$Date: 1997/12/15 19:25:47 $" ICSI
.\" Don't do anything clever in the section below - it messes up the
.\" whatis database that's accessed using "man -k"
.SH NAME
dr_embed \- do embedded training of an MLP
.SH SYNOPSIS
.B dr_embed [options]
.SH DESCRIPTION
.I dr_embed
performs iterative, embedded training of an MLP-based speech recognition
system.  While it currently uses a fixed lexicon, ultimately it will
incorporate lexicon training as well.

The process starts with a training feature file, testing feature file,
lexicon, and initial set of MLP weights (the boot weights) or an
initial labelling of the training set (the boot labels).  It then
performs multiple iterations of the following process:

1) Do a forced-Viterbi labeling of the training set using the most
recent set of MLP weights and the lexicon.

2) Use the labels to train a new MLP.

3) Do a recognition on the testing feature file.

When starting from boot weights, the process begins at (1).  When
starting from boot labels, the process begins at (2).

.SH ENVIRONMENT VARIABLES

The location of the
.BR drspeech(1)
and
.BR drspeech_args(1)
packages must be in your PERLLIB environment variable.

.SH FILES

.I dr_embed
has many inputs and outputs.  Minimally, it requires a boot weights
file, a boot priors file, a training feature file, a training norm file,
a file of word transcriptions for the utterance in the training set, a
testing feature file, a testing norm file, a file of word transcriptions
for the utterance in the testing set, a phoneset file, and a lexicon
file.  For each iteration, it produces a label file, a priors file, a
weight file, a recognition results file, and a recognition score file.
It also produces a log file for the whole embedded training.

.SH FILENAME TEMPLATES AND ESCAPES
.I dr_embed
produces a lot of temporary and output files.  To give the user control
over the naming of these files and the locations to which they are
written, a template mechanism is used.  A template is a string of
regular characters and escapes characters that
.I dr_embed
expands to form a filename.

There are four escapes:

%j expands to the value of the ``job'' command-line parameter.

%i expands to the name of the current iteration of embedded training.
The first iteration is named ``boot'' while subsequent iterations are
named ``iter1'' ``iter2'' and so on.  All templates must contain this
escape.

%p expands to the name of the previous iteration of embedded training.
In the boot iteration, %p expands to ``-''.

%% expands to %

While all of the default filename templates expand to files that are in
the current directory, ii is possible to specify templates that point to
other directories.  Thus, if you wanted to write the recognition scores
from your embedded training into a directory named ``scores'' that is
below the current directory, you could set
.if n .ti +0    \" Useful way of moving on to the next line
scores_filename_template=./scores/%j+%i.score
.if n .ti +0    \" Useful way of moving on to the next line

The %p escape is provided to allow a neat trick.  If 
.if n .ti +0    \" Useful way of moving on to the next line
log_weights_filename_template=%j+%i+%%e.weights
.if n .ti +0    \" Useful way of moving on to the next line
and you set
.if n .ti +0    \" Useful way of moving on to the next line
train_init_weights_filename_template=%j+%p+1.weights
.if n .ti +0    \" Useful way of moving on to the next line
then there is special code to ensure that during the boot iteration the
MLP training will start from random weights, while in subsequent
iterations the MLP training will start from the first set of log weights
generated by the previous MLP training.  The %p escape probably should
not be used for anything else.  If you use this trick, you
should make sure that
.if n .ti +0    \" Useful way of moving on to the next line
clean_log_weightfiles=false
.if n .ti +0    \" Useful way of moving on to the next line

Note that the ``%%e'' string in log_weights_filename_template expands to
``%e'' --- an escape for
.BR swbtrain(1)
that expands to the current training epoch.

.SH OPTIONS
.TP 1i			\" Indent the paragraph
.B help=<any> (optional) [no default]
Print a usage message, then exit without doing any computation.
.TP
.B verbose=<integer> (optional) [default=1]
Set verbosity level for script.  verbose=0 means produce no logging
output.
.TP
.B logname=<filename> (optional) [default=-]
Log file for embedded training.
.TP
.B params=<filename> (optional) [no default]
Name of parameter file to read.  The file is ASCII, with one
<arg>=<value> pair per line.
.TP
.B job=<filename> (required)
Name that uniquely identifies the embedded training job.  This string is
substituted for the %j escape.
.TP
.B niter=<cardinal> (required)
Number of iterations of embedded training to run.
.TP
.B boot_weights=<filename>
MLP weights used to generate inital labeling of the training data.
Note that these weights are used only on the first iteration of
embedded training, to generate labels.  If you want to specify a set
of weights to initialize the MLP each time it is being trained, use
the train_init_weights_filename_template parameter.
.TP
.B boot_labels=<filename>
pfile containing an initial labeling of the training data.
.TP
.B boot_priors=<filename> (required)
Priors used to generate inital labeling of the training data.
.TP
.B ftr_start=<cardinal> (required)
First feature to use during MLP training and MLP forward pass.  Features
are numbered from 0.
.TP
.B ftr_count=<cardinal> (required)
Number of features to use during MLP training and MLP forward pass.
.TP
.B train_ftr_file=<filename> (required)
Feature file that contains the training and cross-validation data.
Currently this file must be in
.BR pfile(5)
format.
.TP
.B train_sent_start=<cardinal> (required)
First utterance to use in MLP training.  Sentences are numbered from 0.
.TP
.B train_sent_count=<cardinal> (required)
Number of utterances to use in MLP training.
.TP
.B fvit_sent_start=<cardinal> (required)
First utterance to use in forced alignment.  Sentences are numbered from
0.  This should usually be the first utterance in the training feature
file.
.TP
.B fvit_sent_count=<cardinal> (required)
Number of utterances to use in forced alignment.  This should usually be
the number of utterances in the training feature file.
.TP
.B cv_sent_start=<cardinal> (required)
First utterance to use in cross-validation during MLP training.
.TP
.B cv_sent_count=<cardinal> (required)
Number of utterances to use in cross-validation during MLP training.
.TP
.B train_norm_file=<filename> (required)
Norm file for training data.  In
.BR norms(5)
format.
.TP
.B train_sents=<filename> (required)
Word transcriptions of the utterances in the training set.
.TP
.B train_cache_frames=<cardinal> (required)
Number of frames in the
.BR swbtrain(1)
training cache.
.TP
.B test_ftr_file=<filename> (required)
Feature file that contains the testing data.  Currently this file must
be in
.BR pfile(5)
format.
.TP
.B test_sent_start=<cardinal> (required)
First utterance to use in testing.
.TP
.B test_sent_count=<cardinal> (required)
Number of sentences to use in testing.
.TP
.B test_norm_file=<filename> (required)
Norm file for testing data.  In
.BR norms(5)
format.
.TP
.B test_sents=<filename> (required)
Word transcriptions of the utterances in the testing set.
.TP
.B phoneset=<filename> (required)
Phoneset used to label the training and testing data.  In
.BR phoneset(5)
format.
.TP
.B lexicon=<filename> (required)
Lexicon of word models used for forced alignment and recognition.
Currently this file must be in
.BR lexicon(5)
format.
.TP
.B mlp3_input_size=<cardinal> (required)
Size of the MLP input layer.
.TP
.B mlp3_hidden_size=<cardinal> (required)
Size of the MLP hidden layer.
.TP
.B mlp3_output_size=<cardinal> (required)
Size of the MLP output layer.
.TP
.B window_extent=<cardinal> (required)
Size of the union of all windows on input features.
.TP
.B future_context=<cardinal> (required)
Number of frames of future context in the MLP input window.
.TP
.B past_context=<cardinal> (required)
Number of frames of past context in the MLP input window.
.TP
.B ftr_window_offset=<cardinal> (required)
Offset in frames on MLP input window.
.TP
.B sample_rate=<float> (required)
Sampling rate for the speech we are training on and recognizing.
.TP
.B frame_step=<cardinal> (required)
Number of milliseconds between adjacent speech frames.
.TP
.B train_init_weights_filename_template=<any> (optional) [default=]
Template for weights file to use for MLP initialization during training.
Default means start with random weights.  The most common use of this
variable is to use it to initialize the current MLP being trained with
the weights from the first epoch of the previous training.  Note that
this variable is not used to specify a starting MLP to create an
initial labeling of the training data.  If that's what you want to
specify, set the boot_weights parameter.
.TP
.B bigram=<opt_filename> (optional) [default=]
Bigram grammar to use during recognition.  If this option is used, the
wpbegin, wpcont, or wpend options cannot be used.  File currently must
be in
.BR bigram(5)
format.
.TP
.B wpbegin=<opt_filename> (optional) [default=]
Word-pair grammar begin file to use during recognition.  If this option
is used, the wpcont and wpend options must also be used, and the bigram
option cannot be used.
.TP
.B wpcont=<opt_filename> (optional) [default=]
Word-pair grammar middle file to use during recognition.  If this
option is used, the wpbegin and wpend options must also be used, and the
bigram option cannot be used.
.TP
.B wpend=<opt_filename> (optional) [default=]
Word-pair grammar end file to use during recognition.  If this option is
used, the wpbegin and wpend options must also be used, and the bigram
option cannot be used.
.TP
.B mlp3_output_type=<label> (optional) [default=softmax]
MLP output unit type.
.TP
.B mlp3_fx=<boolean> (optional) [default=true]
Use fixed-point in MLP evaluation.  If this is false, use_spert should
probably also be false.
.TP
.B use_spert=<boolean> (optional) [default=true]
Run MLP training and forward passes on a SPERT board.  If this is false,
mlp3_fx should probably also be false.
.TP
.B fvit_lmscale=<float> (optional) [default=1]
Language scaling factor to use during forced Viterbi.
.TP
.B fvit_amscale=<float> (optional) [default=1]
Acoustic scaling factor to use during forced Viterbi.
.TP
.B fvit_ptscale=<float> (optional) [default=1]
Phone transition scaling factor to use during forced Viterbi.
.TP
.B fvit_wordtranspenalty=<float> (optional) [default=1]
Word transition penalty to used during forced Viterbi.
.TP
.B fvit_maxactivewords=<cardinal> (optional) [default=30]
Maximum number of active words allowed during forced Viterbi.
.TP
.B fvit_startpruningframe=<integer> (optional) [default=-1]
Frame to start pruning during forced Viterbi.
.TP
.B fvit_transcut=<float> (optional) [default=0]
Word transition pruning during forced Viterbi.
.TP
.B fvit_beamwidth=<float> (optional) [default=120]
Search beam width during forced Viterbi.
.TP
.B fvit_phoneprune=<float> (optional) [default=-1]
Phone pruning during forced Viterbi.
.TP
.B recog_lmscale=<float> (optional) [default=1]
Language scaling factor to use during recognition.
.TP
.B recog_amscale=<float> (optional) [default=1]
Acoustic scaling factor to use during recognition.
.TP
.B recog_ptscale=<float> (optional) [default=1]
Phone transition scaling factor to use during recognition.
.TP
.B recog_wordtranspenalty=<float> (optional) [default=1]
Word transition penalty to use during recognition.
.TP
.B recog_maxactivewords=<cardinal> (optional) [default=30]
Maximum nuber of active words during recognition.
.TP
.B recog_startpruningframe=<integer> (optional) [default=-1]
Frame to start pruning during recognition.
.TP
.B recog_transcut=<float> (optional) [default=0]
Word transition pruning during recognition.
.TP
.B recog_beamwidth=<float> (optional) [default=120]
Search beam width during recognition.
.TP
.B recog_phoneprune=<float> (optional) [default=-1]
Phone pruning during recognition.
.TP
.B weights_filename_template=<any> (optional)
.in 0.5i
.B [default=%j+%i.weights]
.in 1.5i
Template for names of weight files generated during MLP training.
.TP
.B log_weights_filename_template=<any> (optional)
.in 0.5i
.B [default=%j+%i+%%e.weights]
.in 1.5i
Template for names of log weight files generated during MLP training.
.TP
.B priors_filename_template=<any> (optional) [default=%j+%i.priors]
Template for names of priors files generated by the embedded training
process.
.TP
.B wordpaths_filename_template=<any> (optional)
.in 0.5i
.B [default=%j+%i.wpaths]
.in 1.5i
Template for names of wordpath files generated during forced
alignment.
.TP
.B alignbin_filename_template=<any> (optional)
.in 0.5i
.B [default=%j+%i.alignbin]
.in 1.5i
Template for names of binary align files generated during forced
alignment.
.TP
.B align_filename_template=<any> (optional) [default=%j+%i.align]
Template for names of align files files generated during forced
alignment.
.TP
.B label_filename_template=<any> (optional)
.in 0.5i
.B [default=%j+%i+labels.pfile]
.in 1.5i
Template for names of label files generated during forced alignment.
.TP
.B results_filename_template=<any> (optional)
.in 0.5i
.B [default=%j+%i.results]
.in 1.5i
Template for names of result files generated during recognition.
.TP
.B scores_filename_template=<any> (optional) [default=%j+%i.score]
Template for names of score files generated during recognition.
.TP
.B align_prog=<filename> (optional) [default=align]
Program to use for forced alignment.
.TP
.B align_args=<any> (optional) [default=]
Optional arguments to forced alignment program.
.TP
.B pfile_prog=<filename> (optional) [default=pfile]
Program to generate pfiles.
.TP
.B pfile_args=<any> (optional) [default=]
Optional arguments to pfile program.
.TP
.B activation_format=<label> (optional) [default=rapbin]
Format of activations from forward passes.
.TP
.B spertrun_prog=<opt_filename> (optional) [default=spertrun]
Program to execute SPERT programs.
.TP
.B spertrun_args=<any> (optional) [default=]
Optional arguments to SPERT program executor.
.TP
.B swbforward_prog=<filename> (optional) [default=swbforward]
Program to do forward pass on MLP.
.TP
.B swbforward_args=<any> (optional) [default=]
Optional arguments to MLP forward pass program.
.TP
.B swbtrain_prog=<filename> (optional) [default=swbtrain]
Program to train MLPs.
.TP
.B swbtrain_args=<any> (optional) [default=]
Optional arguments to MLP training program.
.TP
.B wordscore_prog=<filename> (optional) [default=wordscore]
Program for scoring recognition output.
.TP
.B wordscore_args=<any> (optional) [default=]
Optional arguments to recognition scoring program.
.TP
.B y0_prog=<filename> (optional) [default=y0]
Speech decoding program.
.TP
.B y0_machine=<hostname> (optional) [default=localhost]
Host to run alignments on.
.TP
.B y0_args=<any> (optional) [default=]
Optional arguments to speech decoder.
.TP
.B y0_fvit_prog=<opt_filename> (optional) [default=]
Speech decoding program to use during forced alignment.  If unspecified,
the value of the y0_prog argument is used.
.TP
.B y0_fvit_args=<any> (optional) [default=]
Optional arguments to forced-alignment speech decoder.  If unspecified,
the value of the y0_args argument is used.
.TP
.B y0_recog_prog=<opt_filename> (optional) [default=]
Speech decoding program to use during recognition.  If unspecified, the
value of the y0_prog argument is used.
.TP
.B y0_recog_args=<any> (optional) [default=]
Optional arguments to recognition speech decoder.  If unspecified,
the value of the y0_args argument is used.
.TP
.B conserve_disk=<boolean> (optional) [default=false]
Conserve disk space by deleting intermediate files.  Setting this flag
true is equivalent to setting all of the following flags true.
.TP
.B clean_alignbinfiles=<boolean> (optional) [default=false]
Delete binary alignment files.
.TP
.B clean_alignfiles=<boolean> (optional) [default=false]
Delete ASCII alignment files.
.TP
.B clean_log_weightfiles=<boolean> (optional) [default=false]
Delete log weightfiles from MLP training.
.TP
.B clean_wordpathfiles=<boolean> (optional) [default=false]
Delete wordpath files from forced alignment.
.TP
.B do_testing=<boolean> (optional) [default=true]
Perform a recognition test after every iteration.
.TP
.B start_iter=(boot|iterN) (optional) [default=boot]
(Re)start training from this iteration.  
.TP
.B reuse_labels=<boolean> (optional) [default=false]
If restarting, check for existing labels and priors files to use.

.SH SAMPLE PARAMETER FILE

Here is a sample parameter file for an MLP with the usual 9-frame
context window that is centered on the current frame.  This is given for
illustrative purposes only.  Don't count on any of the files actually
existing.

.nf
job=plp_baseline
logname=plp_baseline.dr_embed_log
niter=4
conserve_disk=1
train_ftr_file=../../pfile/numbers93+train+plp8+w25s12_5+F+alignable.pfile
test_ftr_file=../../pfile/numbers93+train+plp8+w25s12_5+F+alignable.pfile
train_norm_file=../../pfile/numbers93+train+plp8+w25s12_5+F.norm
test_norm_file=../../pfile/numbers93+train+plp8+w25s12_5+F.norm
ftr_start=1
ftr_count=8
window_extent=15
ftr_window_offset=0
past_context=4
future_context=4
boot_weights=ntimit+plp8+w25s12_5+F+icsi56+epoch1.wts
mlp3_input_size=72
mlp3_hidden_size=512
mlp3_output_size=56
lexicon=../../lex/numbers93.lex
bigram=/u/lingling/NUMBER/grammar/classbigramLogi+bigram
boot_priors=../../list/ntimit.priors
recog_wordtranspenalty=0.1
recog_lmscale=8
train_sents=../../wordtrans/num93+train+alignable.answer
test_sents=../../wordtrans/num93+cv.answer
sample_rate=8000
frame_step=100
phoneset=/u/drspeech/data/phonesets/icsi56.phset
train_sent_start=0
train_sent_count=786
fvit_sent_start=0
fvit_sent_count=875
cv_sent_start=786
cv_sent_count=89
train_cache_frames=60000
test_sent_start=786
test_sent_count=89
.fi

.SH BUGS
Does not do lexicon training.
.SH AUTHOR
Brian Kingsbury <bedk@icsi.berkeley.edu>
.SH SEE ALSO
.BR align(1), pfile(1), spertrun(1), swbforward(1), swbtrain(1),
.BR wordscore(1), y0(1), bigram(5), lexicon(5), norms(5), pfile(5),
.BR phoneset(5)
