/*
 * olfnorms.c
 *
 * Calculate norms from an online feature file.
 * 
 * 1999nov08 dpwe@icsi.berkeley.edu
 * $Header: /u/drspeech/src/ffwd/RCS/olfnorms.c,v 1.1 1999/11/09 05:08:02 dpwe Exp $
 */

#include <stdio.h>
#include <stdlib.h> /* for atoi */
#include <math.h>
#include <assert.h>
#include "io.h"

int width = 0;
char *ifile = NULL;
char *ofile = NULL;

void usage(char *name)
{
  fprintf(stderr,"usage: %s [OPTIONS]\n",name);
  fprintf(stderr,"where [OPTIONS] include:\n");  
  fprintf(stderr,"  -width w                : input ftr width.\n");
  fprintf(stderr,"  -ifile file             : input ftr file (online_ftr format)\n");
  fprintf(stderr,"  -ofile file             : output norms file (qnnorm format)\n");
}

void
parse_args(int argc,char*argv[])
{
  int i;
  if (argc == 1) {
    usage(argv[0]);
    exit(-1);
  }
  for (i=1;i<argc;i++) {
    if (argv[i][0] == '-') {
      if (!strcmp(&(argv[i][1]),"width")) {
	width = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"ifile")) {
	ifile = argv[++i];
      } else if (!strcmp(&(argv[i][1]),"ofile")) {
	ofile = argv[++i];
      } else  {
        fprintf(stderr,"Error: Unknown or invalid argument (%s)\n",argv[i]);
	exit(-1);
      }
    }
  }
  if (width == 0 || ifile == NULL || ofile == NULL) {
    fprintf(stderr,"Error with arguments\n");
    usage(argv[0]);
    exit(-1);
  }
}

int
main(int argc,char *argv[]) 
{
    float *sumx;
    float *sumx2;
    float *buf;

    int i, npats = 0;

    int rc;

    FILE *ipf;

    parse_args(argc,argv);

    ipf = onlftrs_open_r(ifile);

    sumx = (float *)malloc(width*sizeof(float));
    sumx2 = (float *)malloc(width*sizeof(float));
    buf = (float *)malloc(width*sizeof(float));

    for (i=0; i<width; ++i) {
        sumx[i] = 0;
        sumx2[i] = 0;
    }

    while (onlftrs_read(ipf, buf, width) != -1) {
        for (i = 0; i < width; ++i) {
            sumx[i] += buf[i];
            sumx2[i] += buf[i]*buf[i];    
        }
	++npats;
    }

    onlftrs_close(ipf);

    for (i=0; i<width; ++i) {
        sumx[i] /= (float)npats;
        sumx2[i] = sqrt(fabs(sumx[i]*sumx[i] - (sumx2[i]/npats)));
    }

    if (writenorms(ofile, width, sumx, sumx2) == 0) {
	fprintf(stderr, "Error writing norms file '%s'\n", ofile);
	exit(1);
    }

    exit(0);
}
