/*
** $Header: /u/spert/src/fltvec/nfp/vecele.c,v 1.1 1996/03/22 03:13:00 davidj Exp $
**
** "NFP" vector elementwise routines - typically use the integer routines
*/


#include <fltv-config.h>
#include <stddef.h>
#ifdef FLTV_HAVE_LIBINTVEC
#include "intvec.h"
#endif
#include <fltv-defs.h>
#include <fltv-scansrc.h>

/****************************************************************/
/* copy_f_vf implemented using the corresponding int32 routine  */
/****************************************************************/

#ifdef DECL_nfp_copy_f_vf
void nfp_copy_f_vf(size_t len, float from, float* to);
#endif

#ifdef DEFN_nfp_copy_f_vf
void
nfp_copy_f_vf(size_t len, float from, float* to)
{
#if defined(FLTV_HAVE_LIBINTVEC)
    fltv_float_int32_t val;

    val.f = from;
    copy_i32_vi32(len, val.i, (intv_int32_t*) to);
#else
#error You need libintvec
#endif
}
#endif

/****************************************************************/
/* copy_vf_vf implemented using the corresponding int32 routine */
/****************************************************************/

#ifdef DECL_nfp_copy_vf_vf
void nfp_copy_vf_vf(size_t len, const float* from, float* to);
#endif

#ifdef DEFN_nfp_copy_vf_vf
void
nfp_copy_vf_vf(size_t len, const float* from, float* to)
{
#if defined(FLTV_HAVE_LIBINTVEC)
    copy_vi32_vi32(len, (const intv_int32_t*) from, (intv_int32_t*) to);
#else
#error You need libintvec
#endif
}
#endif


/****************************************************************/
/* neg_vf_vf implemented using scalar integer ops               */
/****************************************************************/

#ifdef DECL_nfp_neg_vf_vf
void nfp_neg_vf_vf(size_t len, const float* from, float* to);
#endif

#ifdef DEFN_nfp_neg_vf_vf
void
nfp_neg_vf_vf(size_t len, const float* from, float* to)
{
    size_t i;
    const fltv_uint32_t *fromi = (const fltv_uint32_t *) from;
    fltv_uint32_t *toi = (fltv_uint32_t *) to;

    for (i=0; i<len; i++)
    {
	*toi++ = (*fromi++) ^ 0x80000000ul;
    }
}
#endif

/****************************************************************/
/* abs_vf_vf implemented using scalar integer ops               */
/****************************************************************/

#ifdef DECL_nfp_abs_vf_vf
void nfp_abs_vf_vf(size_t len, const float* from, float* to);
#endif

#ifdef DEFN_nfp_abs_vf_vf
void
nfp_abs_vf_vf(size_t len, const float* from, float* to)
{
    size_t i;
    const fltv_uint32_t *fromi = (const fltv_uint32_t *) from;
    fltv_uint32_t *toi = (fltv_uint32_t *) to;

    for (i=0; i<len; i++)
    {
	*toi++ = (*fromi++) & 0x7ffffffful;
    }
}
#endif
