/*
** $Header: /u/spert/src/fltvec/testsuite/RCS/cmp_gfgf_gi32gb_test.c,v 1.4 1995/12/21 00:39:03 davidj Exp $
*/

#include <stddef.h>
#include <stdlib.h>

#include "fltvec_test.h"

#ifdef CHECK_TOR

enum { MAXVLR = 32 };

static void
one_test(size_t len, const float* a, const float* b, const fltv_int32_t* diff,
	 const fltv_int32_t* unord)
{
    size_t i;
    fltv_uint32_t unordbits;
    fltv_int32_t dres[MAXVLR];
    fltv_int32_t ures[MAXVLR];
    
    unordbits = tor_vlr_cmp_vfvf_vi32i32(len, a, b, dres, ures);
    for (i=0; i<len; i++)
    {
	if (dres[i]>0)
	    dres[i] = 1;
	else if (dres[i]<0)
	    dres[i] = -1;
	else
	    dres[i] = 0;
    }
    rtst_checkeq_vi32vi32(len, dres, diff);
    rtst_checkeq_vi32vi32(len, ures, unord);
    for (i=0; i<len; i++)
	rtst_assert(((unordbits>>i) & 0x01)==ures[i]);
}

/* Check we handle signs and magnitudes properly */

static void
test1(void)
{
    enum { LEN1=12 };
    const float a1[LEN1] =
    {
	1.0, 2.0, -1.0, -2.0, 1.0, -2.0, -1.0, 2.0,
	1.0, 2.0, -1.0, -2.0
    };
    const float b1[LEN1] =
    {
	2.0, 1.0, -2.0, -1.0, -2.0, 1.0, 2.0, -1.0,
	1.0, 2.0, -1.0, -2.0
    };
    const fltv_int32_t diff1[LEN1] =
    {
	-1, 1, 1, -1, 1, -1, -1, 1,
	0, 0, 0, 0
    };
    const fltv_int32_t unord1[LEN1] =
    {
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0
    };
    one_test(LEN1, a1, b1, diff1, unord1);
}

enum {
    INF=0x7f800000u, NINF=0xff800000u,
    MAX=0x7f7fffffu, NMAX=0xff7fffffu,
    NAN1=0x7f800001u, NAN2=0x7f800002u, NANF=0x7fffffffu,
    NNAN1=0xff800001u, NNAN2=0xff800002u, NNANF=0xffffffffu,
    ZERO=0u, NZERO=0x80000000u
};

/* Check infinities and big numbers */

static void
test2(void)
{
    enum { LEN2=16 };
    const fltv_uint32_t a2[] = 
    {
	INF, INF, INF, INF, NINF, NINF, NINF, NINF,
	MAX, MAX, MAX, MAX, NMAX, NMAX, NMAX, NMAX
    };
    const fltv_uint32_t b2[] =
    {
	INF, NINF, MAX, NMAX, INF, NINF, MAX, NMAX,
	INF, NINF, MAX, NMAX, INF, NINF, MAX, NMAX
    };
    const fltv_int32_t diff2[LEN2] =
    {
	0, 1, 1, 1, -1, 0, -1, -1,
	-1, 1, 0, 1, -1, 1, -1, 0
    };
    const fltv_int32_t unord2[LEN2] =
    {
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0
    };
    one_test(LEN2, (const float *)a2, (const float *) b2, diff2, unord2);
}

/* Check we spot different sorts of NaNs */

static void
test3(void)
{
    enum { LEN3=12 };
    const fltv_uint32_t a3[] = 
    {
	0, 0, 0, 0, 0, 0,
	NAN1, NAN2, NANF, NNAN1, NNAN2, NNANF
    };
    const fltv_uint32_t b3[] =
    {
	NAN1, NAN2, NANF, NNAN1, NNAN2, NNANF,
	0, 0, 0, 0, 0, 0
    };
    const fltv_int32_t diff3[LEN3] =
    {
	-1, -1, -1, 1, 1, 1,
	1, 1, 1, -1, -1, -1
    };
    const fltv_int32_t unord3[LEN3] =
    {
	1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1
    };
    one_test(LEN3, (const float *)a3, (const float *) b3, diff3, unord3);
}


#endif


int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);
#ifdef CHECK_TOR
    rtst_start("tor_cmp_gfgf_gi32gb");

    test1();
    test2();
    test3();
    rtst_passed();
#endif
    rtst_exit();

    return 0;
}
