/*
** $Header: /u/spert/src/fltvec/tor/RCS/cmp_gfgf_gi32gb.S,v 1.7 1995/12/21 00:39:09 davidj Exp $
**
** Torrent floating point millicode comparison routine
*/


#include "tor/tordef.h"
#include "m_regdef.h"

	.extern tor_fltv_flags_invalid32,4

#ifdef DECL_tor_cmp_gfgf_gi32gb 
fltv_uint32_t tor_cmp_gfgf_gi32gb(void); 
#endif

#ifdef DEFN_tor_cmp_gfgf_gi32gb

	.globl tor_cmp_gfgf_gi32gb


/*
-------------------------------------------------------------------------------
Routine:  tor_cmp_gfgf_gf_gi32gb

Input:
	vlr
	va0	vA
	va1	vB

Output:
	vz0	vDiff
	vz1	vUnord
	v0	unord

Not preserved:
	vtt
	va0--va3
	vt0--vt5
	v0--v1
	t4--t9

Function:

The single-precision floating-point vectors vA and vB are element-wise
compared.  The result of the comparison is returend in vDiff, a vector
of signed integers.  If the element in vA is greater than the
corresponding element in vB, the element in vDiff is >0.  If the
element in vB is less than the element in vB, the element in vDiff is
<0.  If the two elements are equal (as specified in the IEEE spec. -
i.e. +0==-0), the element in vDiff is set to 0.  If either of vA or vB
is a NaN (i.e. they are unordered), the result in vDiff is undefined
and a 1 is set in the corresponding elelment of vUnord and the
appropriate bit in unord - otherwise these are both set to zero.  The
appropriate flag in fltv_flags_invalid32 is also raised if either input
is a NaN.  No traps are taken for any exception.
*/


#define vA va0			/* Input A */
#define vB va1			/* Input B */
#define vDiff vz0		/* Output of differences */
#define vUnord vz1		/* Output of unordered flags */

#define vNExpSigA vt0		/* The negative of all but the sign of A */
#define vNExpSigB vt1		/* The negative of all but the sign of B */
#define vSExpSigA vNExpSigA	/* A signed version of the exp and sig of A */
#define vSExpSigB vNExpSigB	/* A signed version of the exp and sig of A */
#define vExpA vt2		/* The exponent of A, with the LSB cleared
				   if the significand if A is zero */
#define vExpB vt3		/* The exponent of B, with the LSB cleared
				   if the significand if B is zero */


#define unord v0		/* A bit vector indicating which comparisons
				   were unordered */
#define temp0 t4
#define temp1 t5
#define temp2 t6
#define temp3 t7
#define temp4 t8

	.ent tor_cmp_gfgf_gi32gb
tor_cmp_gfgf_gi32gb:

/* get -expsig(A) and -expsig(B) */

	li	temp0, 0x7fffffff
	li	temp1, FX_NOCLIP | FX_A_AND_B | FX_NOROUND | FX_SHLZA
	fxsub.sv vNExpSigA, temp0, vA, temp1
	fxsub.sv vNExpSigB, temp0, vB, temp1

/* Either choose -expsig(A) or expsig(A) depending on sign(A)
   (and same for B). This gives us a signed number that indicates the
   ordering of A and B
*/

	cmvgez.vv vSExpSigA, va0, va0 
	cmvgez.vv vSExpSigB, va1, va1

/* find the (saturated) difference between the two values - this is
   +ve if A is bigger, -ve if B is bigger, and 0 if they are the same */

	li	temp1, FX_CLIP32 | FX_NOROUND
	fxsub.vv vDiff, vSExpSigA, vSExpSigB, temp1

/* Use the shifters to clear the sign bit of each input, and leave the
   exponent in the least significant bits.  Use the sticky bit to
   clear the LSB of the exponent if all significand bits are zero.
   The end result is to leave 0xff in vExpA if A is a NaN (same for
   B).
*/

	li	temp2, (1*FX_SHLAM) | FX_SHLNR | FX_SEPSHAM | (24*FX_SHRAM) \
			| FX_SHRL | FX_AU | FX_NOCLIP
	fxadd.vs vExpA, vA, zero, temp2
	fxadd.vs vExpB, vB, zero, temp2

/* If vExpA is 0x000000ff, then put a 1 in vUnord.  If vExpB is
   0x000000ff, conditionally move a 1 into vUnord (effectively performing an
   OR operation)
*/

	li	temp3, 0xff
	seq.vs	vUnord, vExpA, temp3
	li	temp4, FX_NOROUND | FX_EQZ | FX_CLIP_SEQZ
	fxsub.vs vUnord, vExpB, temp3, temp4

/* Update the invalid flags and return a bit vector version of vUnord */

	ctvu	zero, vcond
	flt.sv	zero, vUnord
	cfvu	unord, vcond
	lw	v1, tor_fltv_flags_invalid32
	or	v1, v0
	sw	v1, tor_fltv_flags_invalid32

	j	ra
	.end tor_cmp_gfgf_gi32gb

#endif

/****************************************************************
* A function for debugging the above
****************************************************************/

#ifdef DECL_tor_vlr_cmp_vfvf_vi32i32
fltv_uint32_t
tor_vlr_cmp_vfvf_vi32i32(size_t len, const float* a, const float* b,
			 fltv_int32_t* diff_res, fltv_int32_t* unord_res);
#endif
	
#ifdef DEFN_tor_vlr_cmp_vfvf_vi32i32

#define savedRA t0
	
	.globl	tor_vlr_cmp_vfvf_vi32i32

	.ent	tor_vlr_cmp_vfvf_vi32i32
tor_vlr_cmp_vfvf_vi32i32:
	move	savedRA, ra
	ctvu	a0, vlr
	lw.v	va0, a1
	lw.v	va1, a2
	jal	tor_cmp_gfgf_gi32gb
	sw.v	vz0, a3
	lw	v1, 16(sp)
	sw.v	vz1, v1
	j	savedRA
	.end	tor_vlr_cmp_vfvf_vi32i32

#endif
