/*
** $Header: /u/spert/src/fltvec/tor/RCS/tor_mul_gfgf_gf.S,v 1.6 1995/07/09 23:31:14 davidj Exp $
*/

#include "tor/tordef.h"
#include "m_regdef.h"

	.globl tor_mul_gfgf_gf
#ifdef DECL_tor_mul_gfgf_gf 
void tor_mul_gfgf_gf(void); 
#endif

#define vA va0			/* Input A				*/
#define vB va1			/* Input B				*/
#define vZ vv2			/* Output Z				*/

#define vNeitherIsZero vtt	/* Flag for neither A nor B is zero	*/
#define vZSign vt0		/* Sign of result			*/
#define vNeitherIsSubnormal vt1	/* Flag for neither A nor B is		*/
				/*     subnormal			*/
#define vAbsA vt4		/* A shifted up one bit to remove sign	*/
#define vAbsB vt5		/* B shifted up one bit to remove sign	*/
#define vBIsSubnormal vt1	/* Flag for B is subnormal		*/
#define vAbsA2 vt3		/* Temporary storage			*/
#define vCount16 vt1		/* vCountLeadingZerosA & 0x10		*/
#define vCount8 vt2		/* vCountLeadingZerosA & 0x08		*/
#define vLowIndex va0		/* Index into low zeros-count table	*/
#define vHighIndex va1		/* Index into high zeros-count table	*/
#define vCountLeadingZerosA vt3	/* Number of leading zeros of vAbsA	*/
#define vShiftedA va2		/* vAbsA shifted for count of leading	*/
				/*     zeros				*/
#define vAExpOffset vt3		/* Fix-up to A exponent to account for	*/
				/*     a subnormal input		*/
#define vBIsFinite vt1		/* Flag for B is neither infinite nor a	*/
				/*     NaN				*/
#define vBExp2 vt2		/* Trial fixed-up exponent		*/
#define vAHigh va2		/* Leading 8 bits of A shifted down	*/
#define vBHigh vt1		/* Leading 8 bits of B shifted down	*/
#define vALowBHigh vt2		/* Inner partial product		*/
#define vAHighBLow vt3		/* Inner partial product		*/
#define vALowBLow vt2		/* Lowest partial product		*/
#define vAHighBHigh va2		/* Highest partial product		*/
#define vRawZSignif vt2		/* Raw significand of result		*/
#define vAdjustedAExp va0	/* Adjusted exponent of A		*/
#define vAdjustedBExp va1	/* Adjusted exponent of B		*/
#define vZExp vt1		/* Under-biased exponent of result	*/
#define vExpAdjust vt3		/* Normalization adjustment to result	*/
				/*     exponent				*/
#define vBothAreFinite va3	/* Flag for neither A nor B infinite or	*/
				/*     a NaN				*/
#define vZSignif vv2		/* Rounded result significand		*/
#define vInputsCode vtt		/* Coding of NaN/infinity cases in	*/
				/*     inputs				*/
#define vACode vt4		/* Coding of NaN/infinity cases for A	*/
#define vBCode vt5		/* Coding of NaN/infinity cases for B	*/

#define temp0 v0
#define temp1 v1
#define temp2 t4
#define config0 t5
#define config1 t6
#define config2 t7
#define config3 t8
#define config4 t9

	.ent tor_mul_gfgf_gf

/*
-------------------------------------------------------------------------------
Routine:  tor_mul_gfgf_gf

Input:
	vlr
	va0	vA
	va1	vB

Output:
	vv2	vZ

Not preserved:
	vtt
	vv3
	va0--va3
	vt0--vt5
	v0--v1
	t4--t9

Function:

The single-precision floating-point vectors vA and vB are element-wise
multiplied and the single-precision result vector returned in vZ.  The vlr
register controls the elements of vZ actually updated.  This operation conforms
to the IEEE Standard for Binary Floating-point Arithmetic, except as follows:
(1) Only nearest/even rounding is supported; (2) No inexact flags are raised;
and (3) No signalling NaNs are recognized.  All NaNs are quiet.  For all cases
in which the result of the multiplication is a NaN, the default MIPS NaN is
returned.  The bit patterns of input NaNs are not preserved.

Exceptions for any of the element-wise multiplications are signalled by
raising element-wise flags in the global variables fltv_flags_invalid32,
fltv_flags_overflow32, and fltv_flags_underflow32.  No traps are taken for any
exceptions.  Underflow is never signalled for results that are not subnormal.

-------------------------------------------------------------------------------
*/

tor_mul_gfgf_gf:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vNeitherIsZero
|	va0	vA
|	va1	vB
|	vt0	vZSign
|	vt1	vNeitherIsSubnormal
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Find the sign of the result from the exclusive-or of the signs of the inputs.
Shift up A and B each by one bit to eliminate their signs.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	addu.vv vAbsA,vA,vA
	addu.vv vAbsB,vB,vB
	li config1,(FX_SHLZA|FX_A_XOR_B|FX_CLIP_SLTZ)
	fxadd.vv vZSign,vA,vB,config1

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set vNeitherIsZero to a nonzero value if and only if neither A nor B is zero.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	snez.vv vNeitherIsZero,vAbsA
	cmveqz.vv vNeitherIsZero,vAbsB,vzero

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Determine if A or B is a subnormal with neither being zero.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp1,(1<<24)
	sltu.sv vNeitherIsSubnormal,temp1,vAbsA
	li config2,(FX_BU|FX_CLIP_SLTZ|FX_GTZ)
	fxsub.sv vNeitherIsSubnormal,temp1,vAbsB,config2
	ctvu zero,vcond
	flt.vv vNeitherIsSubnormal,vNeitherIsZero
	cfvu temp2,vcond

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set-up a constant and a pipeline configuration for the subsequent significand
multiplication.  (This occurs here only for instruction scheduling purposes.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,0x00FF0000
	li config0,(FX_SHLZA|(25*FX_SHLAM)|FX_A_AND_B\
	            |FX_SEPSHAM|(16*FX_SHRAM)|FX_SHRNR|FX_CLIP_PASS)

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If A or B is subnormal and neither is zero, a subnormal fix-up is needed.  If
a subnormal fix-up is not needed for any vector element, skip the code to
perform the fix-up.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	beqz temp2,inputsNormalized

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vNeitherIsZero
|	vt0	vZSign
|	vt1	vBIsSubnormal
|	vt3	vAbsA2
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If B is subnormal, swap vAbsA and vAbsB.  (The purpose of the swap is to ensure
that a subnormal input is in vAbsA, since only vAbsA will be fixed up.)
    (It is never necessary to fix up more than one subnormal input before
multiplying.  If both inputs are subnormal, their product would correctly
underflow to zero even if both inputs were mis-interpreted as normal.  It would
be possible in fact to avoid _any_ fix-up when both inputs are subnormal.
Rather than test for this case, if both inputs are subnormal the following code
harmlessly performs the fix-up for one but not the other.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,(1<<24)
	sltu.vs vBIsSubnormal,vAbsB,temp0
	mov.vv vAbsA2,vAbsA
	cmvnez.vv vAbsA,vBIsSubnormal,vAbsB
	cmvnez.vv vAbsB,vBIsSubnormal,vAbsA2

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vNeitherIsZero
|	va0	vLowIndex
|	va1	vHighIndex
|	va2	vShiftedA
|	vt0	vZSign
|	vt1	vCount16
|	vt2	vCount8
|	vt3	vCountLeadingZerosA
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set vCountLeadingZerosALeadingZerosA to the number of leading zeros of vAbsA.
    (NOTE:  This block of code could all be replaced by a single count-leading-
zeros instruction.  See countlz_gi32_gi32.S for more refined commentary.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,(1<<16)
	sltu.vs vCount16,vAbsA,temp0
	li temp1,4
	sllv.vs vCount16,vCount16,temp1
	sllv.vv vShiftedA,vAbsA,vCount16
	li temp2,(1<<24)
	sltu.vs vCount8,vShiftedA,temp2
	li temp0,3
	sllv.vs vCount8,vCount8,temp0
	sllv.vv vShiftedA,vShiftedA,vCount8
	addu.vv vCount8,vCount16,vCount8
	li config0,(FX_SHLZA|FX_A_AND_B|(23*FX_SHLAM)|FX_NOROUND|FX_CLIP_PASS)
	li config1,(FX_SHLV|(28*FX_SHLAM)|FX_NOROUND|FX_CLIP_PASS|FX_GEZ)
	li temp1,(7<<25)
	fxadd.vs vLowIndex,vShiftedA,temp1,config0
	li temp2,0x76554444
	fxadd.sv vCountLeadingZerosA,temp2,vLowIndex,config1
	li temp0,(7<<28)
	addu config2,config0,(3*FX_SHLAM)
	fxadd.vs vHighIndex,vShiftedA,temp0,config2
	li temp1,0xF3221111
	fxadd.sv vCountLeadingZerosA,temp1,vHighIndex,config1
	cmvltz.vv vCountLeadingZerosA,vShiftedA,vzero
	addu.vv vCountLeadingZerosA,vCount8,vCountLeadingZerosA

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vNeitherIsZero
|	vt0	vZSign
|	vt3	vCountLeadingZerosA,vAExpOffset
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vAbsA is subnormal, normalize it by shifting up.  Set vAExpOffset to the
number of bits shifted (zero if normal, infinity, or NaN).
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
        li temp2,7
        subu.vs vAExpOffset,vCountLeadingZerosA,temp2
        cmvltz.vv vAExpOffset,vAExpOffset,vzero
        sllv.vv vAbsA,vAbsA,vAExpOffset

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vNeitherIsZero
|	vt0	vZSign
|	vt1	vBIsFinite
|	vt2	vBExp2
|	vt3	vAExpOffset
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vAbsB is finite and not too small, subtract vAExpOffset from the exponent of
vAbsB, so that the product of the normalized vAbsA and the modified vAbsB will
be the same as that of the original inputs.  If vAbsB is a NaN or infinity, or
if the exponent of vAbsB is not large enough to subtract from, leave it alone.
(If vAbsB is too small, multiplying by the normalized vAbsA is bound to
underflow to zero anyway; the exponent adjustment does not matter.  And of
course, if vAbsB is a NaN or infinity, a subnormal vAbsA is inconsequential.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp1,24
	srlv.vs vBExp2,vAbsB,temp1
	li temp2,0xFF
	sltu.vs vBIsFinite,vBExp2,temp2
	subu.vv vBExp2,vBExp2,vAExpOffset
	sllv.vs vAExpOffset,vAExpOffset,temp1
	cmveqz.vv vAExpOffset,vBIsFinite,vzero
	cmvlez.vv vAExpOffset,vBExp2,vzero
	subu.vv vAbsB,vAbsB,vAExpOffset

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set-up a constant and a pipeline configuration for the subsequent significand
multiplication.  (This occurs here only for instruction scheduling purposes.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,0x00FF0000
	li config0,(FX_SHLZA|(25*FX_SHLAM)|FX_A_AND_B\
	            |FX_SEPSHAM|(16*FX_SHRAM)|FX_SHRNR|FX_CLIP_PASS)

inputsNormalized:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vNeitherIsZero
|	va2	vAHigh,vAHighBHigh
|	vt0	vZSign
|	vt1	vBHigh
|	vt2	vALowBHigh,vALowBLow,vRawZSignif
|	vt3	vAHighBLow
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
For both inputs, separate out the high 8 bits of the significand from the
lower 16 bits, and make the leading significand bit explicit.
    Zero, infinity, and NaN inputs are (mis)treated as though they were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	fxadd.vs vAHigh,vAbsA,temp0,config0
	fxadd.vs vBHigh,vAbsB,temp0,config0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Multiply vAbsA by vAbsB and round to odd (jamming) to form vRawZSignif.
    Zero, infinity, and NaN inputs continue to be (mis)treated as though they
were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config1,(FX_AU|FX_BU|FX_SHRL|FX_CLIP_PASS)
	fxmul.vv vAHighBLow,vAHigh,vAbsB,config1
	fxmul.vv vALowBHigh,vAbsA,vBHigh,config1
	addu.vv vAHighBLow,vAHighBLow,vALowBHigh
	addu config2,config1,((16*FX_SHLAM)|FX_RODD)
	fxmul.vv vALowBLow,vAbsA,vAbsB,config2
	li config3,(FX_SEPSHAM|(2*FX_SHRAM)|FX_RODD)
	fxadd.vv vRawZSignif,vAHighBLow,vALowBLow,config3
	fxmul.vv vAHighBHigh,vAHigh,vBHigh,config1
	li config0,((14*FX_SHLAM)|FX_AU|FX_SEPSHAM|FX_NOROUND|FX_CLIP_PASS)
	fxadd.vv vRawZSignif,vAHighBHigh,vRawZSignif,config0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vNeitherIsZero
|	va0	vAdjustedAExp
|	va1	vAdjustedBExp
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vt3	vExpAdjust
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Calcuate the "under-biased" exponent of the result, by summing the exponents of
vA and vB and also taking into account whether vRawZSignif needs to be shifted
up by one bit to be normalized.  (The exponent is "under-biased" because the
integral portion of the significand will be added to it when the result is
packed.  In most cases, a 1 will be added to the exponent during packing.
However, if a carry occurred during rounding of the significand, a value of 2
will be added to the exponent during packing.)
    Zero, infinity, and NaN inputs continue to be (mis)treated as though they
were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config1,(FX_SHLZA|FX_A_AND_B|FX_BU\
	            |FX_SHRL|FX_SEPSHAM|(7*FX_SHRAM)|FX_NOROUND|FX_CLIP_PASS)
	li temp1,0x80000000
	fxadd.vs vExpAdjust,vRawZSignif,temp1,config1
	li config2,(FX_AU|FX_SEPSHAM|(24*FX_SHRAM)|FX_NOROUND|FX_CLIP_PASS)
	fxadd.vv vAdjustedAExp,vAbsA,vExpAdjust,config2
	fxadd.vs vAdjustedBExp,vAbsB,temp1,config2
	addu.vv vZExp,vAdjustedAExp,vAdjustedBExp

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If either input is zero, force the result exponent to zero.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	cmveqz.vv vZExp,vNeitherIsZero,vzero

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vNeitherIsZero
|	vv2	vZSignif
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Normalize vRawZSignif.  The significand product will be close to normalized
anyway, but it can be off by one bit.  If either input is zero, force
vRawZSignif to zero.
    Infinity and NaN inputs continue to be (mis)treated as though they were
normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config0,(FX_AU|FX_BU|FX_CLIP_PASS|FX_GEZ)
	fxadd.vv vRawZSignif,vRawZSignif,vRawZSignif,config0
	cmveqz.vv vRawZSignif,vNeitherIsZero,vzero

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Form the tentative result significand by shifting the raw result significand
down, rounding to nearest/even.
    Infinity and NaN inputs continue to be (mis)treated as though they were
normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config1,(FX_BU|FX_SHRL|(8*FX_SHLAM)|FX_RNEARESTEVEN|FX_CLIP_PASS)
	fxadd.vv vZSignif,vzero,vRawZSignif,config1

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vv2	vZSignif
|	va3	vBothAreFinite
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vZExp is less than 0, the result is a subnormal or has underflowed.  If
vZExp is 0xFD or more, the result might have overflowed.  If any vector element
may have underflowed or overflowed, or if any input was a NaN or infinity, jump
to a special case handler.
    (Although the result has not actually overflowed in most cases in which
vZExp equals 0xFD, overflow does occur when vZExp is 0xFD and vRawZSignif >=
0xFFFFFF80.)
    (An exact zero result can occur only if one of the inputs was zero, in
which case vZExp was forced to zero above.  In all other cases, if vZExp is
zero, the biased exponent of the result will be at least 1.)
    (Note that temp2 is still live at the entrance to the special cases code.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp2,(0xFF<<24)
	sltu.vs vBothAreFinite,vAbsA,temp2
	li config3,(FX_AU|FX_BU|FX_CLIP_SLTZ|FX_GEZ)
	fxsub.vs vBothAreFinite,vAbsB,temp2,config3
	feqz.v vBothAreFinite
	cfvu temp2,vcond
	li temp0,0xFC
	fltu.sv temp0,vZExp
	cfvu temp0,vcond
	or temp0,temp2
	bnez temp0,specialCases

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vv2	vZSignif,vZ
|	vt0	vZSign
|	vt1	vZExp
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Add the integral portion of the significand to the under-biased exponent, and
pack the sign, biased exponent, and significand of the result.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config0,((23*FX_SHLAM)|FX_SEPSHAM|FX_NOROUND)
	fxadd.vv vZ,vZExp,vZSignif,config0
	li config1,((31*FX_SHLAM)|FX_SEPSHAM|FX_NOROUND)
	fxadd.vv vZ,vZSign,vZ,config1

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Return.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	j ra

specialCases:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vv2	vZSignif
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If both inputs are finite, the only possible special cases involve overflow or
underflow of the result.  In that case, skip the code to handle NaN and
infinity inputs.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	beqz temp2,resultSpecialCasesOnly

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vInputsCode
|	vv2	vZSignif
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vt4	vAbsA,vACode
|	vt5	vAbsB,vBCode
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Distill the types of the original inputs into a single value in vInputsCode.
Set vInputsCode to between 3 and 6 inclusive if either input is a NaN; else set
vInputsCode to 0 or 1 if either input is an infinity; else set vInputsCode to
exactly 2.  (This distillation is done by first setting vACode to either 1, 0,
or 3, according to whether the original A input is normal, an infinity, or a
NaN, respectively.  The same is done with vBCode, and then vACode and vBCode
are added to form vInputsCode.  This value is used by the special case handler
called below.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config0,(FX_AU|FX_BU|FX_SHRL|FX_SEPSHAM|(31*FX_SHRAM)|FX_RODD)
	li temp0,(0xFF<<24)
	fxsub.sv vACode,temp0,vAbsA,config0
	fxsub.sv vBCode,temp0,vAbsB,config0
	addu.vv vInputsCode,vACode,vBCode

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vInputsCode
|	vv2	vZSignif
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Continue with the special case input and result handler common to both single-
precision floating-point addition and multiplication.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	j tor_fltv_singlespec

resultSpecialCasesOnly:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vv2	vZSignif
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Continue with the special case result handler common to all single-precision
floating-point operations.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	j tor_fltv_singlerange

	.end tor_mul_gfgf_gf

