//
// FloatVecOps.C
//
// Overloading of arithmetic ops for floatVecs and floatRefs
//
// dpwe@media.mit.edu 1995oct01
//

//#include "FloatVecOps.H"	// basic floatVec and floatRef definitions
#include "nSurface.H"	// pulls in FloatVecOps.H too
#include <math.h>
#include <assert.h>

floatVec operator + (const floatRef& a, const floatRef& b)
{
    floatVec f(a); f += b; return f;
}

floatVec operator + (const floatRef& a, double d)
{
    floatVec f(a); f += d; return f;
}

floatVec operator + (double d, const floatRef& a)
{
    floatVec f(a); f += d; return f;
}

floatVec operator - (const floatRef& a, const floatRef& b)
{
    floatVec f(a); f -= b; return f;
}

floatVec operator - (const floatRef& a, double d)
{
    floatVec f(a); f -= d; return f;
}

floatVec operator - (double d, const floatRef& a)
{
    floatVec f(a); f -= d; f.Scale(-1); return f;
}

floatVec operator * (const floatRef& a, const floatRef& b)
{
    floatVec f(a); f *= b; return f;
}

floatVec operator * (const floatRef& a, double d)
{
    floatVec f(a); f *= d; return f;
}

floatVec operator * (double d, const floatRef& a)
{
    floatVec f(a); f *= d; return f;
}

floatVec operator / (const floatRef& a, const floatRef& b)
{
    floatVec f(a); f /= b; return f;
}

floatVec operator / (const floatRef& a, double d)
{
    floatVec f(a); f /= d; return f;
}

// Also provide divide operator with args in other order since it's not 
// commutative

floatVec operator / (double d, const floatRef& a)
{	// have to code this one explicitly
    floatVec f(a); 
    int i;
    for(i=0; i<f.Size(); ++i)  {
	f[i] = d/a.Val(i);
    }
    return f;
}

floatVec operator ^ (const floatRef& a, const floatRef& b)
{
    floatVec f(a); f ^= b; return f;
}

floatVec operator ^ (const floatRef& a, double d)
{
    floatVec f(a); f ^= d; return f;
}

floatVec operator ^ (double d, const floatRef& a)
{
    floatVec f(a.Size()); f.Fill(d); f ^= a; return f;
}

floatVec Eq(const floatRef& a, const floatRef& b)
{
    floatVec f(a); f.Eq(b); return f;
}

floatVec Eq(const floatRef& a, double d)
{
    floatVec f(a); f.Eq(d); return f;
}

floatVec Eq(double d, const floatRef& a)
{
    floatVec f(a); f.Eq(d); return f;
}

floatVec Lt(const floatRef& a, const floatRef& b)
{
    floatVec f(a.Size());
    f.Fill(0);
    int i;
    for(i=0; i<f.Size(); ++i)  {
	if(a.Val(i) < b.Val(i))	
	    f[i] = 1;
    }
    return f;
}

floatVec Lt(const floatRef& a, double d)
{
    floatVec f(a.Size());
    f.Fill(0);
    int i;
    for(i=0; i<f.Size(); ++i)  {
	if(a.Val(i) < d)	
	    f[i] = 1;
    }
    return f;
}

floatVec Lt(double d, const floatRef& b)
{
    floatVec f(b.Size());
    f.Fill(0);
    int i;
    for(i=0; i<f.Size(); ++i)  {
	if(d < b.Val(i))	
	    f[i] = 1;
    }
    return f;
}

floatVec Gt(const floatRef& a, const floatRef& b)
{
    floatVec f(a.Size());
    f.Fill(0);
    int i;
    for(i=0; i<f.Size(); ++i)  {
	if(a.Val(i) > b.Val(i))	
	    f[i] = 1;
    }
    return f;
}

floatVec Gt(const floatRef& a, double d)
{
    floatVec f(a.Size());
    f.Fill(0);
    int i;
    for(i=0; i<f.Size(); ++i)  {
	if(a.Val(i) > d)	
	    f[i] = 1;
    }
    return f;
}

floatVec Gt(double d, const floatRef& b)
{
    floatVec f(b.Size());
    f.Fill(0);
    int i;
    for(i=0; i<f.Size(); ++i)  {
	if(d > b.Val(i))	
	    f[i] = 1;
    }
    return f;
}

floatVec Floor(const floatRef& a, const floatRef& b)
{
    floatVec f(a); 
    int i; 
    for(i=0; i<f.Size(); ++i) {
	f[i] = max(a.Val(i), b.Val(i));
    }
    return f;
}

floatVec Not(const floatRef& a)
{
    floatVec f(a);
    int i;
    for(i=0; i<f.Size(); ++i) {
	if(f[i])	f[i] = 0;
	else		f[i] = 1;
    }
    return f;
}

floatVec operator == (const floatRef& a, const floatRef& b)
    { return Eq(a, b); }
floatVec operator == (const floatRef& a, double d)
    { return Eq(a, d); }
floatVec operator == (double d, const floatRef& b)
    { return Eq(d, b); }

floatVec operator != (const floatRef& a, const floatRef& b)
    { return Not(Eq(a, b)); }
floatVec operator != (const floatRef& a, double d)
    { return Not(Eq(a, d)); }
floatVec operator != (double d, const floatRef& b)
    { return Not(Eq(d, b)); }

floatVec operator < (const floatRef& a, const floatRef& b)
   { return Lt(a, b); }
floatVec operator < (const floatRef& a, double d)
    { return Lt(a, d); }
floatVec operator < (double d, const floatRef& b)
    { return Lt(d, b); }

floatVec operator > (const floatRef& a, const floatRef& b)
    { return Gt(a, b); }
floatVec operator > (const floatRef& a, double d)
    { return Gt(a, d); }
floatVec operator > (double d, const floatRef& b)
    { return Gt(d, b); }

floatVec operator >= (const floatRef& a, const floatRef& b)
    { return Gt(a, b) + Eq(a, b); }
floatVec operator >= (const floatRef& a, double d)
    { return Gt(a, d) + Eq(a, d); }
floatVec operator >= (double d, const floatRef& b)
    { return Gt(d, b) + Eq(d, b);; }

floatVec operator <= (const floatRef& a, const floatRef& b)
    { return Lt(a, b) + Eq(a, b); }
floatVec operator <= (const floatRef& a, double d)
    { return Lt(a, d) + Eq(a, d); }
floatVec operator <= (double d, const floatRef& b)
    { return Lt(d, b) + Eq(d, b);; }

floatVec operator || (const floatRef& a, const floatRef& b)
    { return ((a != 0)+(b != 0)) != 0; }
floatVec operator && (const floatRef& a, const floatRef& b)
    { return (a != 0) * (b != 0); }

floatVec Floor(const floatRef& a, double d)
{
    floatVec f(a); f.Floor(d); return f;
}

floatVec Floor(double d, const floatRef& a)
{
    floatVec f(a); f.Floor(d); return f;
}

// Max is the same as floor - return the largest value from each pair
floatVec Max(const floatRef& a, const floatRef& b)
    { return Floor(a,b); }
floatVec Max(const floatRef& a, double d)
    { return Floor(a,d); }
floatVec Max(double d, const floatRef& a)
    { return Floor(d,a); }

floatVec Ceil(const floatRef& a, const floatRef& b)
{
    floatVec f(a); 
    int i; 
    for(i=0; i<f.Size(); ++i) {
	f[i] = min(a.Val(i), b.Val(i));
    }
    return f;
}

floatVec Ceil(const floatRef& a, double d)
{
    floatVec f(a); f.Ceil(d); return f;
}

floatVec Ceil(double d, const floatRef& a)
{
    floatVec f(a); f.Ceil(d); return f;
}

// Min is the same as Ceil - return the smaller of each pair
floatVec Min(const floatRef& a, const floatRef& b)
    { return Ceil(a,b); }
floatVec Min(const floatRef& a, double d)
    { return Ceil(a,d); }
floatVec Min(double d, const floatRef& a)
    { return Ceil(d,a); }

floatVec Abs(const floatRef& a)
{
    floatVec f(a); f.FWRectify(); return f;
}
floatVec Exp(const floatRef& a)
{
    floatVec f(a); f.Exp(); return f;
}
floatVec Log(const floatRef& a)
{
    floatVec f(a); f.Log(); return f;
}

floatVec HWRectify(floatRef& fr)
{
    floatVec rslt = fr;
    rslt.HWRectify();
    return rslt;
}

floatVec Sqrt(floatRef& fr)
{
    floatVec rslt = fr;
    rslt.Sqrt();
    return rslt;
}

floatVec SafeDiv(const floatRef& a, const floatRef& b, const floatRef& c)
{   /* return a/b, except where b==0, return c */
    floatVec bzero = (b==0);
    return (bzero*c) + ((1-bzero)*a)/(b + bzero);
}

floatVec SafeDiv(const floatRef& a, const floatRef& b, double d)
{   /* return a/b, except where b==0, return d */
    floatVec bzero = (b==0);
    return (bzero*d) + ((1-bzero)*a)/(b + bzero);
}

floatVec Hypot(const floatRef& a, const floatRef& b)
{
    floatVec f = (a*a) + (b*b); f.Sqrt(); return f;
}

floatVec Hypot(const floatRef& a, double d)
{
    floatVec f = (a*a) + (d*d); f.Sqrt(); return f;
}

floatVec Hypot(double d, const floatRef& a)
{
    return Hypot(a, d);
}


floatVec Pow(const floatRef& a, const floatRef& b)
{
    floatVec f = a; f ^= b; return f;
}

floatVec Pow(const floatRef& a, double d)
{
    floatVec f = a; f ^= d; return f;
}

floatVec Pow(double d, const floatRef& a)
{
    floatVec f(a.Size()); f.Fill(d); f ^= a; return f;
}


#ifdef FVOMAIN
#define ___quote(ARG)      #ARG
#define ___QUOTE(ARG)	___quote(ARG)
#define TEST(num,clause,elseclause)					\
          if(!(clause))	{ cerr << "FAIL Test " << num << ": "	        \
			       << ___QUOTE(clause) << endl;  elseclause ; }


int fEq(float a, float b)
{   // loose version of operator==(float,float) - allow up to 0.001 diff
    if(fabs(a-b)<0.001)	return 1;
    else		return 0;
}

void testFloatVecOps(void)
{
    floatVec fv1(5);
    floatVec fv2(5);
    fv1.Fill(1);
    fv2.Fill(3);
    fv1[1] = 0;
    fv2[3] = 10;
    floatVec fv3 = fv1+fv2;
    fv1.Clear();
    fv2.Clear();
    TEST("1.1", fv3[0]==4 && fv3[1]==3 && fv3[2]==4 && fv3[3]==11 \
	 && fv3[4]==4, cout << fv3);
    fv3 = fv3 + 2;
    TEST("1.2", fv3[0]==6 && fv3[1]==5 && fv3[2]==6 && fv3[3]==13 \
	 && fv3[4]==6, cout << fv3);
    fv3 *= 3;
    TEST("1.3", fv3[0]==18 && fv3[1]==15 && fv3[2]==18 && fv3[3]==39 \
	 && fv3[4]==18, cout << fv3);

    floatVec fv4 = 2/fv3;
    fv3.Clear();
    TEST("1.4", fEq(fv4[0],0.1111) && fEq(fv4[1],0.13333) \
	 && fEq(fv4[3],0.05128), cout << fv4);

    // Test ^= etc
    fv1[0] = 0; fv1[1] = 1; fv1[2] = 2; fv1[3] = 3; fv1[4] = 4;
    fv2[0] = 2; fv2[1] = 1; fv2[2] = 2; fv2[3] = 1; fv2[4] = 2;
    fv3 = fv1 ^ fv2;
    TEST("1.5", fv3[0]==0 && fv3[1]==1 && fv3[2]==4 && fv3[3]==3 &&fv3[4]==16,\
	 cout << fv3);
    fv3 ^= 0.5;
    TEST("1.6", fEq(fv3[0],0) && fEq(fv3[1],1) && fEq(fv3[2],2) \
	 && fEq(fv3[3], sqrt(3)) && fEq(fv3[4],4), cout << fv3);

    // SafeDiv
    fv1.Resize(4);
    fv1.Fill(1);
    fv2 = fv1;
    fv2.Fill(2);
    fv3 = SafeDiv(fv1, fv2, -1);
    TEST("1.7", fv3[0] == 0.5 && fv3[1] == 0.5 && fv3[2] == 0.5 \
	 && fv3[3] == 0.5, cout << fv3);
    fv2[1] = 0;
    fv2[3] = 0;
    fv3 = SafeDiv(fv1, fv2, -1);
    TEST("1.7", fEq(fv3[0],0.5) && fEq(fv3[1],-1) && fEq(fv3[2],0.5) \
	 && fEq(fv3[3],-1), cout << fv3);
    fv3.Scale(2);
    fv3[3] = 99;
    fv3 = SafeDiv(fv1, fv2, fv3);
    TEST("1.8", fEq(fv3[0],0.5) && fEq(fv3[1],-2) && fEq(fv3[2],0.5) \
	 && fEq(fv3[3],99), cout << fv3);
    fv2[1] = 1;
    fv2[3] = 10;
    fv3 = SafeDiv(fv1, fv2, fv3);
    TEST("1.9", fEq(fv3[0],0.5) && fEq(fv3[1],1) && fEq(fv3[2],0.5) \
	 && fEq(fv3[3],0.1), cout << fv3);

    fv3.Print(cout, "Test1done");
}

void main(int argc, char **argv)
{
    testFloatVecOps();
}

#endif /* FVOMAIN */
