//
// FloatVecOps.H
//
// Overloading of arithmetic ops for floatVecs and floatRefs
//
// dpwe@media.mit.edu 1995oct01
//

#ifndef _FLOATVECOPS_H_
#define _FLOATVECOPS_H_

//#include "nSurface.H"	// basic floatVec and floatRef definitions
// Now this file is included in that one, not the other way round
class floatRef;		// forward refs, in case
class floatVec;

floatVec operator + (const floatRef& a, const floatRef& b);
floatVec operator + (const floatRef& a, double d);
floatVec operator + (double d, const floatRef& a);

floatVec operator - (const floatRef& a, const floatRef& b);
floatVec operator - (const floatRef& a, double d);
floatVec operator - (double d, const floatRef& a);

floatVec operator * (const floatRef& a, const floatRef& b);
floatVec operator * (const floatRef& a, double d);
floatVec operator * (double d, const floatRef& a);

floatVec operator / (const floatRef& a, const floatRef& b);
floatVec operator / (const floatRef& a, double d);
floatVec operator / (double d, const floatRef& a);

floatVec operator ^ (const floatRef& a, const floatRef& b);
floatVec operator ^ (const floatRef& a, double d);
floatVec operator ^ (double d, const floatRef& a);

// Eq returns a vector of equality scores for each element
floatVec Eq(const floatRef& a, const floatRef& b);
floatVec Eq(const floatRef& a, double d);
floatVec Eq(double d, const floatRef& b);

// Lt returns a vector of less-than scores for each element
floatVec Lt(const floatRef& a, const floatRef& b);
floatVec Lt(const floatRef& a, double d);
floatVec Lt(double d, const floatRef& b);

// Gt returns a vector of greater-than scores for each element
floatVec Gt(const floatRef& a, const floatRef& b);
floatVec Gt(const floatRef& a, double d);
floatVec Gt(double d, const floatRef& b);

floatVec operator == (const floatRef& a, const floatRef& b);
//    { return Eq(a, b); }
floatVec operator == (const floatRef& a, double d);
//    { return Eq(a, d); }
floatVec operator == (double d, const floatRef& b);
//    { return Eq(d, b); }

floatVec operator != (const floatRef& a, const floatRef& b);
//    { return Eq(a, b); }
floatVec operator != (const floatRef& a, double d);
//    { return Eq(a, d); }
floatVec operator != (double d, const floatRef& b);
//    { return Eq(d, b); }

floatVec Not(const floatRef& a);

floatVec operator < (const floatRef& a, const floatRef& b);
//   { return Lt(a, b); }
floatVec operator < (const floatRef& a, double d);
//    { return Lt(a, d); }
floatVec operator < (double d, const floatRef& b);
//    { return Lt(d, b); }

floatVec operator > (const floatRef& a, const floatRef& b);
//    { return Gt(a, b); }
floatVec operator > (const floatRef& a, double d);
//    { return Gt(a, d); }
floatVec operator > (double d, const floatRef& b);
//    { return Gt(d, b); }

floatVec operator >= (const floatRef& a, const floatRef& b);
floatVec operator >= (const floatRef& a, double d);
floatVec operator >= (double d, const floatRef& b);

floatVec operator <= (const floatRef& a, const floatRef& b);
floatVec operator <= (const floatRef& a, double d);
floatVec operator <= (double d, const floatRef& b);

floatVec operator || (const floatRef& a, const floatRef& b);
floatVec operator && (const floatRef& a, const floatRef& b);

// Floor generates a new vec which is the max of the two
floatVec Floor(const floatRef& a, const floatRef& b);
floatVec Floor(const floatRef& a, double d);
floatVec Floor(double d, const floatRef& a);
// Max is the same as floor - return the largest value from each pair
/* inline */ 
floatVec Max(const floatRef& a, const floatRef& b)   ;
//    { return Floor(a,b); }
/* inline */ 
floatVec Max(const floatRef& a, double d)   ;
//    { return Floor(a,d); }
/* inline */ 
floatVec Max(double d, const floatRef& a)   ;
//    { return Floor(d,a); }

// Ceil generates a new vec which is the min of the two
floatVec Ceil(const floatRef& a, const floatRef& b);
floatVec Ceil(const floatRef& a, double d);
floatVec Ceil(double d, const floatRef& a);
// Min is the same as Ceil - return the smaller of each pair
/* inline */ 
floatVec Min(const floatRef& a, const floatRef& b)    ;
//    { return Ceil(a,b); }
/* inline */ 
floatVec Min(const floatRef& a, double d)   ;
//    { return Ceil(a,d); }
/* inline */ 
floatVec Min(double d, const floatRef& a)   ;
//    { return Ceil(d,a); }

// Abs rectifies
floatVec Abs(const floatRef& a);
// Exp generates a new floatVec as the Exp, and Log does the reverse
floatVec Exp(const floatRef& a);
floatVec Log(const floatRef& a);

floatVec HWRectify(floatRef& fr);

floatVec Sqrt(floatRef& fr);

floatVec SafeDiv(const floatRef& a, const floatRef& b, const floatRef& c);
floatVec SafeDiv(const floatRef& a, const floatRef& b, double d);

// Hypot sums things in the squared domain
floatVec Hypot(const floatRef& a, const floatRef& b);
floatVec Hypot(const floatRef& a, double d);
floatVec Hypot(double d, const floatRef& a);

// Pow raises one number to another's power
floatVec Pow(const floatRef& a, const floatRef& b);
floatVec Pow(const floatRef& a, double d);
floatVec Pow(double d, const floatRef& a);


#endif /* _FLOATVECOPS_H_ */
