/*
 * farrayInit.C
 * 
 * Init file for floatArray parts of surfex library
 * extensions to Tcl/Tk for the surfex blackboard sound understanding system
 *
 * 1997jan16 dpwe@icsi.berkeley.edu after tkutilsdpweInit.cc
 * $Header: /u/drspeech/src/guitools/farray_otcl/RCS/farrayInit.C,v 1.4 1998/12/06 02:42:38 dpwe Exp dpwe $
 */

#include <tcl.h>

extern "C" {
int Tcl_breadfvecCmd(ClientData clientData, Tcl_Interp *interp, 
		  int argc, char **argv);
int Tcl_bwritefvecCmd(ClientData clientData, Tcl_Interp *interp, 
		  int argc, char **argv);
}

#include "floatArray_cdl.H"

/* Initialization commands */
extern "C" int Farray_Init(Tcl_Interp *interp)
{   /* Package opening call */
    
    // for Tcl7.5+, signal the package mechanism
    Tcl_PkgProvide(interp, "Farray", "1.0");

    /* initializations ideally all performed by Otcl static initializers */
#ifdef SHLIB_NO_STATIC_INIT
    _floatRef_Init(interp);
    _floatVec_Init(interp);
    _arrayRef_Init(interp);
    _floatArray_Init(interp);
    _probRef_Init(interp);
    _probVec_Init(interp);
    _probArray_Init(interp);
    _Patch_Init(interp);
    _fPatch_Init(interp);
    /* no longer any global fns since DAT dropped */
    /* _G_Init(interp); */
#endif /* SHLIB_NO_STATIC_INIT */

    /* except these ones */
    Tcl_CreateCommand(interp, "breadfvec", Tcl_breadfvecCmd, 
		      (ClientData) NULL, (void (*)(ClientData)) NULL);
    Tcl_CreateCommand(interp, "bwritefvec", Tcl_bwritefvecCmd, 
		      (ClientData) NULL, (void (*)(ClientData)) NULL);

/*     return TCL_OK; */

    /* Ensure that Otcl is re-kicked to recognize the new cmds */
    return Tcl_Eval(interp, "otcl clear");
}

extern "C" int Farray_SafeInit(Tcl_Interp *interp)
{   /* Package opening call */
/*    fprintf(stderr, "Farray: cannot load under safe interpreter\n");
    return TCL_ERROR; */

    return Farray_Init(interp);
}

#include <Otcl.H>
#include "floatArray.H"
#include "floatArrayTcl.H"

/* function to access farray pointer from Tcl name */
floatArray *FAfromName(char *name) {
    /* Convert a Tcl name into an FA pointer by accessing the Otcl table */
    floatArray *faPtr;
    faPtr = (floatArray*)Otcl::otclPtr->obrefToCpp(name,"floatArray");
    return faPtr;
}

arrayRef *ARfromName(char *name) {
    /* Convert a Tcl name into an AR pointer by accessing the Otcl table */
    arrayRef *faPtr;
    faPtr = (arrayRef*)Otcl::otclPtr->obrefToCpp(name,"arrayRef");
    return faPtr;
}

floatVec *FVfromName(char *name) {
    /* Convert a Tcl name into an FV pointer by accessing the Otcl table */
    floatVec *fvPtr;
    fvPtr = (floatVec*)Otcl::otclPtr->obrefToCpp(name,"floatVec");
    return fvPtr;
}
floatRef *FRfromName(char *name) {
    /* Convert a Tcl name into an FR pointer by accessing the Otcl table */
    floatRef *frPtr;
    frPtr = (floatRef*)Otcl::otclPtr->obrefToCpp(name,"floatRef");
    return frPtr;
}

/* Binary stream access of floatVecs (after bread.c) */

int Tcl_ParseDataSpec(Tcl_Interp *interp, char *arg, 
		      int *ptype, int *pnbytes, int *pbigendian)
{   /* Parse a 0..3 chr specifier for a data format of numbers on
       disk.  The first char can be "f", "i" or "u" for float, int or
       unsigned data.  The second char is the number of bytes, which
       must match sizeof(float,double,char,short,int,long) as
       appropriate.  The third char is "b" for bigendian data, "l" for
       little-endian, or "n" for natural-endian, or "s" for
       swapped-natural.  Pointed-to values are set only if specified
       in the string, else left at defaults.  Return code is TCL_OK on
       success, or an error code.  An error message may be left in the
       interp. *ptype == 0 for float, -1 for int, 1 for unsigned int. */
    int rc = TCL_OK;
    int s = strlen(arg);
    int n = *pnbytes;
#ifdef WORDS_BIGENDIAN
    int  hostbigendian = 1;
#else        /* Must be LITTLEENDIAN */
    int  hostbigendian = 0;
#endif /* BIG_ENDIAN */

    if (s > 3) {
	rc = TCL_ERROR;
    }
    if (s > 0) {
	if (arg[0] == 'f') {
	    *ptype = 0;
	} else if (arg[0] == 'i') {
	    *ptype = -1;
	} else if (arg[0] == 'u') {
	    *ptype = 1;
	} else {
	    rc = TCL_ERROR;
	}
    }
    if (s > 1) {
	if ( (n = arg[1] - '0') < 1 || n > 8 ) {
	    rc = TCL_ERROR;
	}
	if (*ptype == 0) {
	    if ( !(n == sizeof(float) || n == sizeof(double)) ) {
		rc = TCL_ERROR;
	    }
	} else {
	    if ( !(n==sizeof(char) || n==sizeof(short) \
		   || n==sizeof(int) || n==sizeof(long)) ) {
		rc = TCL_ERROR;
	    }
	}
	*pnbytes = n;
    }
    if (s > 2) {
	if (arg[2] == 'b') {
	    *pbigendian = 1;
	} else if (arg[2] == 'l') {
	    *pbigendian = 0;
	} else if (arg[2] == 'n') {
	    *pbigendian = hostbigendian;
	} else if (arg[2] == 's') {
	    *pbigendian = 1 - hostbigendian;
	} else {
	    rc = TCL_ERROR;
	}
    }
    if (rc == TCL_ERROR) {
	Tcl_AppendResult(interp, "format arg \"", arg, 
			 "\" should be {f|i}n{b|l}", NULL);
    }
    return rc;
}

static void swab_ibuf(int *ptr, int n) {
  /* Swap bytes in a buffer of int32s */
  unsigned int d;
  int i;

  assert(sizeof(unsigned int) == 4);
  assert(sizeof(int) == 4);

  for (i=0; i<n; ++i) {
    d = *(unsigned int *)ptr;
    *ptr++ = ((d & 0x000000FF) << 24) | ((d & 0x0000FF00) << 8) | ((d & 0x00FF0000) >> 8) | (0x000000FF & ((d & 0xFF000000) >> 24));
  }
}

static void swab_sbuf(short *ptr, int n) {
  /* Swap bytes in a buffer of int16s */
  int d;
  int i;

  assert(sizeof(short) == 2);

  for (i=0; i<n; ++i) {
      d = *ptr;
      *ptr++ = ((d & 0x00FF) << 8) | (0x00FF & ((d & 0xFF00) >> 8));
  }
}

static void swab(void *buf, int nbytes, int wdsize) {
    /* swap bytes in 2, 4 or 8 byte blocks */
    if (wdsize == 1) {
	return;
    } else if (wdsize == 2) {
	swab_sbuf((short *)buf, nbytes/2);
    } else {
	/* perform 4 byte swapping, then do words if necessary */
	swab_ibuf((int *)buf, nbytes/4);
	if (wdsize == 4) {
	    /* we're all set */
	} else {
	    /* swap pairs of 4 byte words */
	    int i,d;
	    int *iptr = (int *)buf;
	    assert(wdsize==8);
	    assert(sizeof(int) == 4);
	    for (i=0; i<nbytes/8; ++i) {
		d = *iptr; *iptr = *(iptr+1); *(iptr+1) = d;
		iptr += 2;
	    }
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_breadfvecCmd --
 *
 *	Tcl command to read binary-encoded floats from an input stream
 *      into an existing fvec.
 *	Command syntax is: 
 *		breadfvec stream fref {format f4b}
 *      format has up to 3 chrs: the first is "f" or "i" to indicate 
 *      reading float or int values on disk; the second is the number 
 *      of bytes per value; allowable values are f4, f8, i1, i2, i4/
 *      The third char is "b" or "l" for big or little-endian byte order.
 *	This assumes sizeof(float)==4 and sizeof(double)==8.  
 *	sizeof is used instead of constants in the code.
 *
 * Results:
 *      Reads [$fref Size] values from the stream and sets the 
 *      elements of the fref to those values.  Returns the number 
 *      of *bytes* actually read (may be less than Size*$nbytes at EOF).
 *
 * Side effects:
 *	Input file pointer is advanced by $nbytes*[$fref Size] bytes
 *
 *----------------------------------------------------------------------
 */

int Tcl_breadfvecCmd(ClientData clientData, Tcl_Interp *interp, 
		  int argc, char **argv)
{
    Tcl_Channel chan;			/* The channel to read from. */
    int mode;				/* Mode in which channel is opened. */
    int totalread, thistime;		/* byte counters for transfer */
    unsigned char *buf;			/* space to read into */
    floatRef *fr = NULL;
#ifdef WORDS_BIGENDIAN
    int  hostbigendian = 1;
#else        /* Must be LITTLEENDIAN */
    int  hostbigendian = 0;
#endif /* BIG_ENDIAN */

    int  bytesperval = sizeof(float);	/* the number of bytes per value */
    int  bigendian = 1;			/* read vals as bigendian */
    int  readfloats = 1;		/* values are floats not ints */
    int  usigned = 0;
    int  nbytes;

    int  rc = TCL_OK;
    int  i;
    long v;
    float fnum = 0;			/* the answer */
    double dnum = 0;
    int bufallocd = 0;

    int retries = 0;
    int preread = 0;

    if(argc < 3 || argc > 5) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], 
			 " stream fvecref {format f4b} {preread 0}\"", NULL);
	return TCL_ERROR;
    }

    /* handle optional arguments */
    if(argc > 3) {
	int type = 0;
	if ( (rc = Tcl_ParseDataSpec(interp, argv[3], &type, 
				     &bytesperval, &bigendian)) == TCL_ERROR) {
	    return rc;
	}
	if (type == 0) {
	    readfloats = 1;
	} else {
	    readfloats = 0;
	    if (type > 0) {
		usigned = 1;
	    } else {
		usigned = 0;
	    }
	}
    }
    if (argc>4) {
	if (Tcl_GetInt(interp, argv[4], &preread) != TCL_OK) {
	    return TCL_ERROR;
	}
    }


    /* Figure out the channel implicated */
    /* This code is copied from tcl7.5/generic/tclIOCmd.c:Tcl_GetsCmd .. */
    chan = Tcl_GetChannel(interp, argv[1], &mode);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    if ((mode & TCL_READABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[1],
                "\" wasn't opened for reading", (char *) NULL);
        return TCL_ERROR;
    }
    
    /* Get the floatRef */
    if ( (fr = FRfromName(argv[2])) == NULL) {
	Tcl_AppendResult(interp, "could not convert \"", argv[2], 
			 "\" into a floatVec", (char *)NULL);
	return TCL_ERROR;
    }

    /* so how many bytes total do we want? */
    nbytes = fr->Size() * bytesperval;

    /* can we read straight into the fref, or do we need an intermediate 
       buffer? */
    if (bytesperval > sizeof(float) || fr->Step() != 1) {
	buf = (unsigned char *)ckalloc(nbytes);
	bufallocd = 1;
    } else {
	/* write into the final part of the buf space, so we can 
	   convert in-place running forward. */
	buf = ((unsigned char *)&(fr->El(0))) 
	            + fr->Size() * (sizeof(float) - bytesperval) ;
    }

    /* Transfer the bytes */
    thistime = -1;
    totalread = preread;
    if (totalread > 0) {
      fprintf(stderr,"**breadfvec: preread (%d of %d) found in buf (0x%lx=0x%lx...)\n",
	      totalread, nbytes, buf, *(long *)buf);
    }
#define MAX_RETRIES 100
    /*     while(totalread < nbytes && retries < MAX_RETRIES)  */
    while(totalread < nbytes && thistime != 0) {
	thistime = Tcl_Read(chan, (char *)&(buf[totalread]), nbytes-totalread);
	if (thistime == 0) {
	  ++retries;
	}
	if (thistime < 0) {
            Tcl_AppendResult(interp, "error reading \"",
			     Tcl_GetChannelName(chan), "\": ",
			     Tcl_PosixError(interp), (char *) NULL);
            return TCL_ERROR;
	}
	totalread += thistime;
    }

    if (totalread > 0 && totalread < nbytes && bufallocd==0) {
      fprintf(stderr,"**breadfvec: short read (%d of %d) left in buf (0x%lx=0x%lx...)\n",
	      totalread, nbytes, buf, *(long *)buf);
    } else {

    /* maybe reduce size */
    /* fr->Resize(totalread/bytesperval); */
    /* no, just limit where we write.  Already returning bytecount ..? */
    
    /* OK, we got the desired number of bytes into buf */
    /* reverse the bytes if mode is not native */
    if(bigendian != hostbigendian && bytesperval > 1) {
	swab(buf, totalread, bytesperval);
    }

    /* maybe convert vals */
    if (!readfloats || bytesperval != sizeof(float) || fr->Step() != 1) {
	int i;
	if (readfloats) {
	    if (bytesperval == sizeof(float)) {
		for (i = 0; i < fr->Size(); ++i) {
		    fr->El(i) = *(((float *)buf)+i);
		}
	    } else {
		/* must be doubles */
		assert(bytesperval == sizeof(double));
		for (i = 0; i < fr->Size(); ++i) {
		    fr->El(i) = *(((double *)buf)+i);
		}
	    }
	} else {	/* reading ints */
	    /* ok to convert fwds because of how we set up buf */
	    if (usigned) {
		if (bytesperval == sizeof(char)) {
		    for (i = 0; i < fr->Size(); ++i) {
			fr->El(i) = *(((unsigned char *)buf)+i);
		    }
		} else if (bytesperval == sizeof(short)) {
		    for (i = 0; i < fr->Size(); ++i) {
			fr->El(i) = *(((unsigned short*)buf)+i);
		    }
		} else if (bytesperval == sizeof(int)) {
		    for (i = 0; i < fr->Size(); ++i) {
			fr->El(i) = *(((unsigned int*)buf)+i);
		    }
		} else {
		    assert(bytesperval == sizeof(long));
		    for (i = 0; i < fr->Size(); ++i) {
			fr->El(i) = *(((unsigned long*)buf)+i);
		    }
		}
	    } else { /* !usigned */
		if (bytesperval == sizeof(char)) {
		    for (i = 0; i < fr->Size(); ++i) {
			fr->El(i) = *(((char *)buf)+i);
		    }
		} else if (bytesperval == sizeof(short)) {
		    for (i = 0; i < fr->Size(); ++i) {
			fr->El(i) = *(((short*)buf)+i);
		    }
		} else if (bytesperval == sizeof(int)) {
		    for (i = 0; i < fr->Size(); ++i) {
			fr->El(i) = *(((int*)buf)+i);
		    }
		} else {
		    assert(bytesperval == sizeof(long));
		    for (i = 0; i < fr->Size(); ++i) {
			fr->El(i) = *(((long*)buf)+i);
		    }
		}
	    }
	}
    }
    }
    if (bufallocd) {
	free(buf);
    }
    
    /* return val is number of bytes read */
    sprintf(interp->result, "%d", totalread);

    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * Tcl_bwritefvecCmd --
 *
 *	Tcl command to write binary-encoded floats from a floatRef
 *      to an output stream.
 *	Command syntax is: 
 *		bwritefvec stream fref {format f4b}
 *	format as for breadfvec
 *
 * Results:
 *      Sets interp->result to the number of bytes written (0 on error).
 *
 * Side effects:
 *      File pointer is advanced by $nbytes*[$fref Size] bytes; 
 *      data is written.
 *
 *----------------------------------------------------------------------
 */

int Tcl_bwritefvecCmd(ClientData clientData, Tcl_Interp *interp, 
		  int argc, char **argv)
{
    Tcl_Channel chan;			/* The channel to write to. */
    int mode;				/* Mode in which channel is opened. */
    int totaldone, thistime;		/* byte counters for transfer */
    unsigned char *buf;			/* space to write from */
    floatRef *fr = NULL;
#ifdef WORDS_BIGENDIAN
    int  hostbigendian = 1;
#else        /* Must be LITTLEENDIAN */
    int  hostbigendian = 0;
#endif /* BIG_ENDIAN */

    int  bytesperval = sizeof(float);	/* the number of bytes per value */
    int  bigendian = 1;			/* write vals as bigendian */
    int  writefloats = 1;		/* values are floats not ints */
    int  usigned = 0;
    int  nbytes;

    int  rc = TCL_OK;
    int  i;
    long v;
    float fnum = 0;			/* the answer */
    double dnum = 0;
    int bufallocd = 0;

    if(argc < 3 || argc > 4) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], 
			 " stream fvecref {format f4b}\"", NULL);
	return TCL_ERROR;
    }

    /* handle optional arguments */
    if(argc > 3) {
	int type = 0;
	if ( (rc = Tcl_ParseDataSpec(interp, argv[3], &type, 
				     &bytesperval, &bigendian)) == TCL_ERROR) {
	    return rc;
	}
	if (type == 0) {
	    writefloats = 1;
	} else {
	    writefloats = 0;
	    if (type > 0) {
		usigned = 1;
	    } else {
		usigned = 0;
	    }
	}
    }

    /* Figure out the channel implicated */
    /* This code is copied from tcl7.5/generic/tclIOCmd.c:Tcl_GetsCmd .. */
    chan = Tcl_GetChannel(interp, argv[1], &mode);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    if ((mode & TCL_WRITABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[1],
                "\" wasn't opened for writing", (char *) NULL);
        return TCL_ERROR;
    }
    
    /* Get the floatRef */
    if ( (fr = FRfromName(argv[2])) == NULL) {
	Tcl_AppendResult(interp, "could not convert \"", argv[2], 
			 "\" into a floatRef", (char *)NULL);
	return TCL_ERROR;
    }

    /* so how many bytes total do we want to write? */
    nbytes = fr->Size() * bytesperval;

    /* can we write straight from the fref, or do we need an intermediate 
       buffer? */
    if (writefloats && bytesperval == sizeof(float) && fr->Step() == 1) {
	/* just write from memory */
	buf = (unsigned char *)&(fr->El(0));
    } else {
	buf = (unsigned char *)ckalloc(nbytes);
	bufallocd = 1;
	/* convert values */
	if (writefloats) {
	    if (bytesperval == sizeof(float)) {
		for (i = 0; i < fr->Size(); ++i) {
		    *(((float *)buf)+i) = fr->El(i);
		}
	    } else {
		/* must be doubles */
		assert(bytesperval == sizeof(double));
		for (i = 0; i < fr->Size(); ++i) {
		    *(((double *)buf)+i) = fr->El(i);
		}
	    }
	} else {	/* writing ints */
	    if (usigned) {
		if (bytesperval == sizeof(unsigned char)) {
		    for (i = 0; i < fr->Size(); ++i) {
			*(((unsigned char *)buf)+i) = fr->El(i);
		    }
		} else if (bytesperval == sizeof(unsigned short)) {
		    for (i = 0; i < fr->Size(); ++i) {
			*(((unsigned short*)buf)+i) = fr->El(i);
		    }
		} else if (bytesperval == sizeof(unsigned int)) {
		    for (i = 0; i < fr->Size(); ++i) {
			*(((unsigned int*)buf)+i) = fr->El(i);
		    }
		} else {
		    assert(bytesperval == sizeof(unsigned long));
		    for (i = 0; i < fr->Size(); ++i) {
			*(((unsigned long*)buf)+i) = fr->El(i);
		    }
		}
	    } else {
		if (bytesperval == sizeof(char)) {
		    for (i = 0; i < fr->Size(); ++i) {
			*(((char *)buf)+i) = fr->El(i);
		    }
		} else if (bytesperval == sizeof(short)) {
		    for (i = 0; i < fr->Size(); ++i) {
			*(((short*)buf)+i) = fr->El(i);
		    }
		} else if (bytesperval == sizeof(int)) {
		    for (i = 0; i < fr->Size(); ++i) {
			*(((int*)buf)+i) = fr->El(i);
		    }
		} else {
		    assert(bytesperval == sizeof(long));
		    for (i = 0; i < fr->Size(); ++i) {
			*(((long*)buf)+i) = fr->El(i);
		    }
		}
	    }
	}
    }

    /* reverse the bytes if mode is not native */
    if(bigendian != hostbigendian && bytesperval > 1) {
	swab(buf, nbytes, bytesperval);
    }

    /* Transfer the bytes */
    thistime = -1;
    totaldone = 0;
    while(totaldone < nbytes && thistime != 0) {
	thistime = Tcl_Write(chan, (char *)&(buf[totaldone]), 
			     nbytes-totaldone);
	if (thistime < 0) {
            Tcl_AppendResult(interp, "error writing \"",
			     Tcl_GetChannelName(chan), "\": ",
			     Tcl_PosixError(interp), (char *) NULL);
	    rc = TCL_ERROR;
	    goto cleanup;
	}
	totaldone += thistime;
    }

    /* return val is number of bytes done */
    sprintf(interp->result, "%d", totaldone);

cleanup:
    /* swap back if did in place */
    if (buf == (unsigned char *)&(fr->El(0))) {
	if(bigendian != hostbigendian && bytesperval > 1) {
	    swab(buf, nbytes, bytesperval);
	}
    }
    if (bufallocd) {
	free(buf);
    }
    
    return rc;
}
