#ifndef CDL_METHOD_H
#define CDL_METHOD_H

/*  _ __ ___ _
 * | |\ /  /| |  $Id: CdlMethod.H,v 1.7 1998/12/08 05:17:20 dpwe Exp $
 * | | /  / | |  Copyright (C) 1995 IXI Limited.
 * |_|/__/_\|_|  IXI Limited, Cambridge, England.
 *
 * Component   : CdlMethod.H
 *
 * Author      : Dean Sheehan (deans@x.co.uk)
 *  
 * Description : Header file for CdlMethod class that models methods in the
 *               CDL input file.
 *
 * License     :
			Object Tcl License & Copyright
			-----------------------------

IXI Object Tcl software, both binary and source (hereafter, Software) is copyrighted by IXI Limited (IXI), and ownership remains with IXI. 

IXI grants you (herafter, Licensee) a license to use the Software for academic, research and internal business purposes only, without a fee. Licensee may distribute the binary and source code (if required) to third parties provided that the copyright notice and this statement appears on all copies and that no charge is associated with such copies. 

Licensee may make derivative works. However, if Licensee distributes any derivative work based on or derived from the Software, then Licensee will (1) notify IXI regarding its distribution of the derivative work, and (2) clearly notify users that such derivative work is a modified version and not the original IXI Object Tcl distributed by IXI. IXI strongly recommends that Licensee provide IXI the right to incorporate such modifications into future releases of the Software under these license terms. 

Any Licensee wishing to make commercial use of the Software should contact IXI, to negotiate an appropriate license for such commercial use. Commercial use includes (1) integration of all or part of the source code into a product for sale or license by or on behalf of Licensee to third parties, or (2) distribution of the binary code or source code to third parties that need it to utilize a commercial product sold or licensed by or on behalf of Licensee. 

IXI MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE. IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY. IXI SHALL NOT BE LIABLE FOR ANY DAMAGES WHATSOEVER SUFFERED BY THE USERS OF THIS SOFTWARE. 

Copyright (C) 1995, IXI Limited 

By using or copying this Software, Licensee agrees to abide by the copyright law and all other applicable laws of England and the U.S., including, but not limited to, export control laws, and the terms of this license. IXI shall have the right to terminate this license immediately by written notice upon Licensee's breach of, or non-compliance with, any of its terms. Licensee may be held legally responsible for any copyright infringement that is caused or encouraged by Licensee's failure to abide by the terms of this license. 

Comments and questions are welcome and can be sent to
otcl@x.co.uk 

For more information on copyright and licensing issues, contact: 
Legal Department, IXI Limited, Vision Park, Cambridge CB4 4ZR,
ENGLAND. 

 *
 */

// System Includes
#include <fstream.h>

// Local Includes
#include "CdlItem.H"

// Forward Class Declarations
class CdlArg;
class CdlRtn;

// Global fn to quote script strings
void putScript(ostream& of, char *script);

// Public Defines
#define MAX_ARGS 20

// the name for the self-variable in DUMMYTCL code fragments.
// should be "this", but interacts badly with [incr tcl] interp
// no, "this" is ok.
// no, "this" is a problem - if you're using it inside an itcl class, 
// $this is preset and can't be written.
//#define TCL_THIS	"this"
//#define TCL_THIS	"self"
#define TCL_THIS	"__this"
// you'll also have to fix Test 41.2 in tscript


class CdlMethod
{
public: // Public Enumerations

   enum Binding {STATIC, DYNAMIC, DUMMYTCL, DUMMYCPP};

public: // Public Constructor & Destructor

   CdlMethod (char *name);

   virtual ~CdlMethod ();

public: // Public Instance Methods

    int addArg (CdlArg *m);	// returns 0 if arg bad to add

   void setRtn (CdlRtn *r);
   CdlRtn *getRtn (void);

   int isPublic (void);

   void setVoidReturn (void);
   int returnIsVoid (void);

   void setVoidArgs (void);
   int argsAreVoid (void);

   int getNoOfArgs (void);
   int getMinNoOfArgs (void);

   const char *giveName(void)	{ return name; }

    void setScript(const char *s);

   void executeBinding_sub (char *className, char *cname, ofstream &, CdlIndent, CdlClass* pclass, char *objectPrefix);

   void setBinding (Binding b);

   int isDynamic (void);

protected:  // Private Instance Attributes

   CdlArg *arg[MAX_ARGS];
   int noOfArgs;
    int minNoOfArgs;	// fewest acceptable number of args (non-optional)

   CdlRtn *rtn;

   char *name;

    char *script;

   int voidReturn;
   int voidArgs;

   Binding binding;
};

class CdlInstanceMethod : public CdlMethod
{
public:

   CdlInstanceMethod (char *name);

   ~CdlInstanceMethod ();

   void genDec (ofstream &);

   void genDef (char *className, char *cName, ofstream &, int);

   void executeBinding (char *className, char *cname, ofstream &, CdlIndent, CdlClass* pclass);

};

class CdlClassMethod : public CdlMethod
{
public:

   CdlClassMethod (char *name);

   ~CdlClassMethod ();

   void executeBinding (char *className, char *cname, ofstream &, CdlIndent, CdlClass* pclass);
};

class CdlConstructor : public CdlMethod
{
public:
    int castable;	// is the related class castable?

   CdlConstructor (char *name, int castb = 0);

   ~CdlConstructor ();

   void genDec (ofstream &);

   void defineArgs (ofstream &, int includeOtcl = 1);
   void passArgsUpwards (ofstream &);

   void preProcessArgs (ofstream &, CdlIndent);
   void passArgs (ofstream &, char *lastArg);
   void postProcessArgs (ofstream &, CdlIndent);

   int isCopyConstructor(char *cls);	// does this method just take an obref?
    // are any args "-from"? for CdlClass::constructorOK()
    int hasFromArgs(void);
};

#endif // CDL_METHOD
