#! /bin/sh
# -*- Tcl -*-   The next line restarts using our private version of wish \
exec drspeech_wish $0 ${1+"$@"}
 
#!/usr/local/bin/wish
#
# sndview.tcl
#
# Sound file viewer based on Tcl extensions
# (after tsoundg)
# 
# 1998apr20 dpwe@icsi.berkeley.edu
# $Header: /u/drspeech/src/guitools/sound_otcl/RCS/sndview.tcl,v 1.7 2000/02/17 03:33:11 dpwe Exp $

if { $argc < 1 } {
    puts "usage: $argv0 <soundfile>"
    exit -1
}

if {[array names env SPEECH_DIR] == ""} {
    set SPEECH_DIR "/u/drspeech"
} else {
    set SPEECH_DIR $env(SPEECH_DIR)
}
lappend auto_path $SPEECH_DIR/share/lib
package require Dpwe_Utilfns
lappend auto_path $SPEECH_DIR/[SpeechArch]/lib
#lappend auto_path "/u/dpwe/sun4-sunos5/lib"

package require Farray
package require Sound
package require TkSound
package require StftSound
# for timeline widget
package require Dpweutils_tk

if {[file readable libsound_otcl.so]} {
    # For pre-install version, load local sound_otcl if available
    package require Otcl
    otcl list
    floatArray fadummy
    puts "loading local libsound_otcl.so"
    load ./libsound_otcl.so Sound
}

if {[file readable libsound_tk.so]} {
    # For pre-install version, load local sound_tk if available
    puts "loading local libsound_tk.so"
    load ./libsound_tk.so
}

# load the sound
set filename [lindex $argv 0]
set sound so
Sound $sound
$sound open $filename
#$sound play

# Global max frames to show in a view
set gMaxViewLen 1000000

set gSampleRate [$sound getSampleRate]

######### Widget callback functions ###########

proc timelineAndScrollCommand {timeline scrollbar image sound \
				totConts viewSize top bottom} {
    # normally the soundview scrollback command plugs straight into 
    # the scrollbar - but here we steal off a bit to feed the time line 
    # as well.
    # Also, we can take this opportunity to scale the `increment' used 
    # for the end-arrows in the scrollbar.
#    puts "$totConts $viewSize $top $bottom"
#   e.g 52941 10000 8100 18100
    global scrollScale
#    set bottom [expr "$top+$viewSize"]
    # that's what I actually want
    set srate [$sound getSampleRate]
    set tmin  [expr "$top / $srate"]
    set tmax  [expr "$bottom/$srate"]
    # set the timeline to display the new range
    $timeline config -minVal $tmin -maxVal $tmax
    # now move the scrollbar thumb to the right place
    # scroll bar operates in units 1/16th of the total short view width
    set scrollScale [expr "$viewSize / 200"]
    $scrollbar set [expr "$totConts/$scrollScale"] \
		   [expr "$viewSize/$scrollScale"] \
		   [expr "$top/$scrollScale"] [expr "$bottom/$scrollScale"]
    # also update the image to display the right range of columns
    if {$image != ""} {
	set icols [expr 0.0+[$image cget -cols]]
	set ibase [expr "$icols * $top / $totConts"]
	set ilen  [expr "$icols * $viewSize / $totConts"]
	$image config -cbase $ibase -clen $ilen
    }
}

proc scrollbarCommand {soundview baseval} {
    # quick filter to take output from scrollbar callback and scale to samples
    global scrollScale
    set newbase [expr "$baseval * $scrollScale"]
    sndconfigview $soundview $newbase
}

proc nosoundScrollbarCommand {soundview baseval} {
    # scrollback command that just scrolls the spectrogram, not the shortview
    # scanning the soundfile is just too slow
    global scrollScale
    set bv [expr "$baseval * $scrollScale"]
    set sound [$soundview cget -sound]
    set tc [$sound getFrames]
    set vs [$soundview cget -viewLen]
    timelineAndScrollCommand .sv.tl .sb .im $sound $tc $vs $bv \
	                       [expr "$bv + $vs"]
}

# xpad is supposed to be size of scrollbar arrows
set xpad 13
# 16 if scrollbar relief is sunken
set soDuration [expr "[$sound getFrames]/[$sound getSampleRate]"]

#farray fa
#fa load $hexname

##########  Initialize all the widgets #############
#set image [image     .im     -farray fa -padX $xpad -padY 5 -relief flat]
frame     .im     -relief flat
set image ""
#shmimage     .im     -farray fa -padX $xpad -padY 5 -relief flat
#$image config  -rlen [$image cget -rows]

frame 	  .status -borderwidth 2 -relief groove
label     .status.lab -text "File: $filename"
set sellabel [label .status.sel -text "Selection:" -width 40 -justify right -anchor e]
pack .status.lab -side left  -fill y -expand 0
pack .status.sel -side right -fill y -expand 0

# Swap the commenting of the next two lines for faster scrolling 
#   which does not update the shortview until mouseup:
### Slower, prettier scrolling
scrollbar .sb     -orient horiz -relief flat
# -command "scrollbarCommand $sw"

set sndLen [$sound getFrames]
frame     .ov
set ov [sndwvfm   .ov.sw  -sound $sound -padX $xpad -padY 5 -height 100 \
		  -viewLen $sndLen -noscan [expr $sndLen > $gMaxViewLen]]
timeline  .ov.tl  -padX $xpad -padY 0 -height 20 -maxVal $soDuration \
		  -tixTopLeft 0 -relief flat

frame     .sv
timeline  .sv.tl  -padX $xpad -padY 0 -height 20 -maxVal $soDuration \
		  -tixTopLeft 1 -relief flat
set sw [sndwvfm   .sv.sw  -sound $sound -padX $xpad -padY 5 -viewLen 10000 \
		  -scrollCommand "timelineAndScrollCommand .sv.tl .sb \"$image\" \
				  \[.sv.sw cget -sound\]"]
# ".sb set"

scale .sv.ga -orient vert -relief flat -from 9 -to -9 -command "newgain $sw"

proc newgain {sw val} {
    # Called from vert scrollback
    $sw configure -vscale [expr exp($val*log(100.0)/10.0)]
}

.sb config -command "scrollbarCommand $sw"
### Faster, cruder scrolling
#.sb config -command "nosoundScrollbarCommand $sw"

# drag handles - fillers to allow vertical resizing
#frame .drag1 -height 5 -cursor bottom_side
#frame .drag2 -height 5 -cursor bottom_side
#frame .drag3 -height 5 -cursor bottom_side

# global list of views associated with this sound
set $sound.views "$sw $ov"

frame  .buttons
button .buttons.zoomin  -text " < "  -command "zoom $sw .7"
button .buttons.zoomout -text " > "  -command "zoom $sw 1.4"
button .buttons.zoomsel -text "zoom" -command "zoomsel $sw"
button .buttons.quit    -text "quit" -command "exit"
set playbtn [button .buttons.play    -text "play" -command "svplay $sw"]

######### Pack all the widgets ###########
pack .status          -side top -fill x
pack .buttons         -side bottom -fill x -padx 6 -pady 6
pack .buttons.zoomin  -side left
pack .buttons.zoomout -side right
pack .buttons.play .buttons.zoomsel .buttons.quit -side left -expand yes

pack .ov.tl  -side top    -fill x
pack .ov.sw  -side top    -fill x
pack .ov     -side top    -fill x
#pack .drag1  -side top    -fill x

pack .sb     -side top    -fill x

pack .sv.ga  -side left   -fill y
pack .sv.sw  -side top	  -fill both -expand yes
pack .sv.tl  -side bottom -fill x
pack .sv     -side top    -fill both -expand yes
#pack .drag2  -side top    -fill x

# pack .im    -side top    -fill both
# pack .drag3 -side top    -fill x
# pack .swtl  -side top    -fill x
### doesn't allow image panel to grow/shrink when whole window resized

#pack .drag3 -side bottom -fill x
pack .im    -side bottom -fill x

bind .sb <ButtonRelease-1> 	{scrollbarCommand $sw [lindex [.sb get] 2]}

########## dragging bit

proc startDragProc {widg target x y} {
    # widg is the drag widget, target is the widget whose height is 
    # to track the up/down motion over this widget
    $widg config -cursor sb_v_double_arrow
    set my [expr "$y - [winfo rooty .]"]
    global dragStart$widg startHeight$target
#    set dragStart$widg [lindex [local2global $widg $x $y] 1]
    set dragStart$widg $my
    set startHeight$target [$target cget -height]
#    puts "startDrag: y [set dragStart$widg] h [set startHeight$target]"
    xorrect . 0 $my 2000 3
}

proc runDragProc {widg target x y} {
    global dragStart$widg startHeight$target
    set my [expr "$y - [winfo rooty .]"]
#    puts "runDrag: y $my"
#    $target config -height [expr "[set startHeight$target] + $my - [set dragStart$widg]"]
    xorrect . 0 $my 2000 3
}

proc endDragProc {widg target x y} {
    global dragStart$widg startHeight$target
#    set gy [lindex [local2global $widg $x $y] 1]
    set my [expr "$y - [winfo rooty .]"]
    $widg config -cursor bottom_side
#    puts "runDrag: y $my"
    xorrect .
    $target config -height [expr "[set startHeight$target] + $my - [set dragStart$widg]"]
}

#bind .drag1 <1> 			{startDragProc %W $ov %X %Y}
#bind .drag1 <B1-Motion> 		{runDragProc   %W $ov %X %Y}
#bind .drag1 <ButtonRelease-1> 		{endDragProc   %W $ov %X %Y}

#bind .drag2 <1> 			{startDragProc %W $sw %X %Y}
#bind .drag2 <B1-Motion> 		{runDragProc   %W $sw %X %Y}
#bind .drag2 <ButtonRelease-1> 		{endDragProc   %W $sw %X %Y}

#bind .drag3 <1> 			"startDragProc %W $image %X %Y"
#bind .drag3 <B1-Motion> 		"runDragProc   %W $image %X %Y"
#bind .drag3 <ButtonRelease-1> 		"endDragProc   %W $image %X %Y"


######## Selection and autoscrolling functions ############

proc scroll {sndview percent} {
    # shift the current view by some percentage of the current view - 
    #   forward if $percent > 0, else backward
    set viewBase [$sndview cget -viewBase]
    set viewLen  [$sndview cget -viewLen ]
    set newBase [expr "$viewBase + ($viewLen * $percent)/100"]
    sndconfigview $sndview $newBase
}

proc sndconfigview {sndview newBase {newLen 0}} {
    # Configure the sndview to have a new base, but clip it at the ends
    set viewLen  [$sndview cget -viewLen]
    set sndLen   [[$sndview cget -sound] getFrames]
    if {$newBase < 0} {
	set newBase 0
    }
    if {$newLen <= 0} {
	# newLen not specified - use last one, and clip by it
	if {$newBase+$viewLen > $sndLen} {
	    set newBase [expr $sndLen - $viewLen]
	}
	$sndview config -viewBase $newBase
    } else {
	# newLen specified, clip it according to newBase
	if {$newBase+$newLen > $sndLen} {
	    set newLen [expr $sndLen-$newBase]
	}
	$sndview config -viewBase $newBase -viewLen $newLen
    }
    return $newBase
}

proc setAllViews {viewsnames base len} {
    # Setup the ranges on all the appropriate views, also the status
    global sellabel gMaxViewLen gSampleRate $viewsnames
    if {$base < 0} {
	set len [expr $len+$base]
	set base 0
    }
    foreach view [set $viewsnames] {
	$view config -selBase $base -selLen $len \
		-noscan [expr $len > $gMaxViewLen]
    }
    set fstart [format %.3f [expr $base/(0.0+$gSampleRate)]]
    set fend   [format %.3f [expr ($base+$len)/(0.0+$gSampleRate)]]
    if {$len == 0} {
	$sellabel config -text "Selection: $fstart"
    } else {
	$sellabel config -text "Selection: $fstart - $fend"
    }
}

proc startSel {sndview xpix}  {
    # mouse down in view
    global anchor$sndview sellab
    set frame [$sndview frameOfPixel $xpix]
    # set new selection in all concerned views
    set viewsname [$sndview cget -sound].views
    setAllViews $viewsname $frame 0
    set anchor$sndview $frame
}

# semaphore to pre-empt auto-repeats
set sndwvfm_auto_pending 0
set sndwvfm_cancel_auto  0

proc endSel {sndview} {
    # called when the scroll ends: if an auto pending, be sure to cancel it
    global sndwvfm_auto_pending sndwvfm_cancel_auto
    if { $sndwvfm_auto_pending }  {
	set sndwvfm_cancel_auto 1
    }
}

proc scrollGrowSel {sndview scrollamt {growAtEnds 1}}  {
    # scroll the view and act as if the button was moved, since it 
    # has moved, relative to the scrolled view.
    # This command is executed by a timer a little while after the 
    # selection is grown into the 'end zones'
    global sndwvfm_auto_pending sndwvfm_cancel_auto currentX$sndview
    set sndwvfm_auto_pending 0
    if { $sndwvfm_cancel_auto } {
	# this autoscroll event was cancelled before it ever got to happen
	set sndwvfm_cancel_auto 0
    } else {
	scroll $sndview $scrollamt
	# re-execute the scroll .. 
	growSel $sndview [set currentX$sndview] $growAtEnds
    }
}

proc checkForAuto {sndview xpix}  {
    # if we're near the edge of the view, schedule an auto-scroll
    # ('near' is within $border pixels)
    global sndwvfm_auto_pending sndwvfm_cancel_auto
    set border 50
    set lowerlim $border
    set upperlim [expr "[winfo width $sndview] - $border"]
    if { $xpix < $lowerlim || $xpix > $upperlim } {
	# calculate the scroll amount so that it depends how far the 
	# mouse is into the autoscroll region
	if { $xpix < $lowerlim } {
	    set xoffset [expr "$xpix - $lowerlim"]
	} else {
	    set xoffset [expr "$xpix - $upperlim"]
	}
	set scrollamt [expr "(5 * $xoffset)/$border"]
	if { $sndwvfm_auto_pending == 0 } {
	    after 200 "scrollGrowSel $sndview $scrollamt"
	    set sndwvfm_auto_pending 1
	}
	# just in case we have cancelled a pending auto, re-instance it
	set sndwvfm_cancel_auto 0
    } else {
	# we have moved out of the autoscroll zone; cancel any pending scroll
	if { $sndwvfm_auto_pending } {
	    set sndwvfm_cancel_auto 1
	}
    }
}

proc growSel {sndview xpix {growAtEnds 1}}  {
    # mouse drag in view - just extend the selection
    global anchor$sndview currentX$sndview gMaxViewLen
    # save the current x position for the redraw to figure
    set currentX$sndview $xpix
    # better set up for an autoscroll if we need one
    if {$growAtEnds} {
	checkForAuto $sndview $xpix
    }
    # OK, now just grow the selection
    set origPos [set anchor$sndview]
    set newPos  [$sndview frameOfPixel $xpix]
    if { $newPos < $origPos } {
	set tmp $newPos
	set newPos $origPos
	set origPos $tmp
    }
    set newLen [expr "$newPos - $origPos"]
    # set new selection in all involved windows
    set viewsname [$sndview cget -sound].views
    setAllViews $viewsname $origPos $newLen
}

proc startSelExtend {sndview xpix {growAtEnds 1}}  {
    # mouse down in view, but keep more distant of current anchor points
    global anchor$sndview
    set selBase [$sndview cget -selBase]
    set selLen  [$sndview cget -selLen ]
    set selEnd  [expr "$selBase + $selLen"]
    set newPos  [$sndview frameOfPixel $xpix]
    if {$selLen == 0}  {
	# no selection to extend - act like button 1
	startSel $sndview $xpix
    } else {
	# set the anchor to the more distant of the selection points
        if { abs($newPos - $selBase) > abs($newPos - $selEnd) } {
	    set anchor$sndview $selBase
	} else {
	    set anchor$sndview $selEnd
	}
	# and now act as if cursor just moved
	growSel $sndview $xpix $growAtEnds
    }
}

####### Bind all the selection and scroll functions #######

bind $sw <1> 			{startSel $sw %x}
bind $sw <B1-Motion> 		{growSel $sw %x}
bind $sw <B1-ButtonRelease> 	{endSel $sw}
bind $sw <3> 			{startSelExtend $sw %x}
bind $sw <B3-Motion> 		{growSel $sw %x}
bind $sw <B3-ButtonRelease> 	{endSel $sw}
# dspB binding - middle button zooms to mark
bind $sw <2> 			{zoomsel $sw}

bind $ov <1> 			{startSel $ov %x}
bind $ov <B1-Motion> 		{growSel $ov %x 0}
bind $ov <B1-ButtonRelease> 	{endSel $ov}
bind $ov <3> 			{startSelExtend $ov %x 0}
bind $ov <B3-Motion> 		{growSel $ov %x 0}
bind $ov <B3-ButtonRelease> 	{endSel $ov}
bind $ov <2> 			{zoomsel $sw}

##############  Zoom, other button functions ###################

proc sndwvfmSelLims {sndview {noneisall 0}} {
    # return the selection base and len.
    # If $noneisall is present and nonzero, return the entire sound limits 
    # if the selection is zero len
    global gMaxViewLen
    set selBase  [$sndview cget -selBase]
    set selLen   [$sndview cget -selLen ]
    set viewBase [$sndview cget -viewBase]
    set viewLen  [$sndview cget -viewLen ]
    set sound    [$sndview cget -sound]
    set sndLen   [$sound getFrames]
    set maxLen   [min $sndLen $gMaxViewLen]
    if { $noneisall && ($selLen == 0) } {
	# Attempt to center the new section on the old one
#	set selBase [expr $selBase + ($selLen/2) - ($maxLen)/2]
#	if {$selBase < 0} {set selBase 0}
#	if {$selBase + ($maxLen/2) > $sndLen} {
#	    set selBase [expr $sndLen - $maxLen]
#	}
#	set selLen $maxLen
        set selLen [expr $sndLen - $selBase]
    }
    return [list $selBase $selLen $viewBase $viewLen $maxLen]
}

proc zoom {sndview factor} {
    # adjust the display length of a sndview by the specified factor
    # if factor > 1.0, zooming out
    # attempt to keep view centered at same point
    set sellims [sndwvfmSelLims $sndview 1]
    set viewBase [lindex $sellims 2]
    set viewLen  [lindex $sellims 3]
    set newLen   [expr "int($factor * $viewLen)"]
    set newBase  [expr "$viewBase + ($viewLen - $newLen)/2"]
    # newBase may go negative, but $sndview config will clip it nicely
    sndconfigview $sndview $newLen $newBase
}

proc zoomsel {sndview} {
    # zoom sound view to current selection - or whole sound if none
    set sellims [sndwvfmSelLims $sndview 1]
    set selBase [lindex $sellims 0]
    set selLen  [lindex $sellims 1]
    set viewBase [lindex $sellims 2]
    set viewLen  [lindex $sellims 3]
    set maxLen   [lindex $sellims 4]
    # if we're already zoomed to sel, zoom out to full (toggle)
    if { ($viewBase == $selBase) && ($viewLen == $selLen) } {
	set selBase 0
	set selLen $maxLen
    }
    sndconfigview $sndview $selBase $selLen
}

set gPlayingFlag 0
proc svplay {sndview} {
    # play the selected region of the soundview
    global gPlayingFlag playbtn
    if {$gPlayingFlag != 0} {
	set gPlayingFlag 0
    } else {
	# Make sure semaphore is set up to permit stopping
	set gPlayingFlag 1
	$playbtn config -text "Stop"
	set sound   [$sndview cget -sound]
	set sellims [sndwvfmSelLims $sndview 1]
	set selBase [lindex $sellims 0]
	set selLen  [lindex $sellims 1]
	# Also, save actual sel params for restoration
	set eSelBase [$sndview cget -selBase]
	set eSelLen  [$sndview cget -selLen]
#puts stderr    "$sound play $selBase $selLen"
#    $sound play $selBase $selLen
	set blksize 1024
	set viewsname [$sndview cget -sound].views
	set pos $selBase
	set remain $selLen
	$sound setAudioOpen 1
	while {$remain > 0 && $gPlayingFlag != 0} {
	    set frames [min $remain $blksize] 
	    setAllViews $viewsname $pos 0
	    update
	    $sound play $pos $frames
	    incr pos $frames
	    incr remain [expr -$frames]
	}
	$sound setAudioOpen 0
	set gPlayingFlag 0
	# Restore entry selection
	setAllViews $viewsname $eSelBase $eSelLen
	$playbtn config -text "Play"
    }
}

# Startup code
set viewsname ${sound}.views
#setAllViews $viewsname 0 0
