# $Header: /u/drspeech/src/SPRACHcore/H-i586-linux/intvec/../../intvec/RCS/intvec-scansrc.perl,v 1.3 1995/11/21 02:10:43 davidj Exp $
#

$PROGNAME = "intvec-scansrc.perl";
$verbose = "";
$hdrfile = "";
$makefile = "";
$srcdir = "";
@prelist = ();

################################################################
### Initialization and command line arguments
################################################################

sub usage
{
    print(STDERR "Usage: perl $PROGNAME [-v] [-p <path1>,<path2>,...] [-h <outhdr>]\n");
    print(STDERR "            [-m <outmake>] src1 src2 ...\n");
# e.g. perl $PROGNAME -v -p tor,php,nv -h out.h -m Makefile.frag *.c
}

while ($_=$ARGV[0],/^-/)
{
	shift;
	last if /^--$/;
	/^-v$/ && do { $verbose++; next};
	/^-p$/ && do { @prelist = split(/,/,$ARGV[0]); shift; next};
	/^-h$/ && do { $hdrfile = $ARGV[0]; shift; next};
	/^-m$/ && do { $makefile = $ARGV[0]; shift; next};
        /^-s$/ && do { $srcdir = $ARGV[0]; shift; next};
	/-.*/ && do { do &usage; exit 1; };
}
@all_srcs = @ARGV;

if ($verbose)
{
	print "hrdfile=$hdrfile\n";
	print "makefile=$makefile\n";
        print "srcdir=$srcdir\n";
	print "prelist=", join(",", @prelist), "\n";
	print "all_srcs=\n", join("\n", @all_srcs), "\n";
}

################################################################
### Find the sources we will use
################################################################

@lib_srcs = ();
foreach $pre (@prelist)
{
    push(@lib_srcs, grep($_ =~ /^$pre/, @all_srcs));
}
if ($verbose)
{
       print "lib_srcs=\n", join("\n", @lib_srcs), "\n";
}

################################################################
### Scan the source files
################################################################

if ($verbose)
{
    print "Scanning...\n";
}
foreach $file (@lib_srcs)
{
    open(FILE, "$srcdir/$file")
	|| die("$PROGNAME: ERROR - Failed to open $file\n");
    if ($verbose)
    {
	print "$srcdir/$file\n";
    }
    $func = "";
    while (<FILE>)
    {
	if (/^\s*#if.*DEFN_/)
        {
	    /DEFN_([^_]*)_(\w*)/;
	    $func = $1 . "_" . $2;
	    if (grep(/^$1$/, @prelist))
	    {
		$obj2src{$func.".o"} = $file;
	    }
	}
	elsif (/^\s*#if.*DECL_/)
	{
	    /DECL_([^_]*)_(\w*)/;
	    $long = $1 . "_" . $2;
	    if (grep(/^$1$/, @prelist))
	    {
		$macro = "";
		$line = <FILE>;
		# Note we stop when we get #endif (no space between # and e)
		while (!($line =~ /^\s*#endif/))
		{
		    $macro .= $line;
		    $line = <FILE>;
		}
		$long2proto{$long} = $macro;
	    }
        }
    }
    # Here we handle a file with no "#ifdef DEFN_" - use source file name
    # ..to build object file name
    if ($func eq "")
    {
	$file =~ m!.*/([^/]*)[.]\w+!;
	$func = $1;
	$obj2src{$func.".o"} = $file;
    }
}
@long_names = sort(keys(%long2proto));
@lib_objs = sort(keys(%obj2src));
if ($verbose)
{
    print "objs=\n", join("\n", @lib_objs), "\n";
    print "defns=\n", join("\n", @long_names), "\n";
}

################################################################
### Build up aliases for the header file
################################################################

foreach $pfx (reverse(@prelist))
{
	@pfxd_names = grep(/^$pfx[_]/, @long_names);
	foreach $long_name (@pfxd_names)
	{
		$short_name = $long_name;
		$short_name =~ s/^$pfx[_]//;
		$short2macro{$short_name} = $long_name;
	}
}
if ($verbose)
{
    print "aliases=\n";
    foreach $i (keys(%short2macro))
    {
	print $i, "=", $short2macro{$i}, "\n";
    }
}

################################################################
### Output the Makefile
################################################################

if ($makefile)
{
    if ($verbose)
    {
	print "Writing makefile \"$makefile\"...\n";
    }
    open (MAKEFILE, ">" . $makefile)
	|| die("$PROGNAME: ERROR - failed to open $makefile for output\n");
    select MAKEFILE;
    print "### Makefile fragment created by $PROGNAME - do not edit ###\n\n";

    print "all_srcs = \\\n\t", join(" \\\n\t", @all_srcs), "\n";
    print "lib_srcs = \\\n\t", join(" \\\n\t", @lib_srcs), "\n";
    print "lib_objs = \\\n\t", join(" \\\n\t", @lib_objs), "\n";
    print "prelist = ", join(",", @prelist), "\n";
    print "\n";

    foreach $obj (@lib_objs)
    {
	local($src,$suffix,$func);

	$src = $obj2src{$obj};
	$src =~ /[.]([^.]*)$/;
	$suffix = $1;
	$obj =~ /([^.]*)\.o/;
	$func = $1;
        print $obj, ' : $(srcdir)/', $obj2src{$obj}, "\n";
        print "\t", '$(compile.', $suffix, ") -c -DDEFN_$func ",
		'$(srcdir)/', "$src -o $obj\n\n";
    }
    print "\n";

    print "### End of makefile fragment created by $PROGNAME ###\n";
    select STDOUT;
}

################################################################
### Output the header file
################################################################

if ($hdrfile)
{
    if ($verbose)
    {
	print "Writing header file \"$hdrfile\"...\n";
    }
    open (HEADER, ">" . $hdrfile)
	|| die("$progname: ERROR - failed to open $target_headerfile for output\n");
    select HEADER;

    print <<"END";
/*** Header created by $PROGNAME - do not edit ***/

#ifdef __cplusplus
extern "C" {
#endif

END
    foreach $long (@long_names)
    {
	print $long2proto{$long};
    }
    print <<END;

#ifdef __cplusplus
}
#endif

END

    foreach $short (keys(%short2macro))
    {
	print "#define $short $short2macro{$short}\n";
    }
    select STDOUT;
}

if ($verbose)
{
    print "Finished.\n";
}
exit 0;
