/*
   $Header: /u/spert/src/intvec/testsuite/RCS/scalar_i16i32_i16_test.c,v 1.1 1996/01/04 20:07:43 davidj Exp $

   Test scalar 16bit/32 bit binary ops
*/

#include <stddef.h>
#include <stdlib.h>

#include "intvec_test.h"

typedef intv_int16_t (*rtn_i16i32_i16_t)(intv_int16_t a, intv_int32_t b);

void
multitest(rtn_i16i32_i16_t rtn, size_t len, const intv_int16_t* av,
	  const intv_int32_t* bv, const intv_int16_t* corv)
{
    size_t i;
    intv_int16_t a;
    intv_int32_t b;
    intv_int16_t res;
    intv_int16_t correct;

    for (i=0; i<len; i++)
    {
	a = *av++;
	b = *bv++;
	correct = *corv++;
	res = rtn(a, b);
	if (res!=correct)
	{
	    rtst_failed("a=%i b=%i res=%i correct=%i", a, b, res, correct);
	}
    }
}

void
test_srt()
{
    const intv_int16_t val[] = {
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 0, 0, 1, 1, 1, -1, -1, -1, 23, 23, 23,
	INTV_INT16_MAX, INTV_INT16_MAX, INTV_INT16_MAX,
	INTV_INT16_MIN, INTV_INT16_MIN, INTV_INT16_MIN
    };
    const intv_int32_t shift[] = {
	2, 2, 2, 2, 2, 2, 2, 2,
	0, 1, 15, 0, 1, 15, 0, 1, 15, 0, 1, 15,
	0, 1, 15,
	0, 1, 15
    };
    const intv_int16_t correct[] = {
	0, 0, 0, 0, 1, 1, 1, 1,
	0, 0, 0, 1, 0, 0, -1, -1, -1, 23, 11, 0,
	INTV_INT16_MAX, INTV_INT16_MAX/2, 0,
	INTV_INT16_MIN, INTV_INT16_MIN/2, -1
    };

    rtst_start("srt_i16i32_i16");
    multitest(srt_i16i32_i16, sizeof(val)/sizeof(val[0]), val, shift, correct);
    rtst_passed();
}

void
test_srr()
{
    const intv_int16_t val[] = {
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 0, 0, 1, 1, 1, -1, -1, -1, 23, 23, 23,
	INTV_INT16_MAX, INTV_INT16_MAX, INTV_INT16_MAX,
	INTV_INT16_MIN, INTV_INT16_MIN, INTV_INT16_MIN
    };
    const intv_int32_t shift[] = {
	2, 2, 2, 2, 2, 2, 2, 2,
	0, 1, 15, 0, 1, 15, 0, 1, 15, 0, 1, 15,
	0, 1, 15,
	0, 1, 15
    };
    const intv_int16_t correct[] = {
	0, 0, 0, 1, 1, 1, 2, 2,
	0, 0, 0, 1, 0, 0, -1, 0, 0, 23, 12, 0,
	INTV_INT16_MAX, INTV_INT16_MAX/2+1, 1,
	INTV_INT16_MIN, INTV_INT16_MIN/2, -1
    };

    rtst_start("srr_i16i32_i16");
    multitest(srr_i16i32_i16, sizeof(val)/sizeof(val[0]), val, shift, correct);
    rtst_passed();
}

void
test_sll()
{
    const intv_int16_t val[] = {
	0x1ffe, 0x1fff, 0x2000, 0x2001,
	0, 0, 0, 1, 1, 1, -1, -1, -1, 23, 23, 23,
	INTV_INT16_MAX, INTV_INT16_MAX, INTV_INT16_MAX,
	INTV_INT16_MIN, INTV_INT16_MIN, INTV_INT16_MIN
    };
    const intv_int32_t shift[] = {
	2, 2, 2, 2,
	0, 1, 15, 0, 1, 15, 0, 1, 15, 0, 1, 15,
	0, 1, 15,
	0, 1, 15
    };
    const intv_int16_t correct[] = {
	0x7ff8, 0x7ffc, 0x8000, 0x8004,
	0, 0, 0, 1, 2, INTV_INT16_MIN, -1, -2, INTV_INT16_MIN, 23, 46, INTV_INT16_MIN,
	INTV_INT16_MAX, -2, INTV_INT16_MIN,
	INTV_INT16_MIN, 0, 0
    };

    rtst_start("sll_i16i32_i16");
    multitest(sll_i16i32_i16, sizeof(val)/sizeof(val[0]), val, shift, correct);
    rtst_passed();
}

void
test_sls()
{
    const intv_int16_t val[] = {
	0x1ffe, 0x1fff, 0x2000, 0x2001,
	0, 0, 0, 1, 1, 1, -1, -1, -1, 23, 23, 23,
	INTV_INT16_MAX, INTV_INT16_MAX, INTV_INT16_MAX,
	INTV_INT16_MIN, INTV_INT16_MIN, INTV_INT16_MIN
    };
    const intv_int32_t shift[] = {
	2, 2, 2, 2,
	0, 1, 15, 0, 1, 15, 0, 1, 15, 0, 1, 15,
	0, 1, 15,
	0, 1, 15
    };
    const intv_int16_t correct[] = {
	0x7ff8, 0x7ffc, 0x7fff, 0x7fff,
	0, 0, 0, 1, 2, INTV_INT16_MAX, -1, -2, INTV_INT16_MIN, 23, 46, INTV_INT16_MAX,
	INTV_INT16_MAX, INTV_INT16_MAX, INTV_INT16_MAX,
	INTV_INT16_MIN, INTV_INT16_MIN, INTV_INT16_MIN
    };

    rtst_start("sls_i16i32_i16");
    multitest(sls_i16i32_i16, sizeof(val)/sizeof(val[0]), val, shift, correct);
    rtst_passed();
}

int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    test_srt();
    test_srr();
    test_sll();
    test_sls();
    rtst_exit();

    return 0;
}
