/*
   $Header: /u/spert/src/intvec/testsuite/RCS/scalar_i32i32_i32_test.c,v 1.3 1996/01/04 20:08:27 davidj Exp $

   Test scalar 32 bit binary ops
*/

#include <stddef.h>
#include <stdlib.h>

#include "intvec_test.h"

typedef intv_int32_t (*rtn_i32i32_i32_t)(intv_int32_t a, intv_int32_t b);

void
multitest(rtn_i32i32_i32_t rtn, size_t len, const intv_int32_t* av,
	  const intv_int32_t* bv, const intv_int32_t* corv)
{
    size_t i;
    intv_int32_t a, b;
    intv_int32_t res;
    intv_int32_t correct;

    for (i=0; i<len; i++)
    {
	a = *av++;
	b = *bv++;
	correct = *corv++;
	res = rtn(a, b);
	if (res!=correct)
	{
	    rtst_failed("a=%i b=%i res=%i correct=%i", a, b, res, correct);
	}
    }
}

void
test_satadd()
{
    const intv_int32_t a[] = { 0, -1, 1, -1, 1, 0x7fffffff, -0x7fffffff-1,
				   123 };
    const intv_int32_t b[] = { 0, -1, -1, 1, 1, 1, -1, -456 };
    const intv_int32_t correct[] = { 0, -2, 0, 0, 2, 0x7fffffff, 0x80000000,
					 -333 };

    rtst_start("satadd_i32i32_i32");
    multitest(satadd_i32i32_i32, sizeof(a)/sizeof(a[0]), a, b, correct);
    rtst_passed();
}

void
test_satsub()
{
    const intv_int32_t a[] = { 0, -1, 1, -1, 1, INTV_INT32_MIN,
			       INTV_INT32_MAX, 0, 123 };
    const intv_int32_t b[] = { 0, -1, -1, 1, 1, 1, -1, INTV_INT32_MIN, -456 };
    const intv_int32_t correct[] = { 0, 0, 2, -2, 0, INTV_INT32_MIN,
				     INTV_INT32_MAX, INTV_INT32_MAX,579 };

    rtst_start("satsub_i32i32_i32");
    multitest(satsub_i32i32_i32, sizeof(a)/sizeof(a[0]), a, b, correct);
    rtst_passed();
}


void
test_srt()
{
    const intv_int32_t val[] = {
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 0, 0, 1, 1, 1, -1, -1, -1, 23, 23, 23,
	INTV_INT32_MAX, INTV_INT32_MAX, INTV_INT32_MAX,
	INTV_INT32_MIN, INTV_INT32_MIN, INTV_INT32_MIN
    };
    const intv_int32_t shift[] = {
	2, 2, 2, 2, 2, 2, 2, 2,
	0, 1, 31, 0, 1, 31, 0, 1, 31, 0, 1, 31,
	0, 1, 31,
	0, 1, 31
    };
    const intv_int32_t correct[] = {
	0, 0, 0, 0, 1, 1, 1, 1,
	0, 0, 0, 1, 0, 0, -1, -1, -1, 23, 11, 0,
	INTV_INT32_MAX, INTV_INT32_MAX/2, 0,
	INTV_INT32_MIN, INTV_INT32_MIN/2, -1
    };

    rtst_start("srt_i32i32_i32");
    multitest(srt_i32i32_i32, sizeof(val)/sizeof(val[0]), val, shift, correct);
    rtst_passed();
}

void
test_srr()
{
    const intv_int32_t val[] = {
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 0, 0, 1, 1, 1, -1, -1, -1, 23, 23, 23,
	INTV_INT32_MAX, INTV_INT32_MAX, INTV_INT32_MAX,
	INTV_INT32_MIN, INTV_INT32_MIN, INTV_INT32_MIN
    };
    const intv_int32_t shift[] = {
	2, 2, 2, 2, 2, 2, 2, 2,
	0, 1, 31, 0, 1, 31, 0, 1, 31, 0, 1, 31,
	0, 1, 31,
	0, 1, 31
    };
    const intv_int32_t correct[] = {
	0, 0, 0, 1, 1, 1, 2, 2,
	0, 0, 0, 1, 0, 0, -1, 0, 0, 23, 12, 0,
	INTV_INT32_MAX, INTV_INT32_MAX/2+1, 1,
	INTV_INT32_MIN, INTV_INT32_MIN/2, -1
    };

    rtst_start("srr_i32i32_i32");
    multitest(srr_i32i32_i32, sizeof(val)/sizeof(val[0]), val, shift, correct);
    rtst_passed();
}

void
test_sll()
{
    const intv_int32_t val[] = {
	0x1ffffffe, 0x1fffffff, 0x20000000, 0x20000001,
	0, 0, 0, 1, 1, 1, -1, -1, -1, 23, 23, 23,
	INTV_INT32_MAX, INTV_INT32_MAX, INTV_INT32_MAX,
	INTV_INT32_MIN, INTV_INT32_MIN, INTV_INT32_MIN
    };
    const intv_int32_t shift[] = {
	2, 2, 2, 2,
	0, 1, 31, 0, 1, 31, 0, 1, 31, 0, 1, 31,
	0, 1, 31,
	0, 1, 31
    };
    const intv_int32_t correct[] = {
	0x7ffffff8, 0x7ffffffc, 0x80000000, 0x80000004,
	0, 0, 0, 1, 2, INTV_INT32_MIN, -1, -2, INTV_INT32_MIN, 23, 46, INTV_INT32_MIN,
	INTV_INT32_MAX, -2, INTV_INT32_MIN,
	INTV_INT32_MIN, 0, 0
    };

    rtst_start("sll_i32i32_i32");
    multitest(sll_i32i32_i32, sizeof(val)/sizeof(val[0]), val, shift, correct);
    rtst_passed();
}

void
test_sls()
{
    const intv_int32_t val[] = {
	0x1ffffffe, 0x1fffffff, 0x20000000, 0x20000001,
	0, 0, 0, 1, 1, 1, -1, -1, -1, 23, 23, 23,
	INTV_INT32_MAX, INTV_INT32_MAX, INTV_INT32_MAX,
	INTV_INT32_MIN, INTV_INT32_MIN, INTV_INT32_MIN
    };
    const intv_int32_t shift[] = {
	2, 2, 2, 2,
	0, 1, 31, 0, 1, 31, 0, 1, 31, 0, 1, 31,
	0, 1, 31,
	0, 1, 31
    };
    const intv_int32_t correct[] = {
	0x7ffffff8, 0x7ffffffc, 0x7fffffff, 0x7fffffff,
	0, 0, 0, 1, 2, INTV_INT32_MAX, -1, -2, INTV_INT32_MIN, 23, 46, INTV_INT32_MAX,
	INTV_INT32_MAX, INTV_INT32_MAX, INTV_INT32_MAX,
	INTV_INT32_MIN, INTV_INT32_MIN, INTV_INT32_MIN
    };

    rtst_start("sls_i32i32_i32");
    multitest(sls_i32i32_i32, sizeof(val)/sizeof(val[0]), val, shift, correct);
    rtst_passed();
}

int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    test_satadd();
    test_satsub();
    test_srt();
    test_srr();
    test_sll();
    test_sls();
    rtst_exit();

    return 0;
}
