/*
   $Header: /u/spert/src/intvec/testsuite/RCS/vecred_vi32_i32_test.c,v 1.1 1995/10/10 19:24:50 davidj Exp $

   Test vector reduction routines of the form op_vfvf_.
*/

#include <stddef.h>
#include <stdlib.h>

#include "intvec_test.h"


typedef intv_int32_t (*rtn_t)(size_t len, const intv_int32_t* a);

/* Performs one test given input values and expected output. */

void
onetest_normal(rtn_t rtn, size_t len,
	       const intv_int32_t* a, intv_int32_t correct)
{
    /* Create padded copies of inputs. */
    intv_int32_t* c_a = rtst_padvec_clone_vi32(len, a);
    intv_int32_t res;

    /* Run routine. */
    res = rtn(len, c_a);

    /* Check result. */
    rtst_assert(res == correct);

    /* Check input not corrupted. */
    rtst_checkeq_vi32vi32(len, a, c_a);

    /* Free and check padding. */
    rtst_padvec_del_vi32(c_a);
}

void
sum1(rtn_t rtn)
{
    const intv_int32_t a1[] = { 7 };
    const intv_int32_t correct1 = 7;
    const intv_int32_t a2[] = { 1, 2, 3, -4, 0x70000000 };
    const intv_int32_t correct2 = 0x70000002;

    onetest_normal(rtn, sizeof(a1)/sizeof(intv_int32_t), a1, correct1);
    onetest_normal(rtn, sizeof(a2)/sizeof(intv_int32_t), a2, correct2);

}

/* Compare the results of two vector routines */

void
random_test(rtn_t nv_rtn, rtn_t test_rtn)
{
    const int numtests = rtst_numtests;
    const size_t maxlen = rtst_sizetests;
    intv_int32_t *a = rtst_padvec_new_vi32(maxlen);
    intv_int32_t res;

    int i;
    for (i=0; i<numtests; i++)
    {
        size_t len = (size_t) rtst_urand_i32i32_i32(1, maxlen);

        rtst_badrand_vi32(len, a);
        res = nv_rtn(len, a);

        rtst_log("len=%d\n", len);

        onetest_normal(test_rtn, len, a, res);
    }

    rtst_padvec_del_vi32(a);
}

void
test_sum(void)
{
    rtst_start("sum_vi32_i32");
    rtst_log("testing nv.\n");
    sum1(nv_sum_vi32_i32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    sum1(tor_sum_vi32_i32);
    random_test(nv_sum_vi32_i32, tor_sum_vi32_i32);
#endif
    rtst_passed();
}


int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    test_sum();

    rtst_exit();

    return 0;
}
