/*
** $Id: binop_vi32vi32_vi32.S,v 1.2 1996/01/03 20:56:18 davidj Exp $
**
** A general 32 bit vector/vector binary operation routine, plus
** some specific instantiations
** davidj - Sat Oct 15 18:50:22 1994
*/

#include "tor/tordef.h"

/****************************************************************/

#ifdef DECL_tor_add_vi32vi32_vi32
void tor_add_vi32vi32_vi32(size_t len, const intv_int32_t* a,
			   const intv_int32_t* b, intv_int32_t* res);
#endif

#ifdef DEFN_tor_add_vi32vi32_vi32
#define RTNNAME tor_add_vi32vi32_vi32
#define OP_VV addu.vv
#endif

/****************************************************************/


#ifdef DECL_tor_satadd_vi32vi32_vi32
void tor_satadd_vi32vi32_vi32(size_t len, const intv_int32_t* a,
			      const intv_int32_t* b, intv_int32_t* res);
#endif

#ifdef DEFN_tor_satadd_vi32vi32_vi32
#define RTNNAME tor_satadd_vi32vi32_vi32
#define OP_VV sadd.vv
#endif

/****************************************************************/

#ifdef DECL_tor_sub_vi32vi32_vi32
void tor_sub_vi32vi32_vi32(size_t len, const intv_int32_t* a,
			   const intv_int32_t* b, intv_int32_t* res);
#endif

#ifdef DEFN_tor_sub_vi32vi32_vi32
#define RTNNAME tor_sub_vi32vi32_vi32
#define OP_VV subu.vv
#endif

/****************************************************************/

#ifdef DECL_tor_satsub_vi32vi32_vi32
void tor_satsub_vi32vi32_vi32(size_t len, const intv_int32_t* a,
			   const intv_int32_t* b, intv_int32_t* res);
#endif

#ifdef DEFN_tor_satsub_vi32vi32_vi32
#define RTNNAME tor_satsub_vi32vi32_vi32
#define OP_VV ssub.vv
#endif

/****************************************************************/

/* This routine requires 2 macros defines
**
** RTNNAME - the name this routine will be known by
** OP_VV - the binary vector-vector op to perform
*/ 

#ifndef RTNNAME
	.err
#endif
#ifndef OP_VV
	.err
#endif

#define	a_v		vv1
#define b_v		vv2
#define res_v		vv3

#define len_r           a0
#define aptr_r		a1
#define bptr_r		a2
#define resptr_r	a3

#define fourvlr_r	t0
#define done_r		t1

	.globl	RTNNAME


	.ent RTNNAME
RTNNAME:
 # Check the length - jump forward if > 32

        .set    noreorder
        bgtu    len_r, MAXVLEN, morethan32
        ctvu    len_r, vlr              # Assume len <= 32
        .set    reorder

 # The short case follows directly for better cache performance

        lw.v	a_v, aptr_r
        lw.v	b_v, bptr_r
	OP_VV	res_v, a_v, b_v
	sw.v	res_v, resptr_r
	j	ra

        .align  4
morethan32:
        li      v0, MAXVLEN
        ctvu    v0, vlr
        lw.v    a_v, aptr_r
	li	fourvlr_r, MAXVLEN*4
	addu	aptr_r, MAXVLEN*4
	lwai.v	b_v, bptr_r, fourvlr_r
	OP_VV	res_v, a_v, b_v

loop:
	lwai.v	a_v, aptr_r, fourvlr_r
        subu    len_r, MAXVLEN
 	sltu	done_r, len_r, MAXVLEN+1
	lwai.v	b_v, bptr_r, fourvlr_r
	swai.v	res_v, resptr_r, fourvlr_r
	.set	noreorder
	beqz	done_r, loop
	OP_VV	res_v, a_v, b_v
	.set	reorder

lastbit:
	ctvu	len_r, vlr
	sw.v	res_v, resptr_r

	j	ra

	.end RTNNAME



