/*
** $Id: copy.S,v 1.4 1996/10/23 02:36:21 davidj Exp $
** 
** General Torrent copying routines
**
** davidj - Tue Aug 29 18:11:43 1995
*/

#include "machine/tordef.h"

/****************************************************************/
/* Set an 8 bit vector to given value */
/* Hand editted from Krstes tor_memset.S v 1.4
** davidj Mon Oct 21 21:52:17 1996
*/

#ifdef DECL_tor_copy_i8_vi8
void tor_copy_i8_vi8(size_t size, intv_int8_t val, intv_int8_t* vec);
#endif

#ifdef DEFN_tor_copy_i8_vi8

#define len a0
#define char_val a1
#define dest_ptr a2
#define ret_val v0

#define vec_len t9
#define n_bytes t8

	.globl	tor_copy_i8_vi8
        .align 4                        # Entry point aligned in cache.
        .ent tor_copy_i8_vi8
tor_copy_i8_vi8:
        .set noreorder

        sltiu t0, len, MAXVLEN+1        # <=MAXVLEN bytes?
        bnez char_val, nonzero_set      # Common case writes zeros.
        move ret_val, dest_ptr          # Save return value in delay slot.

        beqz t0, big_set                # Jump to bigger copy code.
        li vec_len, MAXVLEN             # This is used in big copy code.

        ctvu len, vlr                   # Move length to vlr.
        j ra                            # Return from routine.
        sb.v vv0, dest_ptr              # Store zeroed bytes to destination.
        .set reorder

 ### len > MAXVLEN.  First store 32 bytes, then move destination pointer up to
 ### a 16-byte aligned boundary. This can write some locations twice
 ### but is quicker than working out the correct length originally.
        .set noreorder
        nop                             # Tweak cache alignment to save one
                                        # cache miss on long non-zero sets.
big_set:
        ctvu vec_len, vlr
        mov.vv vv1, vv0                 # Clear vv1.
        sb.v vv0, dest_ptr              # Write 32 zero bytes.

finish_big_set:
        andi t0, dest_ptr, ALIGNMASK    # Get misalignment offset.
        subu n_bytes, vec_len, t0       # Calculate (32 - misalignment).
        addu dest_ptr, n_bytes          # Advance pointer 32 - misaligment.
        subu len, n_bytes               # Decrement length by bytes stored.
        .set reorder        

 ### Now know that destination aligned to 16 byte boundary.

word_loop:
        .set noreorder
        sltiu t0, len, MAXVLEN*4
        bnezl t0, 1f
        srl vec_len, len, 2             # Adjust vector length on last loop.
1:
        ctvu vec_len, vlr
        sll n_bytes, vec_len, 2         # Number of bytes to be stored.
        swai.v vv1, dest_ptr, n_bytes
        beqz t0, word_loop
        subu len, n_bytes
        
        ctvu len, vlr
        j ra
        sb.v vv1, dest_ptr              # Finish off any remaining bytes.
        .set reorder

        .align 4
nonzero_set:
        .set noreorder
        li vec_len, MAXVLEN
        ctvu vec_len, vlr
        beqz t0, big_nonzero_set
        mov.vs vv1, char_val            # Fill vector register with chars.

        ctvu len, vlr
        j ra
        sb.v vv1, dest_ptr
        .set reorder

 ### Writs out 32 bytes of char to allow subsequent aligning, and sets up vv1
 ### to hold words full of 4 copies of char, before jumping to common word
 ### loop code.
        
        .align 4        
big_nonzero_set:
        .set noreorder
        andi char_val, 0xff             # Mask off any high bits.
        sll t0, char_val, 8
        or char_val, t0
        sb.v vv1, dest_ptr              # Store 32 chars out.
        sll t0, char_val, 16
        or char_val, t0                 # Now has 4 copies of char.

        b finish_big_set
        mov.vs vv1, char_val            # Fill vv1 words with chars.
        .set reorder

	.end	tor_copy_i8_vi8

#endif

/****************************************************************/
/* Set a 16 bit vector to given value */

#ifdef DECL_tor_copy_i16_vi16
void tor_copy_i16_vi16(size_t size, intv_int16_t val, intv_int16_t* vec);
#endif

#ifdef DEFN_tor_copy_i16_vi16

	.globl	tor_copy_i16_vi16

#define	valvec_v	vv1

#define	len_r		a0
#define val_r		a1
#define resptr_r	a2

#define ai_r		t0
#define maxvlr_r	t1
#define offset_r	t2

/*
We handle three possible situations...

i) len_r<=MAXVLEN

In this case, fall straight through to two instruction routine!

ii) MAXVLEN < len_r <= MAXVLEN*2

Again a pretty obvious solution - do 32 first, then the rest

iii) len_r > MAXVLEN*2

In this case, we do a MAXVLEN write first, but maybe increment resptr_r
slightly less to align  our writes to ALIGNLEN boundaries for maximim
efficiency
*/


	.text

	.ent	tor_copy_i16_vi16
tor_copy_i16_vi16:

 # Check the length - jump forward if > MAXVLEN

	.set	noreorder
	bgtu	len_r, MAXVLEN, morethanmax
	ctvu	len_r, vlr		# Assume len <= MAXVLEN
	.set	reorder

 # The short case follows directly for better cache performance

	mov.vs	valvec_v, val_r
	sh.v	valvec_v, resptr_r
	j	ra

 # This is the len_r>MAXVLEN && len_r<=MAXVLEN*2 case
	
morethanmax:
	li	maxvlr_r, MAXVLEN
	ctvu	maxvlr_r, vlr
	mov.vs	valvec_v, val_r
	li	ai_r, MAXVLEN*2
	shai.v	valvec_v, resptr_r, ai_r
	.set	noreorder
	bgtu	len_r, ai_r, needs_align
	subu	len_r, MAXVLEN
	.set	reorder
	ctvu	len_r, vlr
	sh.v	valvec_v, resptr_r
	j	ra

 # This is the len_r>MAXVLEN*2 case - we align the destination

needs_align:
	# These next 3 lines do the alignment

	and	offset_r, resptr_r, ALIGNMASK
	.set	noreorder
	beqzl	offset_r, enter_loop	# Skip this bit if we dont need it
	subu	len_r, MAXVLEN		# (Pulled from start of loop)
	.set	reorder
	subu	resptr_r, offset_r
	srl	offset_r, 1		# Offset is in bytes, length in els.
	addu	len_r, offset_r

loop:	
	subu	len_r, MAXVLEN
	.set	noreorder
enter_loop:
	bgt	len_r, MAXVLEN, loop
	shai.v	valvec_v, resptr_r, ai_r
	.set	reorder

	ctvu	len_r, vlr
	sh.v	valvec_v, resptr_r
	j	ra

	.end	tor_copy_i16_vi16

#endif

/****************************************************************/

/* Set a 32 bit vector to a given value */

#ifdef DECL_tor_copy_i32_vi32
void tor_copy_i32_vi32(size_t size, intv_int32_t val, intv_int32_t* vec);
#endif

#ifdef DEFN_tor_copy_i32_vi32

        .globl  tor_copy_i32_vi32


#define	valvec_v        vv1

#define	len_r           a0
#define val_r           a1
#define resptr_r        a2

#define ai_r            t0
#define maxvlr_r        t1
#define offset_r	t2


        .text

        .ent    tor_copy_i32_vi32
tor_copy_i32_vi32:
 # Check the length - jump forward if > MAXVLEN

	.set	noreorder
	bgtu	len_r, MAXVLEN, morethanmax
	ctvu	len_r, vlr		# Assume len <= MAXVLEN
	.set	reorder

 # The short case follows directly for better cache performance

	mov.vs	valvec_v, val_r
	sw.v	valvec_v, resptr_r
	j	ra

morethanmax:

 # Start some stores going

	li	maxvlr_r, MAXVLEN
	ctvu	maxvlr_r, vlr
	mov.vs	valvec_v, val_r
	sw.v	valvec_v, resptr_r
	li	ai_r, MAXVLEN*4

 # Here we align our writes

	and	offset_r, resptr_r, ALIGNMASK
	.set	noreorder
	beqz	offset_r, loop		# If no alignment needed, skip this bit
	addu	resptr_r, ai_r		# (In delay slot)
	.set	reorder

	subu	resptr_r, offset_r
	srl	offset_r, 2		# Offset is in bytes, length in eles.
	addu	len_r, offset_r

 # Loop around doing the actual writes

loop:	
	subu	len_r, MAXVLEN
	.set	noreorder
	bgtl	len_r, MAXVLEN, loop
	swai.v	valvec_v, resptr_r, ai_r
	.set	reorder

 # Do any left over bits at the end

	ctvu	len_r, vlr
	sw.v	valvec_v, resptr_r
	j	ra

        .end    tor_copy_i32_vi32


#endif

/****************************************************************/
/* Copy an 8 bit vector */
/*
** Source from Krste - a trivially modified version of
** /u/krste/vlsi/t0/src/mem/tor_memcpymov.S version 1.10
*/

#ifdef DECL_tor_copy_vi8_vi8
void tor_copy_vi8_vi8(size_t size, const intv_int8_t* from,	
		      intv_int8_t* to);
#endif


#ifdef DEFN_tor_copy_vi8_vi8

	.globl	tor_copy_vi8_vi8

#define len a0
#define src_ptr a1
#define dest_ptr a2
#define ret_val v0

#define vec_len t9
#define n_bytes t8

 ### Small copies (<=32 bytes) are handled in two I-cache lines.

        .globl tor_copy_vi8_vi8

        .align 4                        # Entry point aligned in cache.
        .ent tor_copy_vi8_vi8
tor_copy_vi8_vi8:
        .set noreorder

        sltiu t0, len, 33               # <=32 bytes?
        beqz t0, big_copy               # Jump to bigger copy code.
        move ret_val, dest_ptr          # Save return value in delay slot.
        ctvu len, vlr                   # Move length to vlr.

        lbu.v vv1, src_ptr              # Load up to 32 bytes from source.
        jr ra                           # Return from routine.
        sb.v vv1, dest_ptr              # Store up to 32 bytes to destination.
        .set reorder

        .align 4                        # Save cache miss by aligning.
        .set noreorder
big_copy:
 ### len > 32.
 ### First transfer enough bytes to align source vector to 16 byte boundary.
 ### Actually transfers 16 bytes more than strictly necessary, since otherwise
 ### memory port would be idle while making decisions.  Transfers 32
 ### bytes if source pointer was already 16 byte aligned.

        li vec_len, 32
        ctvu vec_len, vlr
        lbu.v vv1, src_ptr              # Start load.
        andi t0, src_ptr, 15            # Get misalignment offset.

        subu n_bytes, vec_len, t0       # Bytes to store back (17 <= nb <= 32).
        ctvu n_bytes, vlr
        sbai.v vv1, dest_ptr, n_bytes   # Increment dest ptr.
        addu src_ptr, n_bytes           # Increment src ptr.
        
 ### Now know that source aligned to 16 byte boundary.

 ### Check to see if can use word copies.        

        and t1, dest_ptr, 3             # Is destination word aligned?
        bnez t1, try_halfword           # If not, check for halfword aligned.
        subu len, n_bytes               # Subtract bytes remaining in delay.
        .set reorder

        .set noreorder
 ### Source 16 byte aligned, destination word aligned, len has bytes remaining
 ### to be copied.
word_copy_loop:
        sltiu t0, len, 129              # Last loop?
        bnezl t0, 1f
        srl vec_len, len, 2             # Only change on last loop.
1:
        ctvu vec_len, vlr
        sll n_bytes, vec_len, 2         # Get bytes transferred this loop.

        lwai.v vv1, src_ptr, n_bytes
        subu len, n_bytes
        beqz t0, word_copy_loop         # Only continue if not last loop.
        swai.v vv1, dest_ptr, n_bytes

word_tail:
        ctvu len, vlr                   # Know that len < 4.
        lbu.v vv1, src_ptr
        j ra
        sb.v vv1, dest_ptr              # Store in delay slot.
        .set reorder

 ### Check to see if can do halfword copies.
        .set noreorder
try_halfword:
        and t1, dest_ptr, 1             # Check if halfword aligned.
        bnez t1, byte_copy_loop         # Have to do byte copies.
        
 ### Source 16 byte aligned, destination word halfword aligned, len has bytes
 ### remaining to be copied.
halfword_copy:
        sltiu t0, len, 65               # Last loop?
        bnezl t0, 1f
        srl vec_len, len, 1             # Get last number of halfwords to copy.
1:
        ctvu vec_len, vlr
        sll n_bytes, vec_len, 1         # Get bytes transferred this loop.

        lhuai.v vv1, src_ptr, n_bytes   # Load data.
        bnez t0, halfword_tail          # Only continue if not last loop.
        subu len, n_bytes

        sltiu t0, len, 65               # Last loop?

halfword_copy_loop:
        shai.v vv1, dest_ptr, n_bytes   # Store is never 16B aligned.

        bnezl t0, 1f
        srl vec_len, len, 1             # Get last number of halfwords to copy.
1:
        ctvu vec_len, vlr
        sll n_bytes, vec_len, 1         # Get bytes transferred this loop.

        lhuai.v vv1, src_ptr, n_bytes   # Load data.
        subu len, n_bytes
        beqz t0, halfword_copy_loop     # Only continue if not last loop.
        sltiu t0, len, 65

halfword_tail:
        shai.v vv1, dest_ptr, n_bytes   # Store is never 16B aligned.
        ctvu len, vlr                   # Know that len < 2.
        lbu.v vv1, src_ptr
        j ra
        sb.v vv1, dest_ptr              # Store in delay slot.
        .set reorder

        .align 4                        # Save cache miss.
        .set noreorder
 ### Source 16 byte aligned, but destination not even halfword aligned.
byte_copy_loop:
        sltiu t0, len, 33               # Last loop?
        bnezl t0, 1f
        move vec_len, len               # Only changes on last loop.
1:
        ctvu vec_len, vlr
        lbuai.v vv1, src_ptr, vec_len
        subu len, vec_len
        beqz t0, byte_copy_loop
        sbai.v vv1, dest_ptr, vec_len   # Never 16B aligned.

byte_tail:
        j ra
        nop
        .set reorder

        .end tor_copy_vi8_vi8

#endif

/****************************************************************/
/* Copy a 16 bit vector */

#ifdef DECL_tor_copy_vi16_vi16
void tor_copy_vi16_vi16(size_t size, const intv_int16_t* from,
			intv_int16_t* to);

#endif

#ifdef DEFN_tor_copy_vi16_vi16

	.globl	tor_copy_vi16_vi16


#define	val_v		vv1

#define	len_r		a0
#define fromptr_r	a1
#define toptr_r		a2
#define ai_r		t0
#define maxvlr_r	t1
#define fromoff_r	t2
#define tooff_r		t3

	.text

/*
There are two cases handled...

i) len_r <= MAXVLEN

Drop straight through to code optimized for this case.

ii) len_r > MAXVLEN

If either the source or destination vectors are aligned to the memory port,
just carry on, otherwise align to the source vector.

*/

	.ent	tor_copy_vi16_vi16
tor_copy_vi16_vi16:

 # Check the length - jump forward if > MAXVLEN

	.set	noreorder
	bgtu	len_r, MAXVLEN, morethanmax
	ctvu	len_r, vlr		# Assume len <= MAXVLEN
	.set	reorder

 # The short case follows directly for better cache performance

 	lh.v	val_v, fromptr_r
	sh.v	val_v, toptr_r
	j	ra

 # This is the len_r>MAXVLEN case
	
morethanmax:

 # Start our first load

	li	maxvlr_r, MAXVLEN
	ctvu	maxvlr_r, vlr
	lh.v	val_v, fromptr_r
	and	fromoff_r, fromptr_r, ALIGNMASK
	li	ai_r, MAXVLEN*2
	addu	fromptr_r, ai_r
	shai.v	val_v, toptr_r, ai_r
	subu	len_r, MAXVLEN

 # Check for source misalignment

	beqz	fromoff_r, loop

 # And destination misalignment 

	and	tooff_r, toptr_r, ALIGNMASK
	beqz	tooff_r, loop

 # Here we ignore some of the previous copy to align to the source
	
	subu	fromptr_r, fromoff_r
 	subu	toptr_r, fromoff_r
	srl	fromoff_r, 1		# Convert to length (in elements)
	addu	len_r, fromoff_r

loop:
	lhai.v	val_v, fromptr_r, ai_r
	bltu	len_r, MAXVLEN, done	
	shai.v	val_v, toptr_r, ai_r	
	subu	len_r, MAXVLEN
	b	loop

 # Do the leftover elements here
done:
	ctvu	len_r, vlr
 	sh.v	val_v, toptr_r

	j	ra

	.end	tor_copy_vi16_vi16

#endif

/****************************************************************/
/* Copy a 32 bit vector to another 32 bit vector */

#ifdef DECL_tor_copy_vi32_vi32
void tor_copy_vi32_vi32(size_t size, const intv_int32_t* from,
			intv_int32_t* to);
#endif

#ifdef DEFN_tor_copy_vi32_vi32

	.globl	tor_copy_vi32_vi32

#define	val_v		vv1 	/* Used to transfer values */

#define	len_r		a0	/* Length left to copy */
#define fromptr_r	a1	/* Where we copy from */
#define toptr_r		a2	/* Where we copy to */

#define ai_r		t0	/* Increment for from/to address */
#define maxvlr_r	t1	/* Holds VAXVLEN */
#define fromoff_r	t2	/* Offset of from ptr from natural aligment */
#define tooff_r		t3	/* Offset of to ptr from natural alignment */
#define fromoff_len_r	t4	/* fromoff_r converted to no. of elements */

	.text

/*
There are two cases handled...

i) len_r <= MAXVLEN

Drop straight through to code optimized for this case.

ii) len_r > MAXVLEN

If either the source or destination vectors are aligned to the memory port,
just carry on, otherwise align to the source vector.

*/

	.ent	tor_copy_vi32_vi32
tor_copy_vi32_vi32:

 # Check the length - jump forward if > MAXVLEN

	.set	noreorder
	bgtu	len_r, MAXVLEN, morethanmax
	ctvu	len_r, vlr		# Assume len <= MAXVLEN
	.set	reorder

 # The short case follows directly for better cache performance

 	lw.v	val_v, fromptr_r
	sw.v	val_v, toptr_r
	j	ra

 # This is the len_r>MAXVLEN case
	
morethanmax:

 # Start our first load

	li	maxvlr_r, MAXVLEN
	ctvu	maxvlr_r, vlr
	lw.v	val_v, fromptr_r
	and	fromoff_r, fromptr_r, ALIGNMASK
	and	tooff_r, toptr_r, ALIGNMASK
	srl	fromoff_len_r, fromoff_r, 2 # Convert to length (in elements)
	li	ai_r, MAXVLEN*4
	addu	fromptr_r, ai_r
	subu	len_r, MAXVLEN
	swai.v	val_v, toptr_r, ai_r

 # Check for source misalignment

	beqz	fromoff_r, loop

 # And destination misalignment 

	beqz	tooff_r, loop

 # Here we ignore some of the previous copy to align to the source
	
	subu	fromptr_r, fromoff_r
 	subu	toptr_r, fromoff_r
	addu	len_r, fromoff_len_r

loop:
	lwai.v	val_v, fromptr_r, ai_r
	bltu	len_r, MAXVLEN, done	
	swai.v	val_v, toptr_r, ai_r	
	subu	len_r, MAXVLEN
	b	loop

 # Do the leftover elements here
done:
	ctvu	len_r, vlr
 	sw.v	val_v, toptr_r

	j	ra

	.end	tor_copy_vi32_vi32

#endif

/****************************************************************/
/* Copy a 16 bit vector to a 32 bit one */

#ifdef DECL_tor_conv_vi16_vi32
void tor_conv_vi16_vi32(size_t len, const intv_int16_t* from, \
	intv_int32_t* to);
#endif

#ifdef DEFN_tor_conv_vi16_vi32

	.globl	tor_conv_vi16_vi32

#define	in_v            vv1

#define	len_r           a0
#define inptr_r         a1
#define resptr_r        a2
	
#define twicevlr_r      t0
#define fourvlr_r	t1
#define maxvlr_r        t2
#define done_r          t3


	.ent	tor_conv_vi16_vi32
tor_conv_vi16_vi32:
 # Check the length - jump forward if > 32

        .set    noreorder
        bgtu    len_r, MAXVLEN, morethan32
        ctvu    len_r, vlr              # Assume len <= 32
        .set    reorder

 # The short case (len<32) follows directly for better cache performance

	lh.v	in_v, inptr_r
	sw.v	in_v, resptr_r
	j	ra

morethan32:
        li      maxvlr_r, MAXVLEN
        ctvu    maxvlr_r, vlr

loop:
	lh.v	in_v, inptr_r
	add	inptr_r, 2*MAXVLEN
	sw.v	in_v, resptr_r
	add	resptr_r, 4*MAXVLEN
	subu	len_r, MAXVLEN
 	sltu	done_r, len_r, MAXVLEN+1
	beqz	done_r, loop

	ctvu	len_r, vlr
	lh.v	in_v, inptr_r
	sw.v	in_v, resptr_r
	j	ra

	.end	tor_conv_vi16_vi32

#endif
