// $Header: /n/lasagna/db/davidj/spert/src/quicknet/H-spert0/../RCS/QN_SentDB_Narrow.h,v 1.3 1996/03/01 03:37:40 davidj Exp $

#ifndef QN_SentDB_Narrow_h_INCLUDED
#define QN_SentDB_Narrow_h_INCLUDED

/* Must include the config.h file first */
#include <QN_config.h>
#include <assert.h>
#include <stddef.h>
#include <stdio.h>
#ifdef QN_HAVE_LIMITS_H
#include <limits.h>
#endif
#include "QN_types.h"
#include "QN_SentDB.h"

// QN_SentDB_Narrow is used to select various features, labels and
// targets from a QN_SentDB

class QN_SentDB_Narrow : public QN_SentDB
{
public:
    // Constructor.
    // a_db - the database we are subsetting
    // a_first_ftr - number of the first feature we want
    // a_n_ftrs - number of features we want (INT_MAX if all remaining)
    // buf_size - size of buffer (in sentences)
    // logf - stream to send logging information to
    QN_SentDB_Narrow(int dbg, QN_SentDB& a_db,
		     size_t a_first_ftr,
		     size_t a_n_ftrs,
		     size_t a_first_label = 0,
		     size_t a_num_labels = 1,
		     size_t a_first_tgt = 0,
		     size_t a_num_tgts = QN_SentDB::ALL_TARGETS,
		     size_t a_buf_size = 5);
    ~QN_SentDB_Narrow();

    // Return the total number of frames in the database
    size_t num_frames();
    // Return the total number of sentences in the database
    size_t num_sentences();
    // Return the number of frames in a range of sentences
    size_t num_frames_in_sents(size_t first_sent, size_t num_sents);
    // Return the number of reatures in a frame
    size_t num_features_in_frame();
    // Return the number of labels in a frame
    size_t num_labels_in_frame();
    // Return the number of targets in a frame
    size_t num_targets_in_frame();

    // Return the labels and features for a given sentence.
    // If ftrs is NULL, do not get features.
    // If labs is NULL, do not get labels.
    size_t get_sentences(size_t sentno, size_t num_sents,
			 QNUInt32* sentnos, QNUInt32* framenos,
			 float* ftrs, QNUInt32* labs, float* tgts);
private:
    int debug;			// Level of verbosity
    QN_SentDB& db;		// The database we are subsetting
    size_t first_ftr;		// The first feature we supply
    size_t n_ftrs;		// The number of features we supply
    size_t first_lbl;		// The first label we supply
    size_t n_lbls;		// The number of labels we supply
    size_t first_tgt;		// The first target we supply
    size_t n_tgts;		// The number of targets we supply
    const size_t n_db_ftrs;	// Number of features in the database we use 
    const size_t n_db_lbls;	// Number of labels in the database
    const size_t n_db_tgts;	// Number of targets in the database
    const size_t buf_size;	// Size of buffer (in sentences)
    const size_t n_sents;	// The number of sentences in the database
    size_t buf_frames;		// Length of buffer in frames
    float* ftr_buf;		// Buffer to use for features
    QNUInt32* lbl_buf;		// Buffer to use for labels
    int narrow_ftrs;		// 1 if narrowing features
    int narrow_lbls;		// 1 if narrowing labels
    static const char* CLASSNAME; // Name of the class (for error reporting) 
};


#endif // #ifdef QN_SentDB_Narrow_h_INCLUDED
