// $Header: /n/lasagna/db/davidj/spert/src/quicknet/H-spert0/../RCS/QN_SentDB_Norm.h,v 1.3 1996/03/01 03:37:40 davidj Exp $

#ifndef QN_SentDB_Norm_h_INCLUDED
#define QN_SentDB_Norm_h_INCLUDED

/* Must include the config.h file first */
#include <QN_config.h>
#include <assert.h>
#include <stddef.h>
#include <stdio.h>
#include "QN_types.h"
#include "QN_SentDB.h"

// QN_SentDB_Norm is a class used to normalize a QN_SentDB (i.e.
// multiply each of the feature values by a constant specific to that
// feature number)

class QN_SentDB_Norm : public QN_SentDB
{
public:
    // Constructor.
    // a_db - the database we are norming
    // norm_mean - the mean of the input vector - subtract this element wise
    //             from all frames
    // norm_invsd - the standard deviation of the input vector - multiply
    //              all values by this element-wise
    QN_SentDB_Norm(int dbg, QN_SentDB& a_db,
		   const float* norm_mean, const float* norm_sdevs);
    ~QN_SentDB_Norm();

    // Return the total number of frames in the database.
    size_t num_frames();
    // Return the total number of sentences in the database.
    size_t num_sentences();
    // Return the number of frames in a range of sentences.
    size_t num_frames_in_sents(size_t first_sent, size_t num_sents);
    // Return the number of features in a frame..
    size_t num_features_in_frame();
    // Return the number of labels in a frame.
    size_t num_labels_in_frame();
    // Return the number of targets in a frame.
    size_t num_targets_in_frame();

    // Return the labels and features for a given sentence.
    // If ftrs is NULL, do not get features.
    // If labs is NULL, do not get labels.
    size_t get_sentences(size_t sentno, size_t num_sents,
			 QNUInt32* sentnos, QNUInt32* framenos,
			 float* ftrs, QNUInt32* labs, float* tgts);
private:
    int debug;			// Level of verbosity
    QN_SentDB& db;		// The database we are subsetting.
    const size_t n_features;	// The number of features in the db.
    float* const norm_nmean;	// The negated mean - add this.
    float* const norm_invsd;	// The inverse standard deviation - multiply
				// by this.

    static const char* CLASSNAME; // Name of the class (for error
				  // reporting).
};



#endif // #ifdef QN_SentDB_Norm_h_INCLUDED
