const char* QN_mat_rcsid =
    "$Header: /n/lasagna/db/davidj/spert/src/quicknet/RCS/QN_mat.cc,v 1.3 1996/09/24 20:41:19 davidj Exp $";

/* Must include the config.h file first */
#include <QN_config.h>
#include <assert.h>
#include <errno.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "intvec.h"
#include "fltvec.h"
#include "QN_types.h"
#include "QN_mat.h"

inline static size_t
min(size_t a, size_t b)
{
    if (a<=b)
	return a;
    else
	return b;
}

inline static void
copy_svf_vf(size_t len, size_t stride, const float* in, float* out)
{
    intv_int32_t* from = (intv_int32_t*) in;
    intv_int32_t* to = (intv_int32_t*) out;

    while(len--)
    {
	*to = *from;
	to++;
	from += stride;
    }
}

QN_Mat_Header::QN_Mat_Header(QNUInt32 a_type, QNUInt32 a_mrows,
			     QNUInt32 a_mcols, QNUInt32 a_imagf,
			     QNUInt32 a_namlen)
{
    type = a_type;
    mrows = a_mrows;
    mcols = a_mcols;
    imagf = a_imagf;
    namlen = a_namlen;
}

QN_Mat_OutMatrix::QN_Mat_OutMatrix(FILE* a_file, const char* a_name,
				   QNUInt32 a_type,
				   size_t a_rows, size_t a_cols,
				   int a_complex,
				   size_t a_bufsize)
    : // We save in native format and rely on the reader to convert.
      hdr(a_type + QN_MAT_TYPE_NATIVE + QN_MAT_TYPE_FULL,
	  a_rows, a_cols, a_complex, strlen(a_name)+1),
      file(a_file),
      name(a_name),
      count(0),
      bufsize(a_bufsize),
      floatbuf(NULL)
{
    size_t cnt;			// Number of items written.

    // Currently only support floats.
    assert(a_type==QN_MAT_TYPE_FLOAT);
    assert (hdr.mrows!=0 && hdr.mcols!=0);

    floatbuf = new float[bufsize];
    cnt = fwrite(&hdr, sizeof(hdr), 1, file);
    if (cnt!=1)
    {	
	QN_ERROR("QN_Mat_OutMatrix",
		 "failed to write matrix '%s' header - %s.",
		 name, strerror(errno));
    }
    cnt = fwrite(name, hdr.namlen, 1, file);
    if (cnt!=1)
    {	
	QN_ERROR("QN_Mat_OutMatrix",
		 "failed to write matrix '%s' name - %s.",
		 name, strerror(errno));
    }
}

QN_Mat_OutMatrix::~QN_Mat_OutMatrix()
{
    // Check that we output all values.
    if (count != (hdr.mrows*hdr.mcols))
    {
	QN_ERROR("QN_Mat_OutMatrix",
		 "incorrect number of values written to matrix '%s'.",
		 name);
    }
    if (floatbuf!=NULL)
	delete[] floatbuf;
}

void
QN_Mat_OutMatrix::output(size_t a_count, const float* a_vals)
{
    size_t cnt;			// Number of items written.

    cnt = fwrite(a_vals, sizeof(float), a_count, file);
    if (cnt!=a_count)
    {	
	QN_ERROR("QN_Mat_OutMatrix",
		 "failed to write contents of matrix '%s'- %s.",
		 name, strerror(errno));
    }
    count += a_count;
}

void
QN_Mat_OutMatrix::dump(size_t rows, size_t cols, float* mat)
{
    size_t i;			// Current column.

    for (i=0; i<cols; i++)	// Iterate over columns.
    {
	size_t j;		// Current row within column.
	float* col;		// Pointer to current part of column.

	col = &(mat[i]);
	for (j=0; j<rows; j+=bufsize) // Iterate over chunks of one column.
	{
	    size_t len = min(rows-j, bufsize); // 
	    
	    copy_svf_vf(len, cols, col, floatbuf);
	    output(len, floatbuf);
	    col += len*cols;
	}
    }
}
