// $Header: /u/drspeech/src/SPRACHcore/quicknet/RCS/QN_mat.h,v 1.4 1999/11/07 06:52:51 dpwe Exp $

#ifndef QN_mat_h_INCLUDED
#define QN_mat_h_INCLUDED

// Some stuff to aid in the reading and writing of Matlab ".mat" files.

/* Must include the config.h file first */
#include <QN_config.h>
#include <stddef.h>
#include <stdio.h>
#include "QN_types.h"
#include "QN_Logger.h"


// Some definitions for the "type" field in the matrix header.
enum {

// The machine type.
    QN_MAT_TYPE_IEEELITTLE = 0,
    QN_MAT_TYPE_IEEEBIG = 1000,
#ifdef INTV_WORDS_BIGENDIAN
    QN_MAT_TYPE_NATIVE = QN_MAT_TYPE_IEEEBIG,
#else
    QN_MAT_TYPE_NATIVE = QN_MAT_TYPE_IEEELITTLE,
#endif

// The data type.
    QN_MAT_TYPE_DOUBLE = 0,
    QN_MAT_TYPE_FLOAT = 10,
    QN_MAT_TYPE_INT32 = 20,
    QN_MAT_TYPE_INT16 = 30,
    QN_MAT_TYPE_UINT16 = 40,
    QN_MAT_TYPE_UINT8 = 50,

// The layout.
    QN_MAT_TYPE_FULL = 0,
    QN_MAT_TYPE_TEXT = 2,
    QN_MAT_TYPE_SPARSE = 3
    
};

// The header that goes at the start of each matrix.
struct QN_Mat_Header
{
    QNUInt32 type;
    QNUInt32 mrows;
    QNUInt32 mcols;
    QNUInt32 imagf;
    QNUInt32 namlen;
    // Just a constructor for the header.
    QN_Mat_Header(QNUInt32 a_type, QNUInt32 a_mrows, QNUInt32 a_mcols,
		 QNUInt32 a_imagf, QNUInt32 a_namlen);
};

// A class that is used to write a matrix to a file.

class QN_Mat_OutMatrix
{
public:
    enum { DEFAULT_BUFSIZE = 256 }; // The size of our internal buffer.
    // Create a matrix.
    // a_file = stream to create matrix on.
    // a_name = name of matrix.
    // a_type = the type of the values in the file
    // (typically QN_MAT_TYPE_FLOAT).
    // a_rows = number of rows.
    // a_cols = number of cols.
    // a_complex = set to one for complex matrix.
    QN_Mat_OutMatrix(FILE* a_file, const char* a_name, QNUInt32 a_type,
		     size_t a_rows, size_t a_cols,
		     int a_complex = 0,
		     size_t a_bufsize = QN_Mat_OutMatrix::DEFAULT_BUFSIZE);
    // Finish off the matrix.
    ~QN_Mat_OutMatrix();

    // Write a given number of values to the output matrix.  Note that
    // output is column major - i.e. we have to transpose traditional C
    // matrices.
    void output(size_t a_count, const float* a_vals);
    // Output a whole floating point matrix.
    void dump(size_t rows, size_t cols, float* mat);
private:
    const QN_Mat_Header hdr;	// The header information.
    FILE* file;			// The file on which we write the data.
    const char* name;		// The name of the matrix.
    size_t count;		// Count of values output so far.
    size_t bufsize;		// Size of the buffer used for conversions.
    float* floatbuf;		// A buffer for conversions.
};

// Some functions to encapsulate the whole matrix output operation.
static inline void
QN_Mat_dump(FILE* file, const char* name, size_t rows, size_t cols,
	    float* mat, size_t bufsize = QN_Mat_OutMatrix::DEFAULT_BUFSIZE)
{
    QN_Mat_OutMatrix om(file, name, QN_MAT_TYPE_FLOAT,
			rows, cols, 0, bufsize);
    om.dump(rows, cols, mat);
}


#endif // #define QN_mat_h_INCLUDED

